/*
* Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include <linux/printk.h>
#if defined(__linux__)
 #include <linux/version.h>
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0))
  #include <linux/stdarg.h>
 #else
  #include <stdarg.h>
 #endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0) */
#else
 #include <stdarg.h>
#endif /* __linux__ */
#include <linux/device.h>
#include <linux/kern_levels.h>

#include "inno_plat_dev.h"

#define LOG_MAX_PRINT_TIMES 20

void inno_dev_printk(char *debug_level, inno_dev *dev, const char *format, ...);

void inno_printk(const char *format, ...);

void inno_printk_ratelimited(const char *format, ...);

#define inno_error(fmt, ...) \
	do { \
		inno_printk(KERN_ERR fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_info(fmt, ...) \
	do { \
		inno_printk(KERN_INFO fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_emerg(fmt, ...) \
	do { \
		inno_printk(KERN_EMERG fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_alert(fmt, ...) \
	do { \
		inno_printk(KERN_ALERT fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_crit(fmt, ...) \
	do { \
		inno_printk(KERN_CRIT fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_warning(fmt, ...) \
	do { \
		inno_printk(KERN_WARNING fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_notice(fmt, ...) \
	do { \
		inno_printk(KERN_NOTICE fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_debug(fmt, ...) \
	do { \
		inno_printk(KERN_DEBUG fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_error_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_ERR fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_info_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_INFO fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_emerg_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_EMERG fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_alert_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_ALERT fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_crit_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_CRIT fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_warning_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_WARNING fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_notice_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_NOTICE fmt, ##__VA_ARGS__); \
	} while (0);

#define inno_debug_ratelimited(fmt, ...) \
	do { \
		inno_printk_ratelimited(KERN_DEBUG fmt, ##__VA_ARGS__); \
	} while (0);


