/*
* Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <linux/moduleparam.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <asm/div64.h>
#include <linux/bug.h>
#include <linux/err.h>
#include <linux/utsname.h>
#include <linux/kfifo.h>
#include <linux/module.h>
#include <linux/random.h>
#include "inno_misc.h"

uint32_t gPVRDebugLevel = 0;
module_param(gPVRDebugLevel, uint, 0644);
MODULE_PARM_DESC(gPVRDebugLevel,
		"Sets the level of debug output (default 0x7)");
INNO_EXT_SYM(gPVRDebugLevel);

int inno_memcmp(void *buf0, void *buf1, uint32_t size)
{
	return memcmp(buf0, buf1, size);
}
INNO_EXT_SYM(inno_memcmp);

uint64_t inno_strlcat(char *dst, const char *src, uint64_t size)
{
	return strlcat(dst, src, size);
}
INNO_EXT_SYM(inno_strlcat);

char *inno_strchr(const char *s, int c)
{
	return strchr(s, c);
}
INNO_EXT_SYM(inno_strchr);

char *inno_strrchr(const char *s, int c)
{
	return strrchr(s, c);
}
INNO_EXT_SYM(inno_strrchr);

char *inno_strstr(const char *s1, const char *s2)
{
	return strstr(s1, s2);
}
INNO_EXT_SYM(inno_strstr);

int inno_sprintf(char *buf, size_t size, const char *fmt, ...)
{
	va_list arg_list;
	int count;

	va_start(arg_list, fmt);
	count = inno_vsnprintf(buf, size, fmt, arg_list);
	va_end(arg_list);

	return count;
}
INNO_EXT_SYM(inno_sprintf);

int inno_vsnprintf(char *buf, uint64_t size, const char *fmt, va_list args)
{
	return vsnprintf(buf, size, fmt, args);
}
INNO_EXT_SYM(inno_vsnprintf);

char *inno_kasprintf(gfp_t gfp, const char *fmt, ...)
{
	va_list ap;
	char *p;

	va_start(ap, fmt);
	p = kvasprintf(gfp, fmt, ap);
	va_end(ap);

	return p;
}
INNO_EXT_SYM(inno_kasprintf);

int inno_snprintf(char *buf, size_t size, const char *fmt, ...)
{
	va_list args;
	int i;

	va_start(args, fmt);
	i = vsnprintf(buf, size, fmt, args);
	va_end(args);

	return i;
}
INNO_EXT_SYM(inno_snprintf);


uint64_t inno_strlen(const char *str)
{
	return strlen(str);
}
INNO_EXT_SYM(inno_strlen);

uint64_t inno_strnlen(const char *str, uint64_t cnt)
{
	return strnlen(str, cnt);
}
INNO_EXT_SYM(inno_strnlen);

int inno_strncmp(const char *str1, const char *str2, uint64_t size)
{
	return strncmp(str1, str2, size);
}
INNO_EXT_SYM(inno_strncmp);

int inno_strcmp(const char *str1, const char *str2)
{
	return strcmp(str1, str2);
}
INNO_EXT_SYM(inno_strcmp);

int inno_kstrtou32(const char *s, unsigned int base, uint32_t *res)
{
	return kstrtou32(s, base, res);
}
INNO_EXT_SYM(inno_kstrtou32);

int inno_kstrtou64(const char *s, unsigned int base, u64 *res)
{
	return kstrtou64(s, base, res);
}
INNO_EXT_SYM(inno_kstrtou64);

char *inno_strsep(char **s, const char *ct)
{
	return strsep(s, ct);
}
INNO_EXT_SYM(inno_strsep);

int inno_strcasecmp(const char *s1, const char *s2)
{
	return strcasecmp(s1, s2);
}
INNO_EXT_SYM(inno_strcasecmp);

int inno_strlcpy(char *dest, const char *src, int size)
{
	return strlcpy(dest, src, size);
}
INNO_EXT_SYM(inno_strlcpy);

char *inno_strncpy(char *dest, const char *src, int count)
{
	return strncpy(dest, src, count);
}
INNO_EXT_SYM(inno_strncpy);

int inno_vsscanf(const char *buf, const char *fmt, va_list args)
{
	return vsscanf(buf, fmt, args);
}
INNO_EXT_SYM(inno_vsscanf);

int inno_sscanf(const char *buf, const char *fmt, ...)
{
	va_list args;
	int i;

	va_start(args, fmt);
	i = vsscanf(buf, fmt, args);
	va_end(args);

	return i;
}
INNO_EXT_SYM(inno_sscanf);

bool inno_is_err(const void *ptr)
{
	return IS_ERR(ptr);
}
INNO_EXT_SYM(inno_is_err);

bool inno_is_err_or_null(const void *ptr)
{
	return IS_ERR_OR_NULL(ptr);
}
INNO_EXT_SYM(inno_is_err_or_null);

long inno_ptr_err(const void *ptr)
{
	return PTR_ERR(ptr);
}
INNO_EXT_SYM(inno_ptr_err);

void inno_bug(void)
{
	BUG();
}
INNO_EXT_SYM(inno_bug);

void inno_bug_on(int condition)
{
	BUG_ON(condition);
}
INNO_EXT_SYM(inno_bug_on);

int inno_warn_on(int condition)
{
	return WARN_ON(condition);
}
INNO_EXT_SYM(inno_warn_on);

int inno_warn_on_once(int condition)
{
	return WARN_ON_ONCE(condition);
}
INNO_EXT_SYM(inno_warn_on_once);

const char *inno_utsname_sysname(void)
{
	return utsname()->sysname;
}
INNO_EXT_SYM(inno_utsname_sysname);

const char *inno_utsname_release(void)
{
	return utsname()->release;
}
INNO_EXT_SYM(inno_utsname_release);

const char *inno_utsname_version(void)
{
	return utsname()->version;
}
INNO_EXT_SYM(inno_utsname_version);

const char *inno_utsname_machine(void)
{
	return utsname()->machine;
}
INNO_EXT_SYM(inno_utsname_machine);

bool inno_kernel_version_ge(int x, int y, int z)
{
	return LINUX_VERSION_CODE >= KERNEL_VERSION(x, y, z);
}
INNO_EXT_SYM(inno_kernel_version_ge);

bool inno_kernel_version_gt(int x, int y, int z)
{
	return LINUX_VERSION_CODE > KERNEL_VERSION(x, y, z);
}
INNO_EXT_SYM(inno_kernel_version_gt);

bool inno_kernel_version_lt(int x, int y, int z)
{
	return LINUX_VERSION_CODE < KERNEL_VERSION(x, y, z);
}
INNO_EXT_SYM(inno_kernel_version_lt);

void inno_atomic64_set(atomic64_t *v, long long i)
{
	atomic64_set(v, i);
}
INNO_EXT_SYM(inno_atomic64_set);

void inno_atomic64_add(long long i, atomic64_t *v)
{
	atomic64_add(i, v);
}
INNO_EXT_SYM(inno_atomic64_add);

void inno_atomic64_sub(long long i, atomic64_t *v)
{
	atomic64_sub(i, v);
}
INNO_EXT_SYM(inno_atomic64_sub);

long long inno_atomic64_read(const atomic64_t *v)
{
	return atomic64_read(v);
}
INNO_EXT_SYM(inno_atomic64_read);

void inno_kill_fasync(void **fp)
{
	kill_fasync((struct fasync_struct **)fp, SIGIO, POLL_IN);
}
INNO_EXT_SYM(inno_kill_fasync);

unsigned int inno_kfifo_len(void *fifo)
{
	return kfifo_len((struct kfifo *)fifo);
}
INNO_EXT_SYM(inno_kfifo_len);

int inno_kfifo_is_full(void *fifo)
{
	return kfifo_is_full((struct kfifo *)fifo);
}
INNO_EXT_SYM(inno_kfifo_is_full);

void inno_kfifo_reset(void *fifo)
{
	kfifo_reset((struct kfifo *)fifo);
}
INNO_EXT_SYM(inno_kfifo_reset);

unsigned int inno_kfifo_in_spinlocked(void *kfifo, void *buf, unsigned long n, void *lock)
{
	return kfifo_in_spinlocked((struct kfifo *)kfifo, buf, n, (spinlock_t *)lock);
}
INNO_EXT_SYM(inno_kfifo_in_spinlocked);

unsigned int inno_kfifo_out_spinlocked(void *kfifo, void *buf, unsigned long n, void *lock)
{
	return kfifo_out_spinlocked((struct kfifo *)kfifo, buf, n, (spinlock_t *)lock);
}
INNO_EXT_SYM(inno_kfifo_out_spinlocked);

unsigned long inno_simple_strtoul(const char *cp, char **endp, unsigned int base)
{
	return simple_strtoull(cp, endp, base);
}
INNO_EXT_SYM(inno_simple_strtoul);

void inno_this_module_put(void)
{
	module_put(THIS_MODULE);
}
INNO_EXT_SYM(inno_this_module_put);

void inno_smp_mb(void)
{
	smp_mb();
}
INNO_EXT_SYM(inno_smp_mb);

void inno_smp_wmb(void)
{
	smp_wmb();
}
INNO_EXT_SYM(inno_smp_wmb);

void inno_smp_rmb(void)
{
	smp_rmb();
}
INNO_EXT_SYM(inno_smp_rmb);

bool inno_is_aligned(uint64_t val, int aligned)
{
	return IS_ALIGNED((unsigned long)val, aligned);
}
INNO_EXT_SYM(inno_is_aligned);

int inno_num_possible_cpus(void)
{
	return num_possible_cpus();
}
INNO_EXT_SYM(inno_num_possible_cpus);

void inno_get_random_bytes(void *buf, int nbytes)
{
	get_random_bytes(buf, nbytes);
}
INNO_EXT_SYM(inno_get_random_bytes);

#if defined (__i386__) || defined(__x86_64__)
extern unsigned int tsc_khz;
unsigned int inno_tsc_khz(void)
{
	return tsc_khz;
}
INNO_EXT_SYM(inno_tsc_khz);
#endif

uint64_t inno_div_round_up(uint64_t val, uint64_t div)
{
	return DIV_ROUND_UP(val, div);
}
INNO_EXT_SYM(inno_div_round_up);

bool inno_test_and_set_bit(unsigned long nr, volatile unsigned long *addr)
{
	return test_and_set_bit(nr, addr);
}
INNO_EXT_SYM(inno_test_and_set_bit);

bool inno_test_and_clear_bit(unsigned long nr, volatile unsigned long *addr)
{
	return test_and_clear_bit(nr, addr);
}
INNO_EXT_SYM(inno_test_and_clear_bit);

void inno_set_bit(unsigned long nr, volatile unsigned long *addr)
{
	set_bit(nr, addr);
}
INNO_EXT_SYM(inno_set_bit);

void inno_clear_bit(unsigned long nr, volatile unsigned long *addr)
{
	clear_bit(nr, addr);
}
INNO_EXT_SYM(inno_clear_bit);
