/*
* Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
* Dual MIT/GPLv2
*
* The contents of this file are subject to the MIT license as set out below.
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
*
* Alternatively, the contents of this file may be used under the terms of
* the GNU General Public License Version 2 ("GPL") in which case the provisions
* of GPL are applicable instead of those above.
*
* If you wish to allow use of your version of this file only under the terms of
* GPL, and not to allow others to use your version of this file under the terms
* of the MIT license, indicate your decision by deleting the provisions above
* and replace them with the notice and other provisions required by GPL as set
* out in the file called "GPL-COPYING" included in this distribution. If you do
* not delete the provisions above, a recipient may use your version of this file
* under the terms of either the MIT license or GPL.
*
* This License is also included in this distribution in the file called
* "MIT-COPYING".
*
* EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
* PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
* BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
* PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
* COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
* IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
* CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#ifndef __INNO_MISC_H__
#define __INNO_MISC_H__

#if defined(__linux__)
 #include <linux/version.h>
 #if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0))
  #include <linux/stdarg.h>
 #else
  #include <stdarg.h>
 #endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0) */
#else
 #include <stdarg.h>
#endif /* __linux__ */

#include <linux/types.h>

/* for compatibility: move few math functions from inno_misc.c to inno_math.c */
#include "inno_math.h"

#if defined(SEPARATE_BUILD) || defined(CONFIG_DRM_INNO_APU)
 #define INNO_EXT_SYM(f) EXPORT_SYMBOL(f)
 #define INNO_EXT_TRACEPOINT_SYM(f) EXPORT_TRACEPOINT_SYMBOL(f)
#else
 #define INNO_EXT_SYM(f)
 #define INNO_EXT_TRACEPOINT_SYM(f)
#endif

#define INNO_ALWAYS_EXT_SYM(f) EXPORT_SYMBOL(f)

extern uint32_t gPVRDebugLevel;

int inno_memcmp(void *buf0, void *buf1, uint32_t size);

uint64_t inno_strlcat(char *dst, const char *src, uint64_t size);

char *inno_strchr(const char *s, int c);

char *inno_strrchr(const char *s, int c);

char *inno_strstr(const char *s1, const char *s2);

int inno_sprintf(char *buf, size_t size, const char *fmt, ...);

int inno_vsnprintf(char *buf, uint64_t size, const char *fmt, va_list args);

uint64_t inno_strlen(const char *str);

uint64_t inno_strnlen(const char *str, uint64_t cnt);

int inno_strcmp(const char *str1, const char *str2);

int inno_strncmp(const char *str1, const char *str2, uint64_t size);

int inno_kstrtou32(const char *s, unsigned int base, uint32_t *res);

int inno_kstrtou64(const char *s, unsigned int base, u64 *res);

char *inno_strsep(char **s, const char *ct);

int inno_strcasecmp(const char *s1, const char *s2);

int inno_strlcpy(char *dest, const char *src, int size);

char *inno_strncpy(char *dest, const char *src, int count);

int inno_vsscanf(const char *buf, const char *fmt, va_list args);

int inno_sscanf(const char *buf, const char *fmt, ...);

bool inno_is_err(const void *ptr);

bool inno_is_err_or_null(const void *ptr);

long inno_ptr_err(const void *ptr);

void inno_bug(void);

void inno_bug_on(int condition);

int inno_warn_on(int condition);

int inno_warn_on_once(int condition);

void inno_dump_stack(void);

const char *inno_utsname_sysname(void);

const char *inno_utsname_release(void);

const char *inno_utsname_version(void);

const char *inno_utsname_machine(void);

bool inno_kernel_version_ge(int x, int y, int z);

bool inno_kernel_version_gt(int x, int y, int z);

bool inno_kernel_version_lt(int x, int y, int z);

void inno_atomic64_set(atomic64_t *v, long long i);

void inno_atomic64_add(long long i, atomic64_t *v);

void inno_atomic64_sub(long long i, atomic64_t *v);

long long inno_atomic64_read(const atomic64_t *v);

void inno_kill_fasync(void **fp);

unsigned int inno_kfifo_len(void *fifo);

int inno_kfifo_is_full(void *fifo);

void inno_kfifo_reset(void *fifo);

unsigned int inno_kfifo_in_spinlocked(void *kfifo, void *buf, unsigned long n, void *lock);

unsigned int inno_kfifo_out_spinlocked(void *kfifo, void *buf, unsigned long n, void *lock);

char *inno_kasprintf(gfp_t gfp, const char *fmt, ...);

int inno_snprintf(char *buf, size_t size, const char *fmt, ...);

unsigned long inno_simple_strtoul(const char *cp, char **endp, unsigned int base);

void inno_this_module_put(void);

void inno_smp_mb(void);

void inno_smp_wmb(void);

void inno_smp_rmb(void);

bool inno_is_aligned(uint64_t val, int aligned);

int inno_num_possible_cpus(void);

int inno_get_unused_fd_flags_o_cloexec(void);

#if defined (__i386__) || defined(__x86_64__)
unsigned int inno_tsc_khz(void);
#endif

uint64_t inno_div_round_up(uint64_t val, uint64_t div);

bool inno_test_and_set_bit(unsigned long nr, volatile unsigned long *addr);

bool inno_test_and_clear_bit(unsigned long nr, volatile unsigned long *addr);

void inno_set_bit(unsigned long nr, volatile unsigned long *addr);

void inno_clear_bit(unsigned long nr, volatile unsigned long *addr);
#endif
