/*
 * Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
 * Dual MIT/GPLv2
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License Version 2 ("GPL") in which case the provisions
 * of GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms of
 * GPL, and not to allow others to use your version of this file under the terms
 * of the MIT license, indicate your decision by deleting the provisions above
 * and replace them with the notice and other provisions required by GPL as set
 * out in the file called "GPL-COPYING" included in this distribution. If you do
 * not delete the provisions above, a recipient may use your version of this file
 * under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT-COPYING".
 *
 * EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
 * PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <generated/utsrelease.h>
#include <linux/pci.h>
#include <uapi/linux/pci.h>
#include <uapi/linux/pci_regs.h>
#include <linux/version.h>
#include "inno_misc.h"
#include "inno_pci.h"
#include "inno_debug.h"
#define KBUILD_INNO_PCI "inno_pci"
#define pr_fmt_inno_pci(fmt) "[%s][%s:%d]" fmt,KBUILD_INNO_PCI,__func__,__LINE__

#define inno_pci_notice(dev,fmt, ...)\
		inno_dev_printk(KERN_NOTICE,dev,pr_fmt_inno_pci(fmt), ##__VA_ARGS__)
#define inno_pci_error(dev,fmt, ...)\
		inno_dev_printk(KERN_ERR,dev,pr_fmt_inno_pci(fmt), ##__VA_ARGS__)

int INNO_IRQ_MODE_LEGACY = PCI_IRQ_LEGACY;
int INNO_IRQ_MODE_MSI = PCI_IRQ_MSI;

int inno_pci_enable_device(inno_pci_dev *dev)
{
	return pci_enable_device((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_enable_device);

void inno_pci_set_master(inno_pci_dev *dev)
{
	pci_set_master((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_set_master);

int inno_pci_enable_msi(inno_pci_dev *dev)
{
	return pci_enable_msi((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_enable_msi);

inno_pci_dev * inno_pci_get_device(unsigned int vendor, unsigned int device)
{
	struct pci_dev *dev;
	dev = pci_get_device(vendor, device, NULL);
	return dev;
}
INNO_EXT_SYM(inno_pci_get_device);

unsigned long long inno_pci_resource_len(inno_pci_dev *dev, int bar)
{
	return pci_resource_len((struct pci_dev *)dev, bar);
}
INNO_EXT_SYM(inno_pci_resource_len);

unsigned long long inno_pci_resource_start(inno_pci_dev *dev, int bar)
{
	return pci_resource_start((struct pci_dev *)dev, bar);
}
INNO_EXT_SYM(inno_pci_resource_start);

unsigned long long inno_pci_resource_end(inno_pci_dev *dev, int bar)
{
	return pci_resource_end((struct pci_dev *)dev, bar);
}
INNO_EXT_SYM(inno_pci_resource_end);

unsigned long inno_pci_resource_flags(inno_pci_dev *dev, int bar)
{
	return pci_resource_flags((struct pci_dev *)dev, bar);
}
INNO_EXT_SYM(inno_pci_resource_flags);

int inno_pci_request_region(inno_pci_dev *dev, int bar, const char *res_name)
{
	return pci_request_region((struct pci_dev *)dev, bar, res_name);
}
INNO_EXT_SYM(inno_pci_request_region);

void inno_pci_release_region(inno_pci_dev *dev, int bar)
{
	pci_release_region((struct pci_dev *)dev, bar);
}
INNO_EXT_SYM(inno_pci_release_region);

void inno_pci_disable_msi(inno_pci_dev *dev)
{
	pci_disable_msi((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_disable_msi);

void inno_pci_clear_master(inno_pci_dev *dev)
{
	pci_clear_master((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_clear_master);

void inno_pci_restore_state(inno_pci_dev *dev)
{
	pci_restore_state((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_restore_state);

void inno_pci_disable_device(inno_pci_dev *dev)
{
	pci_disable_device((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_disable_device);

int inno_pci_save_state(inno_pci_dev *dev)
{
	return pci_save_state((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_save_state);

int inno_pci_set_power_state(inno_pci_dev *dev, enum inno_pmsg_state state)
{
	switch (state) {
	case INNO_PMSG_ON:
		return pci_set_power_state((struct pci_dev *)dev, pci_choose_state((struct pci_dev *)dev, PMSG_ON));
	case INNO_PMSG_SUSPEND:
		return pci_set_power_state((struct pci_dev *)dev, pci_choose_state((struct pci_dev *)dev, PMSG_SUSPEND));
	default:
		return INNO_PMSG_NONE;
	}
}
INNO_EXT_SYM(inno_pci_set_power_state);

bool inno_pci_is_busmaster(inno_pci_dev *dev)
{
	return ((struct pci_dev*)dev)->is_busmaster;
}
INNO_EXT_SYM(inno_pci_is_busmaster);

unsigned int inno_get_pci_irq(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->irq;
}
INNO_EXT_SYM(inno_get_pci_irq);

unsigned short inno_get_pci_vendor(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->vendor;
}
INNO_EXT_SYM(inno_get_pci_vendor);

unsigned short inno_get_pci_device(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->device;
}
INNO_EXT_SYM(inno_get_pci_device);

unsigned short inno_get_pci_subvendor(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->subsystem_vendor;
}
INNO_EXT_SYM(inno_get_pci_subvendor);

unsigned short inno_get_pci_subdevice(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->subsystem_device;
}
INNO_EXT_SYM(inno_get_pci_subdevice);

void inno_arch_phys_wc_del(int handle)
{
	arch_phys_wc_del(handle);
}
INNO_EXT_SYM(inno_arch_phys_wc_del);

void inno_arch_io_free_memtype_wc(unsigned long long start, unsigned long long size)
{
	arch_io_free_memtype_wc(start, size);
}
INNO_EXT_SYM(inno_arch_io_free_memtype_wc);

void * inno_request_region(unsigned long long start, unsigned long long n, const char * name)
{
	return request_region(start, n, name);
}
INNO_EXT_SYM(inno_request_region);

void * inno_request_mem_region(unsigned long long start, unsigned long long n, const char *name)
{
	return request_mem_region(start, n, name);
}
INNO_EXT_SYM(inno_request_mem_region);

void inno_release_region(unsigned long long start, unsigned long long n)
{
	release_region(start, n);
}
INNO_EXT_SYM(inno_release_region);

void inno_release_mem_region(unsigned long long start, unsigned long long n)
{
	release_mem_region(start, n);
}
INNO_EXT_SYM(inno_release_mem_region);

int inno_pci_irq_vector(inno_pci_dev *dev, unsigned int nr)
{
	return pci_irq_vector((struct pci_dev *)dev, nr);
}
INNO_EXT_SYM(inno_pci_irq_vector);

inno_pci_dev *inno_to_pci_dev(void *dev)
{
	return (inno_pci_dev *)to_pci_dev((struct device *)dev);
}
INNO_EXT_SYM(inno_to_pci_dev);

uint8_t inno_pci_revision_id(inno_pci_dev *dev)
{
	return ((struct pci_dev *)dev)->revision;
}
INNO_EXT_SYM(inno_pci_revision_id);

int inno_pci_alloc_irq_vectors(inno_pci_dev *dev, unsigned int min_vecs, unsigned int max_vecs, int mode)
{
	return pci_alloc_irq_vectors((struct pci_dev *)dev, min_vecs, max_vecs, mode);
}
INNO_EXT_SYM(inno_pci_alloc_irq_vectors);

void inno_pci_free_irq_vectors(inno_pci_dev *dev)
{
	return pci_free_irq_vectors((struct pci_dev *)dev);
}
INNO_EXT_SYM(inno_pci_free_irq_vectors);

/*
 * When detect KylinOS2203-55 kernel and loongson 3a5000/7a2000 pcie bridge, return true (meaning pcie intx irq).
 * */
bool inno_loongarch_kylin2203_55_intx_irq(inno_pci_dev *dev)
{
#if defined(CONFIG_LOONGARCH) && defined(CONFIG_KYLINOS_DESKTOP)
#if (LINUX_VERSION_CODE == KERNEL_VERSION(5, 4, 18))
#if (UTS_KYLINOS_RELEASE_ABI == 55)
	struct pci_dev *pdev = (struct pci_dev *)dev;
	struct pci_bus *root = pdev->bus;

	while (root && root->self) {
		/*
		 * when root is pcie bridge, root->parent and root->self is non-null
		 * when root is pcie rc, root->parent and root->self is null
		 * 7a2000 pcie bridge vendorid: 0x14 deviceid: 0x7a59
		 * 7a1000 pcie bridge vendorid: 0x14 deviceid: 0x7a29
		 */
		if (root->self->vendor == 0x14 && root->self->device == 0x7a59) {
			return true;
		}

		root = root->parent;
	}
#endif
#endif
#endif

	return false;
}
INNO_EXT_SYM(inno_loongarch_kylin2203_55_intx_irq);


extern struct resource *pci_bus_resource_n(const struct pci_bus *bus, int n);
#define inno_pci_bus_for_each_resource(bus, res, i)				\
	for (i = 0;							\
		(res = pci_bus_resource_n(bus, i)) || i < PCI_BRIDGE_RESOURCE_NUM; \
		i++)

void inno_pci_release_resource(struct pci_dev *dev, int resno)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	pci_release_resource(dev, resno);
#else
	// null
#endif

}

int inno_pci_resize_resource(struct pci_dev *dev, int resno, int size)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	return pci_resize_resource(dev, resno, size);
#else
	return 1;
#endif
}

int inno_check_rc_size(inno_pci_dev *dev)
{
	struct pci_bus *root;
	struct resource *res;
	struct pci_dev *pdev;
	unsigned i;

	if (dev == NULL)
		return -1;

	pdev = (struct pci_dev *)dev;
	/* Check if the root BUS has 64bit memory resources */
	root = pdev->bus;
	while (root->parent)
		root = root->parent;

	inno_pci_bus_for_each_resource(root, res, i) {
		if (res && res->flags & (IORESOURCE_MEM | IORESOURCE_MEM_64) &&
			res->start > 0x100000000ull)
			break;
	}

	/* Trying to resize is pointless without a root hub window above 4GB */
	if (!res)
		return 1;

	return 0;
}

INNO_EXT_SYM(inno_check_rc_size);

int inno_check_resize_version(void)
{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	return 0;
#else
	return 1;
#endif
}
INNO_EXT_SYM(inno_check_resize_version);

#define INNO_RESIZE_NUMS (0x1)
int inno_resize_bar(struct pci_dev *pdev, int resno, int rbar_size)
{
	u16 cmd;
	int ret = 0;
	u32 i;

	/* Disable memory decoding while we change the BAR addresses and size */
	pci_read_config_word(pdev, PCI_COMMAND, &cmd);
	pci_write_config_word(pdev, PCI_COMMAND,
				cmd & ~PCI_COMMAND_MEMORY);

	inno_pci_release_resource(pdev, resno);
	inno_pci_release_resource(pdev, 0);
	//inno_pci_error(&pdev->dev, " -ENOTSUPP:%d, -EBUSY:%d, -EINVAL:%d \n", (-ENOTSUPP), (-EBUSY), (-EINVAL) );
	for (i = 0; i < INNO_RESIZE_NUMS; i++)
	{

		ret = inno_pci_resize_resource(pdev, resno, rbar_size);
		if(ret == -ENOSPC)
		{
			inno_pci_notice(&pdev->dev, "Not enough PCI address space for a large BAR.\n");
		}else if (ret && ret != -ENOTSUPP) {
				inno_pci_notice(&pdev->dev, "Problem resizing BAR2 (%d).retry:%d\n", ret, i);
		}else{
			inno_pci_notice(&pdev->dev, "pci_resize_resource cfg success \n");
			break;
		}

	}

	pci_assign_unassigned_bus_resources(pdev->bus);
	pci_write_config_word(pdev, PCI_COMMAND, cmd);

	return ret;
}

int inno_gpu_device_resize_fb_bar(inno_pci_dev *dev, int resno, int rbar_size, int *repo)
{
	struct pci_dev *pdev;
	int ret;

	if ((dev == NULL) || (repo == NULL))
		return -1;

	pdev = (struct pci_dev *)dev;
	ret = inno_resize_bar(pdev, resno, rbar_size);
	*repo = ret;

	return 0;
}

INNO_EXT_SYM(inno_gpu_device_resize_fb_bar);
