/*
 * Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
 * Dual MIT/GPLv2
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License Version 2 ("GPL") in which case the provisions
 * of GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms of
 * GPL, and not to allow others to use your version of this file under the terms
 * of the MIT license, indicate your decision by deleting the provisions above
 * and replace them with the notice and other provisions required by GPL as set
 * out in the file called "GPL-COPYING" included in this distribution. If you do
 * not delete the provisions above, a recipient may use your version of this file
 * under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT-COPYING".
 *
 * EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
 * PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef __INNO_PCI_H__
#define __INNO_PCI_H__

enum inno_pmsg_state{
	INNO_PMSG_ON,
	INNO_PMSG_SUSPEND,
	INNO_PMSG_NONE,
};

extern int INNO_IRQ_MODE_LEGACY;
extern int INNO_IRQ_MODE_MSI;

typedef void inno_pci_dev;

int inno_pci_enable_device(inno_pci_dev *dev);

void inno_pci_set_master(inno_pci_dev *dev);

int inno_pci_enable_msi(inno_pci_dev *dev);

inno_pci_dev * inno_pci_get_device(unsigned int vendor, unsigned int device);

int inno_check_resize_version(void);

int inno_check_rc_size(inno_pci_dev *dev);

int inno_gpu_device_resize_fb_bar(inno_pci_dev *dev, int resno, int rbar_size, int *repo);

unsigned long long inno_pci_resource_len(inno_pci_dev *dev, int bar);

unsigned long long inno_pci_resource_start(inno_pci_dev *dev, int bar);

unsigned long long inno_pci_resource_end(inno_pci_dev *dev, int bar);

unsigned long inno_pci_resource_flags(inno_pci_dev *dev, int bar);

int inno_pci_request_region(inno_pci_dev *dev, int bar, const char *res_name);

void inno_pci_release_region(inno_pci_dev *dev, int bar);

void inno_pci_disable_msi(inno_pci_dev *dev);

void inno_pci_clear_master(inno_pci_dev *dev);

void inno_pci_restore_state(inno_pci_dev *dev);

void inno_pci_disable_device(inno_pci_dev *dev);

int inno_pci_save_state(inno_pci_dev *dev);

int inno_pci_set_power_state(inno_pci_dev *dev, enum inno_pmsg_state state);

bool inno_pci_is_busmaster(inno_pci_dev *dev);

unsigned int inno_get_pci_irq(inno_pci_dev *dev);

unsigned short inno_get_pci_vendor(inno_pci_dev *dev);

unsigned short inno_get_pci_subvendor(inno_pci_dev *dev);

unsigned short inno_get_pci_device(inno_pci_dev *dev);

unsigned short inno_get_pci_subdevice(inno_pci_dev *dev);

int inno_pci_irq_vector(inno_pci_dev *dev, unsigned int nr);

void inno_arch_phys_wc_del(int handle);

void inno_arch_io_free_memtype_wc(unsigned long long start, unsigned long long size);

void * inno_request_region(unsigned long long start, unsigned long long n, const char * name);

void * inno_request_mem_region(unsigned long long start, unsigned long long n, const char *name);

void inno_release_region(unsigned long long start, unsigned long long n);

void inno_release_mem_region(unsigned long long start, unsigned long long n);

inno_pci_dev *inno_to_pci_dev(void *dev);

uint8_t inno_pci_revision_id(inno_pci_dev *dev);

bool inno_loongarch_kylin2203_55_intx_irq(inno_pci_dev *dev);

int inno_pci_alloc_irq_vectors(inno_pci_dev *dev, unsigned int min_vecs, unsigned int max_vecs, int mode);

void inno_pci_free_irq_vectors(inno_pci_dev *dev);
#endif
