/*
 * Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
 * Dual MIT/GPLv2
 *
 * The contents of this file are subject to the MIT license as set out below.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * the GNU General Public License Version 2 ("GPL") in which case the provisions
 * of GPL are applicable instead of those above.
 *
 * If you wish to allow use of your version of this file only under the terms of
 * GPL, and not to allow others to use your version of this file under the terms
 * of the MIT license, indicate your decision by deleting the provisions above
 * and replace them with the notice and other provisions required by GPL as set
 * out in the file called "GPL-COPYING" included in this distribution. If you do
 * not delete the provisions above, a recipient may use your version of this file
 * under the terms of either the MIT license or GPL.
 *
 * This License is also included in this distribution in the file called
 * "MIT-COPYING".
 *
 * EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
 * PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#include <linux/platform_device.h>
#include "inno_misc.h"
#include "inno_plat_dev.h"

inno_platform_device *inno_platform_device_register_full(
		struct inno_plat_dev_info *_info)
{
	struct resource res = DEFINE_RES_MEM_NAMED(0, 0, "empty");
	struct platform_device_info info = {
		.parent    = _info->parent,
		.name      = _info->name,
		.id        = _info->id,
		.data      = _info->data,
		.size_data = _info->data_size,
		.dma_mask  = _info->dma_mask,
	};

	if (_info->res) {
		res.start    = _info->res->start;
		res.end      = _info->res->start + _info->res->size - 1;
		res.name     = _info->res->name;
		info.res     = &res;
		info.num_res = 1;
	}

	return platform_device_register_full(&info);
}
INNO_EXT_SYM(inno_platform_device_register_full);

void inno_platform_device_unregister(inno_platform_device *dev)
{
	platform_device_unregister((struct platform_device *)dev);
}
INNO_EXT_SYM(inno_platform_device_unregister);

void *inno_platform_get_parent_dev(inno_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return pdev->dev.parent;
}
INNO_EXT_SYM(inno_platform_get_parent_dev);

void *inno_platform_get_dev(inno_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return &pdev->dev;
}
INNO_EXT_SYM(inno_platform_get_dev);

inno_dev *inno_dev_get_parent(inno_dev *dev)
{
	return ((struct device *)dev)->parent;
}
INNO_EXT_SYM(inno_dev_get_parent);

inno_platform_device *inno_to_platform_device(void *dev)
{
	struct device *_dev = (struct device *)dev;
	return to_platform_device(_dev);
}
INNO_EXT_SYM(inno_to_platform_device);

void *inno_platform_get_data(inno_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return pdev->dev.platform_data;
}
INNO_EXT_SYM(inno_platform_get_data);

uint64_t inno_platform_get_dma_mask(inno_platform_device *dev)
{
	struct platform_device *pdev = (struct platform_device *)dev;
	return *pdev->dev.dma_mask;
}
INNO_EXT_SYM(inno_platform_get_dma_mask);

inno_resource *inno_platform_get_resource_byname(inno_platform_device *dev,
		unsigned int type, const char *name)
{
	return platform_get_resource_byname((struct platform_device *)dev,
		type, name);
}
INNO_EXT_SYM(inno_platform_get_resource_byname);

uint64_t inno_resource_start(inno_resource *res)
{
	return ((struct resource *)res)->start;
}
INNO_EXT_SYM(inno_resource_start);

uint64_t inno_resource_size(inno_resource *res)
{
	return resource_size((struct resource *)res);
}
INNO_EXT_SYM(inno_resource_size);

static void inno_rsrc_devres_release(struct device *dev, void *res)
{
	/* No extra cleanup needed */
}

void *inno_rsrc_devres_find(inno_dev *dev)
{
	return devres_find(dev, inno_rsrc_devres_release, NULL, NULL);
}
INNO_EXT_SYM(inno_rsrc_devres_find);

void *inno_rsrc_devres_alloc(size_t size)
{
	return devres_alloc(inno_rsrc_devres_release, size, GFP_KERNEL);
}

struct device *inno_dev_to_dev(inno_dev *dev)
{
	return (struct device *)dev;
}
INNO_EXT_SYM(inno_dev_to_dev);

void inno_put_device(inno_dev *dev)
{
	put_device((struct device *)dev);
}
EXPORT_SYMBOL(inno_put_device);

inno_dev *inno_get_device(inno_dev *dev)
{
	return (inno_dev *)get_device((struct device *)dev);
}
EXPORT_SYMBOL(inno_get_device);

void inno_platform_set_drvdata(void *pdev, void *data)
{
	platform_set_drvdata((struct platform_device *)pdev, data);
}
INNO_EXT_SYM(inno_platform_set_drvdata);

void *inno_platform_get_drvdata(void *pdev)
{
	return platform_get_drvdata((struct platform_device *)pdev);
}
INNO_EXT_SYM(inno_platform_get_drvdata);

void *inno_dev_get_platdata(void *pdev)
{
	return (void *)dev_get_platdata((struct device *)pdev);
}
INNO_EXT_SYM(inno_dev_get_platdata);

void *inno_get_dev_ofnode(inno_dev *dev)
{
	return ((struct device *)dev)->of_node;
}
EXPORT_SYMBOL(inno_get_dev_ofnode);

void *inno_get_platdev_drvdata(void *pdev)
{
	return dev_get_drvdata(&((struct platform_device *)pdev)->dev);
}
INNO_EXT_SYM(inno_get_platdev_drvdata);

void *inno_get_dev_drvdata(inno_dev *dev)
{
	return dev_get_drvdata((struct device *)dev);
}
INNO_EXT_SYM(inno_get_dev_drvdata);

void inno_set_dev_drvdata(inno_dev *dev, void *data)
{
	dev_set_drvdata((struct device *)dev, data);
}
INNO_EXT_SYM(inno_set_dev_drvdata);

int inno_platform_get_numa_node(inno_platform_device *dev)
{
	int node_node = 0;
#ifdef CONFIG_NUMA
	struct platform_device *pdev = (struct platform_device *)dev;
	node_node = pdev->dev.numa_node;
#endif
	return node_node;
}
INNO_EXT_SYM(inno_platform_get_numa_node);

