/*************************************************************************/ /*!
@File			innodpu_connector.h
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#ifndef __INNODPU_CONNECTOR_H__
#define __INNODPU_CONNECTOR_H__
#include <drm/drm_modes.h>
#include "innodpu_common.h"

#define DPU_CONNECT_NUM  (8)

#define DIMENSION_G1_SOC	(0)
#define DIMENSION_G0_SOC	(1)
#define DIMENSION_G1P_SOC	(2)
#define DIMENSION_G0M_SOC	(3)
#define DIMENSION_MAX		(4)

#define CONNECTOR_MAP_ITEM(type, width, height, crtc) [type] = {width, height, BIT(crtc)}
#define CONNECTOR_MAP_EMPTY_ITEM(type) [type] = {0, 0, 0}

#define CONNECTOR_ID_ITEM(type, id) [type] = (id)

#define CUSTOM_OUTPUT_MODE_0  (0)
#define CUSTOM_OUTPUT_MODE_1  (1)
#define CUSTOM_OUTPUT_MODE_2  (2)
#define CUSTOM_OUTPUT_MODE_3  (3)
#define CUSTOM_OUTPUT_MODE_4  (4)
#define CUSTOM_OUTPUT_MODE_5  (5)
#define CUSTOM_OUTPUT_MODE_6  (6)
#define CUSTOM_OUTPUT_MODE_7  (7)
#define CUSTOM_OUTPUT_MODE_8  (8)
#define CUSTOM_OUTPUT_MODE_9  (9)
#define CUSTOM_OUTPUT_MODE_10 (10)
#define CUSTOM_OUTPUT_MODE_11 (11)
#define CUSTOM_OUTPUT_MODE_12 (12)
#define CUSTOM_OUTPUT_MODE_13 (13)
#define CUSTOM_OUTPUT_MODE_14 (14)
#define CUSTOM_OUTPUT_MODE_15 (15)
#define CUSTOM_OUTPUT_MODE_16 (16)
#define CUSTOM_OUTPUT_MODE_MAX (17)

#define DISPLAY_MODE_LEN 	(32)
#define MONITOR_NAME_LEN 	(13)

#define ODM_INFO_LEN		(10)

enum {
	HDMI0_ZOOM_ENABLE = 0,
	HDMI1_ZOOM_ENABLE,
	DP0_ZOOM_ENABLE,
	LVDS_ZOOM_ENABLE,
	HDMI2_ZOOM_ENABLE,
	HDMI3_ZOOM_ENABLE,
	DP1_ZOOM_ENABLE,
	VGA_ZOOM_ENABLE,
};

enum connector_module {
	CONNECTOR_M_HDMI0 = 0,
	CONNECTOR_M_HDMI1,
	CONNECTOR_M_HDMI2,
	CONNECTOR_M_HDMI3,
	CONNECTOR_M_DP0,
	CONNECTOR_M_DP1,
	CONNECTOR_M_VGA,
	CONNECTOR_M_LVDS,
};

enum connector_backlight_mode {
	CONNECTOR_BACKLIGHT_DDCCI = 0,
	CONNECTOR_BACKLIGHT_PWM0,
	CONNECTOR_BACKLIGHT_AUX_VESA,
	CONNECTOR_BACKLIGHT_AUX_HDR,
};

struct connector_map_prop {
	unsigned int max_width;
	unsigned int max_height;
	unsigned int possible_crtc;
};

struct monitor_non_replace {
	char monitor_name[MONITOR_NAME_LEN];
	char non_replace_name[DISPLAY_MODE_LEN];
	int clock;
};

struct innodpu_mode_bypass_list {
	const char *monitor;
	const char *odm_info;
	chip_type_e plat;
	int hdisplay;
	int vdisplay;
	int vrefresh;
};

struct connector_output_mode {
	int mode_connector_item;
	unsigned char convert_name[20];

#define CONNECTOR_FLAGS_NONE       (0x0U)
	/*
	 * for some conversion chip, such as CS5801(HDMI2DP) etc.
	 * when DP cable is removed and the DP2HDMI's HPD has been pulled down,
	 * we can still read EDID from the chip.
	 * So we must skip ddc detect, and wait longer time after HPD IRQ.
	 * */
#define CONNECTOR_FLAGS_SKIP_DDC   (0x1U)
#define CONNECTOR_FLAGS_LONGER_HPD (0x2U)

	/* 1.kylin system for hot-plugging processing,single-screen, dual-screen phenomenon is
	 * inconsistent, single-screen unplug the monitor, the display interface will not go
	 * disable, source will always send stream data.
	 * 2. ch7517 firmware, hpd pin high and low by the display is connected, whether to read
	 * the display edid, dp source this side of the main link whether there is a data stream
	 * of the three conditions, only all three are satisfied, ch7517 unreset after the hpd pin will
	 * be pulled high!
	 * 3. bug id:7013
	 * */
#define CONNECTOR_FLAGS_EN_PLUGIN (0x4U)

	unsigned int flags;

	/*
	 * Optional.
	 * when flags is CONNECTOR_FLAGS_LONGER_HPD, the data is delay_ms.
	 * etc.
	 * */
	int data;
};


struct hw_board_info {
	const char *odm_vendor;
	const char *pcb_version;
};

#define HW_ODM_VENDOR_LEN    (4U)
#define HW_PCB_VERSION_LEN  (28U)
#define HW_BOARD_INFO_ITEM(_ov, _pv) {.odm_vendor=(_ov), .pcb_version=(_pv)}

bool is_output_type_hdmi(struct connector_output_mode *om);
bool is_output_type_dvi(struct connector_output_mode *om);
bool is_output_type_vga(struct connector_output_mode *om);
bool is_output_type_dp(struct connector_output_mode *om);
bool is_output_type_edp(struct connector_output_mode *om);
bool is_output_type_lvds(struct connector_output_mode *om);
bool is_flags_skip_ddc_detect(struct connector_output_mode *om);
bool is_flags_wait_longer_hpd(struct connector_output_mode *om);
bool is_flags_data_en_plugin(struct connector_output_mode *om);

bool innodpu_is_odm_pcb_match(inno_dev *pcie_dev, struct hw_board_info *board_info);
bool inno_mode_equal(const struct drm_display_mode *mode1,
		    const struct drm_display_mode *mode2);
int innodpu_custom_init(struct drm_device *drm_dev, struct innodpu_drm_private *dev_priv);
struct connector_map_prop *innodpu_find_connector_map_module(struct device *dev, enum connector_module module);
int innodpu_str_push_edid(const char *buf, struct drm_connector *connector);
int innodpu_modes_noedid(struct drm_connector *connector);
int innodpu_get_display_mode(void);

bool innodpu_modes_is_match(const struct drm_display_mode *mode1,
			    const struct drm_display_mode *mode2);
int innodpu_modes_drop_repeat(struct drm_connector *connector);
int innodpu_modes_replace(struct drm_connector *connector,
		const struct monitor_non_replace *non_replace_ptr, int count);

int innodpu_get_mfc_name(struct device *device, char *name, unsigned char *edid);
int innodpu_get_odm_info(struct device *device, char *odm_info, inno_dev *dev);
bool innodpu_mode_need_bypass(const char *monitor, char *omd_info,
	chip_type_e plat, struct innodpu_mode_bypass_list *mode_list,
	int list_len, int hdisplay, int vdisplay, int vrefresh);

int innodpu_connector_add_common_modes(struct drm_connector *connector,
		struct drm_display_mode *native_mode, bool support_force);

struct drm_display_mode *
innodpu_get_native_mode(struct drm_connector *connector,
		bool (*filter)(struct drm_connector *pconnector, struct drm_display_mode *pmode, bool priv_filter));

struct connector_output_mode *innodpu_get_connector_output_mode(inno_dev *dev, unsigned int reg_module);
int innodpu_get_connector_backlight_mode(inno_dev *dev, unsigned int reg_module);

int innodpu_detailed_block_monitor_range_replace(u8 *raw_edid);
int innodpu_conn_get_monitor_max_clk(u8 *raw_edid);
void innodpu_connector_clear_eld(struct drm_connector *connector);

#endif

