/*************************************************************************/ /*!
@File			innodpu_drm_gem.h
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#ifndef __INNODPU_DRM_GEM_H
#define __INNODPU_DRM_GEM_H

#include <linux/dma-buf.h>
#include "inno_drm_version.h"

#if (DRM_VERSION >= KERNEL_VERSION(5, 5, 0))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#endif
#if (DRM_VERSION >= KERNEL_VERSION(5, 5, 0))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif

#if (DRM_VERSION >= KERNEL_VERSION(3, 18, 0))
#include <drm/drm_gem.h>
#endif

#include "innodpu_common.h"

#include "pvr_dma_resv.h"
#include "hal_interface.h"
#include "kernel_compatibility.h"
#define SCHEDULE_FREE
#define DUMB_USED_GTT			BIT(27)
#define DUMB_USED_INVISIABLE	BIT(28)
#define GEM_MSYNC_PER_SIZE 		(0x100000)
#define GEM_ZERO_SIZE			(0x100000)
#define GEM_USED_VMALLOC


/**
 * DRM_MODE_FMT - printf string for &struct inno_gem_object
 */
#define DRM_VRAM_FMT "%pK:visiable-%d dev_addr %lld"

/**
 * DRM_VRAM_ARG - printf arguments for &inno_gem_object
 * @m: display mode
 */
#define DRM_VRAM_ARG(m) \
	(m), (m)->gem_priv->visiable, &((m)->dev_addr)


extern struct vm_operations_struct inno_gem_vm_ops;

struct gem_vm_list {
	struct list_head list;
	unsigned long vm_vaddr;
	unsigned long offset;
	unsigned long dev_addr;
	void * vaddr;
	int size;
};

struct inno_gem_object {
	char *name;
	struct drm_gem_object base;
	struct list_head vm_head;
	struct mutex vm_lock;
	struct list_head gem_mm_node;
	void *vm_wsync_dst[256];
	void *vm_wsync_src[256];
	int vm_wsync_len[256];
	struct gem_vm_list *pvm_arr[256];
	void *axi_vram_dst[256];
	void *axi_vram_src[256];

	/* 如果当前gem_pvir使用drm mm方式，则保留节点信息 */
	struct drm_mm_node node;

	/* Non-null if backing was imported */
	struct sg_table *sgt;

	bool dma_map_export_host_addr;
	phys_addr_t cpu_addr;
	dma_addr_t dev_addr;

#if (DRM_VERSION < KERNEL_VERSION(5, 2, 0))
	struct dma_resv _resv;		// 4.14 版本使用这个resv，5.2版本及以上使用drm_gem_object中的_resv;
#endif
	struct dma_resv *resv;

	bool cpu_prep;
	spinlock_t lock;
	int dfd;
	struct dma_buf *tmp_dbuf;
	struct list_head fb_free_node;
	struct list_head fb_pm_node;

	void *suspend_data;
	struct inno_gem_private *gem_priv;

	bool is_fbdev_obj;

	u32 flags;
	/* use for control gtt memory */
	struct gtt_item *gtt;
};
#define to_inno_obj(obj) container_of(obj, struct inno_gem_object, base)



// gem init
struct inno_gem_private *inno_gem_init(struct drm_device *drm_dev, bool visiable,
		unsigned char flags);
void inno_gem_fini(struct drm_device *drm_dev, struct inno_gem_private *gem_priv);

struct dma_resv *inno_gem_get_resv(struct drm_gem_object *obj);
unsigned long inno_gem_get_dev_addr(struct drm_gem_object * obj);
unsigned long inno_gem_get_cpu_addr(struct drm_gem_object *obj);



// buffer_create and free
int inno_gem_dumb_create(struct drm_file *file_priv, struct drm_device *drm_dev,
		struct inno_gem_private *gem_priv, struct drm_mode_create_dumb *args);

int inno_gem_create_priv(struct drm_file *file_priv, struct drm_device *drm_dev,
		struct inno_gem_private *gem_priv, void *data);

void inno_gem_object_free_priv(struct drm_gem_object *obj);

#if (DRM_VERSION <= KERNEL_VERSION(5, 2, 0))
struct reservation_object * inno_gem_prime_res_obj( struct drm_gem_object *gem_obj);
#endif
void inno_gem_object_free_priv(struct drm_gem_object *gem_obj);







int inno_gem_vram_count(int id, void *ptr, void *data);

void inno_sysmem_thaw_free(struct inno_gem_private *gem_priv);
void inno_gem_recover(struct inno_gem_private *gem_priv);
void inno_gem_suspend(struct inno_gem_private *gem_priv);
void inno_gem_backup(struct inno_gem_private *gem_priv);
void inno_gem_backup_shmem(struct inno_gem_private *gem_priv);

void inno_gem_backup_zero_vram(struct inno_gem_private *gem_priv);
void inno_gem_recover_zero_vram(struct inno_gem_private *gem_priv);

void inno_gem_prime_close_object(struct drm_gem_object *obj, struct drm_file *filp);


int inno_gem_object_cpu_prep_ioctl(struct drm_device *dev, void *data, struct drm_file *file);
int inno_gem_object_cpu_fini_ioctl(struct drm_device *dev, void *data, struct drm_file *file);
int inno_gem_object_dump_vram_ioctl(struct drm_device *dev, void *data, struct drm_file *file);
int inno_gem_object_inv_get_ioctl(struct drm_device *dev, void *data, struct drm_file *file);

/* drm driver functions */
struct drm_gem_object *inno_gem_object_create(struct drm_device *dev,
											  struct inno_gem_private *gem_priv,
											  size_t size, u32 flags);

struct dma_buf *inno_gem_prime_export(
#if (DRM_VERSION < KERNEL_VERSION(5, 4, 0))
										 struct drm_device *dev,
#endif
										 struct drm_gem_object *obj, int flags);

struct drm_gem_object *inno_gem_prime_import(struct drm_device *dev, struct dma_buf *dma_buf);

struct drm_gem_object *inno_gem_prime_import_sg_table(struct drm_device *dev,
													  struct dma_buf_attachment *attach,
													  struct sg_table *sgt);



#if (DRM_VERSION < KERNEL_VERSION(5, 4, 0))

// 生产者给dma-buf创建efence
//extern int inno_gem_efence_create(struct dma_buf *dbuf);

// 根据fd 获取cpu_addr和dev_addr
static inline int inno_gem_pfd_2_addr(int pfd, unsigned long *baddr, unsigned long *caddr,
									  unsigned long *psize)
{
	struct dma_buf *dbuf;
	struct inno_gem_object *iobj;
	struct drm_gem_object *obj;

	dbuf = dma_buf_get(pfd);

	if (!dbuf)
		return -ENOENT;
	dma_buf_put(dbuf);

	obj = dbuf->priv;
	if (!obj)
		return -ENOENT;
	iobj = to_inno_obj(obj);

	if (baddr)
		*baddr = iobj->dev_addr;
	if (caddr)
		*caddr = iobj->cpu_addr;
	if (psize)
		*psize = obj->size;

	return 0;
}

// 根据fd 去清理掉fence
static inline int inno_gem_pfd_clear_fence(int pfd)
{
	struct dma_fence *fence_excl;

	struct dma_buf *dbuf = dma_buf_get(pfd);
	if (!dbuf) {
		return -ENOENT;
	}
	if (!dbuf->resv) {
		inno_error("pfd %d dbuf->resv is null\n", pfd);
		dma_buf_put(dbuf);
		return 0;
	}
	inno_printk("pfd %d dbuf->resv, dbuf->resv %p\n", pfd, dbuf->resv);
	fence_excl = dma_resv_get_excl(dbuf->resv);
	fence_excl->flags = 0;		// clear excl fence

	// fixed me : need clear shared  fence
	dma_buf_put(dbuf);

	return 0;
}

// 根据handle 获取cpu_addr和dev_addr
static inline int inno_gem_handle_2_devaddr(struct drm_file *file, u32 handle, dma_addr_t * addr)
{
	struct inno_gem_object *iobj;
	struct drm_gem_object *obj;

	obj = drm_gem_object_lookup(file, handle);
	if (!obj)
		return -ENOENT;
	iobj = to_inno_obj(obj);
	drm_gem_object_put(obj);

	*addr = iobj->dev_addr;
	return 0;
}

static inline int inno_gem_handle_2_cpuaddr(struct drm_file *file, u32 handle, phys_addr_t * addr)
{
	struct inno_gem_object *iobj;
	struct drm_gem_object *obj;

	obj = drm_gem_object_lookup(file, handle);
	if (!obj)
		return -ENOENT;
	iobj = to_inno_obj(obj);
	drm_gem_object_put(obj);

	*addr = iobj->cpu_addr;
	return 0;
}
#endif

static size_t __maybe_unused inno_gem_handle_get_size(struct drm_file *file, u32 handle)
{
	struct drm_gem_object *obj;

	obj = drm_gem_object_lookup(file, handle);
	if (!obj)
		return -ENOENT;
	drm_gem_object_put(obj);

	return obj->size;
}

#if 0
// 发送signal信号
static inline int inno_gem_signal_efence(struct dma_buf *dbuf)
{
	struct dma_fence *fence_excl;

	// inno_error("!!!!!!!!!!!!!may be kernel panic....1\n");
	// struct dma_buf *dbuf;
	// dbuf = dma_buf_get(pfd); // 这个接口好像会引起休眠，导致卡死
	if (s_drm_debug) {
#if (DRM_VERSION < KERNEL_VERSION(4, 16, 0))
		inno_printk("[%s %d]: dbuf is %pK",  __func__, __LINE__, dbuf);
#else
		inno_printk("[%s %d]: dbuf is %pK",  __func__, __LINE__, dbuf);
#endif
	}
	if (!dbuf) {
		inno_error("!!!!!!!!!!!!!not wake\n");
		return -ENOENT;
	}

	fence_excl = dma_resv_get_excl(dbuf->resv);

	if (s_drm_debug)
		inno_error("!!!!!!!!!!!!!wake excl fence flag %#lx\n", fence_excl->flags);
	//if (fence_excl && test_bit(DMA_FENCE_FLAG_ENABLE_SIGNAL_BIT, &fence_excl->flags)){
	if (fence_excl) {
		dma_fence_signal(fence_excl);	// excl will return pollin
		// inno_error("!!!!!!!!!!!!!excl fence flag %#lx\n", fence_excl->flags);
		dma_fence_put(fence_excl);	// 释放fence
	}

	dma_buf_put(dbuf);

	return 0;
}
#endif

#endif //__INNODPU_DRM_GEM_H
