/*************************************************************************/ /*!
@File			innodpu_connector.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include "innodpu_connector.h"
#include "innodpu_common.h"
#include "innodpu_dp_common.h"
#include "innogpu_drm.h"

extern unsigned int s_dpu_match;

static const struct drm_display_mode edid_cvtrb_modes[] = {
	/* 1920x1080@60Hz 16:9 */
	{DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008,
			  2052, 2200, 0, 1080, 1084, 1089, 1125, 0,
			  DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	 .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9,},

	{ DRM_MODE("2240x1400", DRM_MODE_TYPE_DRIVER, 212500, 2240, 2240+48,
			   2240+48+32, 2240+220, 0, 1400, 1400+3, 1400+3+6, 1400+40, 0,
			   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC) },

	{ DRM_MODE("2240x1400", DRM_MODE_TYPE_DRIVER, 141696, 2240, 2240+48,
			   2240+48+32, 2240+220, 0, 1400, 1400+3, 1400+3+6, 1400+40, 0,
			   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC) },
	/*
		support for F4
		CVT-RB Modeline 	   Modeline "2160x1400_29.98" 98.75 2160 2208 2240 2320 1400 1403 1413 1420 +HSync -VSync
		CVT-RB Modeline 	   Modeline "2160x1400_59.94" 200.25 2160 2208 2240 2320 1400 1403 1413 1440 +HSync -VSync
		CVT-RB Modeline        Modeline "2160x1400_74.93" 252.25 2160 2208 2240 2320 1400 1403 1413 1451 +HSync -VSync
	*/
	/* 2160x1400@30Hz 16:9 */
	{ DRM_MODE("2160x1400", DRM_MODE_TYPE_DRIVER, 98750, 2160, 2208,
		   2240, 2320, 0, 1400, 1403, 1413, 1420, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = 0, }, // invalid picture_aspect_ratio

	/* 2160x1400@60Hz 16:9 */
	{ DRM_MODE("2160x1400", DRM_MODE_TYPE_DRIVER, 200250, 2160, 2208,
		   2240, 2320, 0, 1400, 1403, 1413, 1440, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = 0, }, // invalid picture_aspect_ratio

	/* 2160x1400@75Hz 16:9 */
	{ DRM_MODE("2160x1400", DRM_MODE_TYPE_DRIVER, 252250, 2160, 2208,
		   2240, 2320, 0, 1400, 1403, 1413, 1451, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = 0, }, // invalid picture_aspect_ratio


	/* 2560x1440@30Hz 16:9 */
	{ DRM_MODE("2560x1440", DRM_MODE_TYPE_DRIVER, 119000, 2560, 2608,
		   2640, 2720, 0, 1440, 1443, 1448, 1461, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },

	/* 2560x1440@60Hz 16:9 */
	{ DRM_MODE("2560x1440", DRM_MODE_TYPE_DRIVER, 241500, 2560, 2608,
		   2640, 2720, 0, 1440, 1443, 1448, 1481, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },

	/* 2560x1440@75Hz 16:9 */
	{ DRM_MODE("2560x1440", DRM_MODE_TYPE_DRIVER, 304250, 2560, 2608,
		   2640, 2720, 0, 1440, 1443, 1448, 1492, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },

	/* 2560x1600@30Hz 16:9 */
	{ DRM_MODE("2560x1600", DRM_MODE_TYPE_DRIVER, 128541, 2560, 2568,
		   2600, 2640, 0, 1600, 1609, 1617, 1623, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
	/* 2560x1600@60Hz 16:9 */
	{ DRM_MODE("2560x1600", DRM_MODE_TYPE_DRIVER, 260726, 2560, 2568,
		   2600, 2640, 0, 1600, 1632, 1640, 1646, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
	/* 2560x1600@75Hz 16:9 */
	{ DRM_MODE("2560x1600", DRM_MODE_TYPE_DRIVER, 328284, 2560, 2568,
		   2600, 2640, 0, 1600, 1644, 1652, 1658, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_NVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
	/* 3840x2160@30Hz 16:9 */
	{ DRM_MODE("3840x2160", DRM_MODE_TYPE_DRIVER, 297000, 3840, 4016,
		   4104, 4400, 0, 2160, 2168, 2178, 2250, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
	/* 3840x2160@60Hz 16:9 */
	{ DRM_MODE("3840x2160", DRM_MODE_TYPE_DRIVER, 594000, 3840, 4016,
		   4104, 4400, 0, 2160, 2168, 2178, 2250, 0,
		   DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	  .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
	/* 3840x1080@60Hz 32:9*/
	{ DRM_MODE("3840x1080", DRM_MODE_TYPE_DRIVER, 292350, 3840, 4018,
		  4178, 4388, 0, 1080, 1083, 1093, 1111, 0,
		  DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
	 .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
};

static bool innodpu_drm_match_timings(const struct drm_display_mode *mode1,
				const struct drm_display_mode *mode2)
{
	return mode1->hdisplay == mode2->hdisplay &&
		mode1->hsync_start == mode2->hsync_start &&
		mode1->hsync_end == mode2->hsync_end &&
		mode1->htotal == mode2->htotal &&
		mode1->vdisplay == mode2->vdisplay &&
		mode1->vsync_start == mode2->vsync_start &&
		mode1->vsync_end == mode2->vsync_end &&
		mode1->vtotal == mode2->vtotal;
}
static bool innodpu_modes_match_clock(const struct drm_display_mode *mode1,
			    const struct drm_display_mode *mode2)
{
	if (mode1->clock && mode2->clock) {
		return KHZ2PICOS(mode1->clock) == KHZ2PICOS(mode2->clock);
	} else {
		return mode1->clock == mode2->clock;
	}
}

bool inno_mode_equal(const struct drm_display_mode *mode1,
		    const struct drm_display_mode *mode2)
{
	if (!mode1 || !mode2) {
		return false;
	}

	if (!innodpu_drm_match_timings(mode1, mode2)) {
		return false;
	}

	if (!innodpu_modes_match_clock(mode1, mode2)) {
		return false;
	}
	return true;
}

bool innodpu_modes_is_match(const struct drm_display_mode *mode1,
			    const struct drm_display_mode *mode2)
{
	if (mode1->hdisplay == mode2->hdisplay && \
		mode1->vdisplay == mode2->vdisplay && \
		abs(drm_mode_vrefresh(mode1) - drm_mode_vrefresh(mode2)) <= 1) {
		return true;
	}

	return false;
}


int innodpu_modes_drop_repeat(struct drm_connector *connector)
{
	struct drm_display_mode *mode1 = NULL;
	struct drm_display_mode *mode2 = NULL;
	struct drm_display_mode *m = NULL;
	int nums = 0;

	list_for_each_entry(mode1, &connector->probed_modes, head) {
		list_for_each_entry_safe(mode2, m, &connector->probed_modes, head) {
			if (!mode1 || !mode2 || mode1 == mode2)
				continue;

			if (innodpu_modes_is_match(mode1, mode2)) {
				list_del(&mode2->head);
				drm_mode_destroy(connector->dev, mode2);
				nums ++;
			}
		}
	}

	return nums;
}


static void innodpu_modes_replace_timing(struct drm_display_mode *mode)
{
	int i = 0;
	const struct drm_display_mode *rmode = NULL;

	for (i = 1; i < ARRAY_SIZE(edid_cvtrb_modes); i++) {
		rmode = &edid_cvtrb_modes[i];

		/*the code could be deleted, because it could be replace when equal.*/
		if (inno_mode_equal(mode, rmode))
			continue;

		if (innodpu_modes_is_match(mode, rmode)) {
			mode->clock = rmode->clock;
			mode->hdisplay = rmode->hdisplay;
			mode->hsync_start = rmode->hsync_start;
			mode->hsync_end = rmode->hsync_end;
			mode->htotal = rmode->htotal;
			mode->hskew = rmode->hskew;
			mode->vdisplay = rmode->vdisplay;
			mode->vsync_start = rmode->vsync_start;
			mode->vsync_end = rmode->vsync_end;
			mode->vtotal = rmode->vtotal;
			mode->vscan = rmode->vscan;
			mode->flags = rmode->flags;
			mode->picture_aspect_ratio = rmode->picture_aspect_ratio;
			mode->type = (mode->type & DRM_MODE_TYPE_PREFERRED) ? \
						 (rmode->type | DRM_MODE_TYPE_PREFERRED) : rmode->type;
			break;
		}
	}
}

int innodpu_modes_replace(struct drm_connector *connector,
		const struct monitor_non_replace *p, int count)
{
	int i = 0;
	bool replace = true;
	struct drm_display_mode *n = NULL;
	struct drm_display_mode *mode1 = NULL;

	list_for_each_entry_safe(mode1, n, &connector->probed_modes, head) {
		replace = true;
		for (i = 0; i < count; i++) {
			if (!inno_strcmp(p[i].non_replace_name, mode1->name) &&
				p[i].clock == mode1->clock) {
				replace = false;
				break;
			}
		}

		if (replace && !drm_match_cea_mode(mode1) && \
			drm_mode_validate_size(mode1, INNODPU_COMBINE_WIDTH, INNODPU_COMBINE_HEIGHT)) {
			innodpu_modes_replace_timing(mode1);
		}
	}

	return 0;
}

int innodpu_modes_noedid(struct drm_connector *connector)
{
	int index = 0;
	int num_modes = 0;
	const struct drm_display_mode *ptr = NULL;
	struct drm_display_mode *mode = NULL;
	struct drm_device *drm_dev = connector->dev;

	for (index = 0, num_modes = 0; index < 1; ++index) {
		ptr = &edid_cvtrb_modes[index];
		if ((ptr->hdisplay > drm_dev->mode_config.max_width) ||
				(ptr->vdisplay > drm_dev->mode_config.max_height)) {
			continue;
		}
		mode = drm_mode_duplicate(connector->dev, ptr);
		if (mode) {
			drm_mode_probed_add(connector, mode);
			++num_modes;
		}
	}
	return num_modes;
}

int innodpu_str_push_edid(const char *buf, struct drm_connector *connector)
{
	struct inno_drm_display_mode {
		int clock;
		u16 hdisplay;
		u16 hsync_start;
		u16 hsync_end;
		u16 htotal;
		u16 hskew;
		u16 vdisplay;
		u16 vsync_start;
		u16 vsync_end;
		u16 vtotal;
		u16 vscan;
		u32 flags;
	};

	int i = 0;
	int num_modes = 0;
	struct drm_display_mode *ptr = NULL;
	struct drm_display_mode *mode = NULL;
	struct inno_drm_display_mode inno_drm_mode = {0};
	int inno_display_size = sizeof(struct inno_drm_display_mode);

	ptr = kzalloc(sizeof(struct drm_display_mode), GFP_KERNEL);
	if (IS_ERR(ptr)) {
		innodpu_err(connector->dev->dev, "[%s %d] \n", __func__, __LINE__);
		return 0;
	}

	for (i = 0; (i < INNODP_EDID_BUF_LEN) &&
		((INNODP_EDID_BUF_LEN - i) >= inno_display_size);
		i += inno_display_size) {
		memset(&inno_drm_mode, 0, inno_display_size);
		memcpy(&inno_drm_mode, buf + i, inno_display_size);
		if (inno_drm_mode.clock) {
			memset(ptr, 0, sizeof(struct drm_display_mode));
			ptr->clock       = be32_to_cpu(inno_drm_mode.clock);
			ptr->hdisplay    = be16_to_cpu(inno_drm_mode.hdisplay);
			ptr->hsync_start = be16_to_cpu(inno_drm_mode.hsync_start);
			ptr->hsync_end   = be16_to_cpu(inno_drm_mode.hsync_end);
			ptr->htotal      = be16_to_cpu(inno_drm_mode.htotal);
			ptr->hskew       = be16_to_cpu(inno_drm_mode.hskew);
			ptr->vdisplay    = be16_to_cpu(inno_drm_mode.vdisplay);
			ptr->vsync_start = be16_to_cpu(inno_drm_mode.vsync_start);
			ptr->vsync_end   = be16_to_cpu(inno_drm_mode.vsync_end);
			ptr->vtotal      = be16_to_cpu(inno_drm_mode.vtotal);
			ptr->vscan       = be16_to_cpu(inno_drm_mode.vscan);
			ptr->flags       = be32_to_cpu(inno_drm_mode.flags);
			snprintf(ptr->name, sizeof(ptr->name), "%dx%d", ptr->hdisplay, ptr->vdisplay);

			mode = drm_mode_duplicate(connector->dev, ptr);
			if (mode) {
				drm_mode_probed_add(connector, mode);
				++num_modes;
			}
		}
	}

	kfree(ptr);

	return num_modes;
}

static struct connector_map_prop s_connector_map[CUSTOM_OUTPUT_MODE_MAX][DIMENSION_MAX][DPU_CONNECT_NUM] = {
	[CUSTOM_OUTPUT_MODE_1] = {
		[DIMENSION_G1_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G1P_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_2] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G1P_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_3] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G1P_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_4] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),

		},
		[DIMENSION_G1P_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_5] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 0),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_6] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 0),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_7] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 2),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_8] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 0),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_9] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 0),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_10] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 2),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_11] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 0),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_12] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 1),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI0),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 4096, 4096, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},

	[CUSTOM_OUTPUT_MODE_13] = {
		[DIMENSION_G0_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_VGA),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_LVDS, 2048, 2048, 2),
		},
		[DIMENSION_G0M_SOC] = {
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI0, 2048, 2048, 3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_HDMI1, 2048, 2048, 1),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_HDMI3),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_DP0, 2048, 2048, 2),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_DP1),
			CONNECTOR_MAP_ITEM(CONNECTOR_M_VGA, 1920, 1080, 0),
			CONNECTOR_MAP_EMPTY_ITEM(CONNECTOR_M_LVDS),
		},
	},
};

static int g0_soc_connector_map_get(int sk_select, int sk_support, int dpu_match)
{
	int ret = 0;
	unsigned long bit_map_sk = 0;

	if (dpu_match > 0 || s_dpu_match > 0) {
		if (s_dpu_match > 0) {
			ret = s_dpu_match;
		} else {
			ret = dpu_match;
			s_dpu_match = ret;
		}
	 } else {
		if (sk_select >= 0) {
			bit_map_sk = sk_select;
		} else {
			bit_map_sk = sk_support;
		}

		if (bit_map_sk & BIT(CONNECTOR_M_HDMI1)) {
			ret = CUSTOM_OUTPUT_MODE_3;
		} else if (bit_map_sk & BIT(CONNECTOR_M_HDMI0)) {
			ret = CUSTOM_OUTPUT_MODE_1;
		} else if (bit_map_sk & BIT(CONNECTOR_M_DP0)) {
			ret = CUSTOM_OUTPUT_MODE_2;
		} else {
			ret = CUSTOM_OUTPUT_MODE_3;
		}
	}

	return ret;
}

static int g0m_soc_connector_map_get(int sk_select, int sk_support, int dpu_match)
{
	int ret = 0;

	if (s_dpu_match || dpu_match <= 0) {
		if (s_dpu_match) {
			ret = s_dpu_match;
		} else {
			ret = CUSTOM_OUTPUT_MODE_8;
		}
	} else {
		ret = dpu_match;
	}

	s_dpu_match = ret;

	return ret;
}

static void innodpu_logo_extime_blacklist(struct drm_device *drm_dev)
{
	struct hw_board_info board1 = HW_BOARD_INFO_ITEM("ZY", "NH32_1");
	struct hw_board_info board2 = HW_BOARD_INFO_ITEM("ZY", "NH32_2");

	if (!drm_dev || s_logo_extime <= 0)
		return;

	if (innodpu_is_odm_pcb_match(drm_dev->dev, &board1) ||
		innodpu_is_odm_pcb_match(drm_dev->dev, &board2)) {
		s_logo_extime = 0;
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "No logo delay\n");
	}
}

static int custom_mode = 0;
int innodpu_custom_init(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	chip_type_e plat;
	int retcode = 0;
	int dpu_match = 0;
	int dimension = 0;
	int sk_select = 0;
	int sk_support = 0;
	int ret = 0;
	plat_data_t *drm_plat_data = NULL;

	ret = innogpu_drm_get_platdata(drm_dev, &drm_plat_data);
	if (ret) {
		return ret;
	};

	plat = hal_get_chiptype(drm_dev->dev);
	sk_select = hal_select_4k(drm_dev->dev);
	sk_support = hal_is_support_4k(drm_dev->dev);
	dpu_match = hal_get_dpu_match(drm_dev->dev);

	innodpu_logo_extime_blacklist(drm_dev);

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "select 4k: %d\n", sk_select);
	innodpu_info(drm_dev->dev, DPU_UT_INIT, "support 4k:%d\n", sk_support);
	innodpu_info(drm_dev->dev, DPU_UT_INIT, "dpu_match:%d\n", dpu_match);

	switch (plat) {
	case CHIP_G1_SOC:
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "DPU start init g1 soc.\n");
		s_dpu_nums = 0xc;
		dimension = DIMENSION_G1_SOC;
		custom_mode = CUSTOM_OUTPUT_MODE_1;
		dev_priv->role_id = drm_plat_data->dev_idx * 2;
	break;
	case CHIP_G1P_SOC:
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "DPU start init g1p soc.\n");
		s_dpu_nums = 0x3;
		dimension = DIMENSION_G1P_SOC;
		custom_mode = CUSTOM_OUTPUT_MODE_1;
		dev_priv->role_id = drm_plat_data->dev_idx * 2;
	break;
	case CHIP_G0_SOC:
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "DPU start init g0 soc.\n");
		s_dpu_nums = 0x07;
		dev_priv->role_id = 0;
		dimension = DIMENSION_G0_SOC;
		custom_mode = g0_soc_connector_map_get(sk_select, sk_support, dpu_match);
	break;
	case CHIP_G0M_SOC:
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "DPU start init g0m soc.\n");
		s_dpu_nums = 0x0f;
		dev_priv->role_id = 0;
		dimension = DIMENSION_G0M_SOC;
		custom_mode = g0m_soc_connector_map_get(sk_select, sk_support, dpu_match);
	break;
	default:
		innodpu_err(drm_dev->dev, "DPU does not currently support %d platform.\n", plat);
		retcode = -EINVAL;
	break;
	}

	if (custom_mode >= CUSTOM_OUTPUT_MODE_MAX ||
		custom_mode <= CUSTOM_OUTPUT_MODE_0) {
		innodpu_err(drm_dev->dev, "custom mode parse error!,Use display mode1\n");
		s_dpu_match = custom_mode = CUSTOM_OUTPUT_MODE_1;
	}

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "custom_mode:%d\n"
			"hdmi-0 max_width:%d max_height:%d possible_crtc:%d\n"
			"hdmi-1 max_width:%d max_height:%d possible_crtc:%d\n"
			"hdmi-2 max_width:%d max_height:%d possible_crtc:%d\n"
			"hdmi-3 max_width:%d max_height:%d possible_crtc:%d\n"
			"dp - 0 max_width:%d max_height:%d possible_crtc:%d\n"
			"dp - 1 max_width:%d max_height:%d possible_crtc:%d\n"
			"   vga max_width:%d max_height:%d possible_crtc:%d\n"
			"  lvds max_width:%d max_height:%d possible_crtc:%d\n",
			custom_mode,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI0].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI0].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI0].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI1].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI1].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI1].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI2].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI2].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI2].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI3].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI3].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_HDMI3].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP0].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP0].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP0].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP1].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP1].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_DP1].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_VGA].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_VGA].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_VGA].possible_crtc,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_LVDS].max_width,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_LVDS].max_height,
			s_connector_map[custom_mode][dimension][CONNECTOR_M_LVDS].possible_crtc);

	return retcode;

}

struct connector_map_prop *innodpu_find_connector_map_module(struct device *dev, enum connector_module module)
{
	chip_type_e plat = hal_get_chiptype(dev->parent);
	struct connector_map_prop *connector_map_module = NULL;

	switch (plat) {
	case CHIP_G1_SOC:
		connector_map_module = &s_connector_map[custom_mode][DIMENSION_G1_SOC][module];
	break;
	case CHIP_G1P_SOC:
		connector_map_module = &s_connector_map[custom_mode][DIMENSION_G1P_SOC][module];
	break;
	case CHIP_G0_SOC:
		connector_map_module = &s_connector_map[custom_mode][DIMENSION_G0_SOC][module];
	break;
	case CHIP_G0M_SOC:
		connector_map_module = &s_connector_map[custom_mode][DIMENSION_G0M_SOC][module];
	break;
	default:
		connector_map_module = NULL;
		innodpu_err(dev, "DPU does not currently support %d platform.\n", plat);
	break;
	};

	if (connector_map_module != NULL) {
		innodpu_info(dev, DPU_UT_INIT, "connector_mode:%d\n"
				"max_width:%d max_height:%d possible_crtc:%d\n",
				module, connector_map_module->max_width,
				connector_map_module->max_height,
				connector_map_module->possible_crtc);
	}

	return connector_map_module;
}

int innodpu_get_display_mode(void)
{
	return custom_mode;
}

int innodpu_get_mfc_name(struct device *device, char *name, unsigned char *edid)
{
	int offset = 'A'-1;

	if (name == NULL) {
		return -1;
	}

	if (edid == NULL) {
		inno_memset(name, '*', 4);
		name[3] = '\0';
		innodpu_warn(device, "edid pointer null, use default mfc name[%s]\n", name);
	} else {
		name[0] = (edid[0x08] >> 2) + offset;
		name[1] = (((edid[0x08] << 3) | (edid[0x09] >> 5)) & 0x1f)+ offset;
		name[2] = (edid[0x09] & 0x1f) + offset;
		name[3] = '\0';

		if (inno_strlen(name) == 0) {
			inno_memset(name, '*', 4);
			name[3] = '\0';
			innodpu_warn(device, "edid pointer null, use default mfc name[%s]\n", name);
		} else {
			innodpu_info(device, DPU_UT_INIT, "get mfc name[%s]\n", name);
		}
	}

	return 0;
}

int innodpu_get_odm_info(struct device *device, char *odm_info, inno_dev *pcie_dev)
{
	int ret = 0;

	ret = hal_get_odm_vendor(pcie_dev, odm_info);

	if (ret || (inno_strlen(odm_info) == 0)) {
		inno_memset(odm_info, '*', ODM_INFO_LEN);
		odm_info[ODM_INFO_LEN - 1] = '\0';
		innodpu_warn(device, "get odm info err, use default odm info[%s]\n", odm_info);
		return ret;
	}

	innodpu_info(device, DPU_UT_INIT, "get odm info[%s]\n", odm_info);
	return ret;
}

bool innodpu_mode_need_bypass(const char *monitor, char *odm_info,
	chip_type_e plat, struct innodpu_mode_bypass_list *mode_list,
	int list_len, int hdisplay, int vdisplay, int vrefresh)
{
	int i = 0;
	uint64_t monitor_name_len = 0 , odm_info_len = 0;

	monitor_name_len = inno_strlen(monitor);
	odm_info_len = inno_strlen(odm_info);

	for (i = 0; i < list_len; i++) {

		/* check plat */
		if ((mode_list[i].plat != 0) && (mode_list[i].plat != plat)) {
			continue;
		}

		/* check odm info */
		if ((inno_strncmp(odm_info, mode_list[i].odm_info, odm_info_len) != 0) &&
			(inno_strncmp("*", mode_list[i].odm_info, 1) != 0)) {
			continue;
		}

		/* check monitor info */
		if ((inno_strncmp(monitor, mode_list[i].monitor, monitor_name_len) != 0) &&
			(inno_strncmp("*", mode_list[i].monitor, 1) != 0)) {
			continue;
		}

		/* check mode info */
		if ((hdisplay == mode_list[i].hdisplay) &&
			(vdisplay == mode_list[i].vdisplay) &&
			(vrefresh == mode_list[i].vrefresh)) {
				return true;
		}
	}

	return false;
}

bool innodpu_is_odm_pcb_match(inno_dev *pcie_dev, struct hw_board_info *board_info)
{
	int ret;
	char odm_vendor[HW_ODM_VENDOR_LEN] = {0};
	char pcb_version[HW_PCB_VERSION_LEN] = {0};

	if (!pcie_dev || !board_info) {
		goto out;
	}

	if (!board_info->odm_vendor || !board_info->pcb_version) {
		goto out;
	}

	ret = hal_get_odm_vendor(pcie_dev, odm_vendor);
	if (-1 == ret)
		goto out;

	ret = hal_get_pcb_version(pcie_dev, pcb_version);
	if (-1 == ret)
		goto out;

	if (!strcmp(odm_vendor, board_info->odm_vendor) &&
		!strcmp(pcb_version, board_info->pcb_version)) {
		return true;
	}

out:
	return false;
}

struct drm_display_mode *
innodpu_get_native_mode(struct drm_connector *connector,
		bool (*filter)(struct drm_connector *pconnector, struct drm_display_mode *pmode, bool priv_filter))
{
	struct drm_display_mode *mode = NULL;
	struct drm_display_mode *native_mode = NULL;
	struct drm_display_mode *minimum_mode = NULL;
	struct drm_display_mode *preferred_mode = NULL;

	/*
	 * choose Wmax as native mode, find minimum mode.
	 */
	list_for_each_entry(mode, &connector->probed_modes, head) {
		if (mode && (mode->type & DRM_MODE_TYPE_PREFERRED)) {
			preferred_mode = mode;
		} else {
			preferred_mode = NULL;
		}

		if (!minimum_mode || (minimum_mode->hdisplay > mode->hdisplay)) {
			if (!filter(connector, mode, false))
				minimum_mode = mode;
		}

		if (filter(connector, mode, true))
			continue;

		if (preferred_mode && preferred_mode->hdisplay < INNODPU_COMBINE_WIDTH &&
			preferred_mode->vdisplay < INNODPU_COMBINE_WIDTH) {
			native_mode = preferred_mode;
			break;
		}

		if (!native_mode || (native_mode->hdisplay < mode->hdisplay)) {
			native_mode = mode;
			continue;
		}

		if (native_mode->hdisplay > mode->hdisplay)
			continue;

		if ((native_mode->hdisplay == mode->hdisplay) &&
			(native_mode->vdisplay < mode->vdisplay))
			native_mode = mode;
	}

	/*
	 * choose Wmin as native mode if min_mode > combine
	 */
	if ((!native_mode) && minimum_mode) {
		bool mode_priv_support = filter(connector, minimum_mode, true); // true indicate  need combine
		if (mode_priv_support) {
			native_mode = minimum_mode;
		}
	}

	return native_mode;
}


static struct drm_display_mode *
innodpu_create_common_mode(struct drm_connector *connector,
		struct drm_display_mode *native_mode, char *name, int hdisplay, int vdisplay)
{
	struct drm_device *drm_dev = connector->dev;
	struct drm_display_mode *mode = NULL;

	mode = drm_mode_duplicate(drm_dev, native_mode);
	if (mode == NULL)
		return NULL;

	mode->hdisplay = hdisplay;
	mode->vdisplay = vdisplay;
	mode->type &= ~DRM_MODE_TYPE_PREFERRED;
	strscpy(mode->name, name, DRM_DISPLAY_MODE_LEN);

	return mode;
}

int innodpu_connector_add_common_modes(struct drm_connector *connector,
		struct drm_display_mode *native_mode, bool support_force)
{
	struct drm_display_mode *mode = NULL;
	int i;
	int n;
	int num_modes = 0;
	struct mode_size {
		char name[DRM_DISPLAY_MODE_LEN];
		int w;
		int h;
		bool force;
	} common_modes[] = {
#if 0 // RK
		{  "640x480",  640,  480, false},
		{  "720x480",  720,  480, false},
		{  "800x600",  800,  600, false},
		{ "1024x768", 1024,  768, true},
		{ "1152x864", 1152,  864, true},
		{ "1280x600", 1280,  720, true},
		{ "1280x720", 1280,  720, true},
		{ "1280x768", 1280,  768, true},
		{ "1280x800", 1280,  800, true},
		{ "1280x960", 1280,  960, true},
		{"1280x1024", 1280, 1024, true},
		{ "1360x768", 1360,  768, true},
		{ "1366x768", 1366,  768, true},
		{ "1440x900", 1440,  900, false},
		{"1440x1050", 1440, 1050, false},
		{"1680x1050", 1680, 1050, false},
		{" 1600x900", 1600,  900, false},
		{"1600x1200", 1600, 1200, false},
		{"1680x1050", 1600, 1050, false},
		{"1920x1080", 1920, 1080, false},
		{"1920x1200", 1920, 1200, false}
#else
		{  "640x480",  640,  480, false},
		{  "720x480",  720,  480, false},
		{  "800x600",  800,  600, false},
		{ "1024x768", 1024,  768, true},
		{ "1280x720", 1280,  720, true},
		{ "1280x800", 1280,  800, true},
		{ "1280x960", 1280,  960, true},
		{"1280x1024", 1280, 1024, true},
		/* { "1366x768", 1366,  768, false}, */
		{ "1440x900", 1440,  900, false},
		{"1680x1050", 1680, 1050, false},
		{"1600x900",  1600,  900, false},
		{"1600x1200", 1600, 1200, false},
		{"1920x1080", 1920, 1080, false},
		{"1920x1200", 1920, 1200, false},
		{"2560x1440", 2560, 1440, false},
		{"2560x1600", 2560, 1600, false},
		{"2560x1920", 2560, 1920, false},
		{"2560x2048", 2560, 2048, false}
#endif
	};

	n = ARRAY_SIZE(common_modes);

	for (i = 0; i < n; i++) {
		struct drm_display_mode *curmode = NULL;
		bool mode_existed = false;

		if (common_modes[i].w > native_mode->hdisplay ||
			common_modes[i].h > native_mode->vdisplay ||
			(common_modes[i].w == native_mode->hdisplay &&
			common_modes[i].h == native_mode->vdisplay))
			continue;

		list_for_each_entry(curmode, &connector->probed_modes, head) {
			if (common_modes[i].w == curmode->hdisplay &&
				common_modes[i].h == curmode->vdisplay) {
				if (support_force && (common_modes[i].force == true)) {
					drm_mode_copy(curmode, native_mode);
					curmode->hdisplay = common_modes[i].w;
					curmode->vdisplay = common_modes[i].h;
					curmode->type &= ~DRM_MODE_TYPE_PREFERRED;
					strscpy(curmode->name, common_modes[i].name, DRM_DISPLAY_MODE_LEN);
				}
				mode_existed = true;
				break;
			}
		}

		if (mode_existed)
			continue;

		mode = innodpu_create_common_mode(connector, native_mode,
				common_modes[i].name, common_modes[i].w,
				common_modes[i].h);
		if (!mode)
			continue;

		drm_mode_probed_add(connector, mode);
		num_modes++;
	}
	return num_modes;
}

static struct connector_output_mode dp_output_mode[] = {
	{DRM_MODE_CONNECTOR_DisplayPort, "DP", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA, "CH7517", CONNECTOR_FLAGS_EN_PLUGIN, 0},
	{DRM_MODE_CONNECTOR_VGA, "CM3166", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA, "CS5212", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA, "RT2166", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA, "LT8711V", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_DisplayPort, "DP", CONNECTOR_FLAGS_NONE, 0},/*reserver id:6, as DP now*/
	{DRM_MODE_CONNECTOR_DisplayPort, "DP", CONNECTOR_FLAGS_NONE, 0},/*reserver id:7, as DP now*/
	{DRM_MODE_CONNECTOR_DisplayPort, "DP", CONNECTOR_FLAGS_NONE, 0},/*reserver id:8, as DP now*/
	{DRM_MODE_CONNECTOR_eDP, "eDP", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_LVDS,"LT7911D", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_LVDS,"CS5211", CONNECTOR_FLAGS_NONE, 0},
};

static struct connector_output_mode hdmi_output_mode[] = {
	{DRM_MODE_CONNECTOR_HDMIA, "HDMI", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_DVII,  "DVI", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA,   "CS5213", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_DisplayPort, "CS5801", \
		CONNECTOR_FLAGS_SKIP_DDC | CONNECTOR_FLAGS_LONGER_HPD, 1600},
	{DRM_MODE_CONNECTOR_LVDS,  "LT6911C", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA,  "CS5210", CONNECTOR_FLAGS_NONE, 0},
	{DRM_MODE_CONNECTOR_VGA,  "LT8511A", CONNECTOR_FLAGS_NONE, 0},
};

static int chip_connector_id[DIMENSION_MAX][DPU_CONNECT_NUM] = {
	[DIMENSION_G1_SOC] = {
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI0, 0),
		CONNECTOR_ID_ITEM(CONNECTOR_M_DP0,   1),
	},
	[DIMENSION_G0_SOC] = {
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI0, 0),
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI1, 1),
		CONNECTOR_ID_ITEM(CONNECTOR_M_DP0,   2),
		CONNECTOR_ID_ITEM(CONNECTOR_M_LVDS,  3),
	},
	[DIMENSION_G1P_SOC] = {
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI0, 0),
		CONNECTOR_ID_ITEM(CONNECTOR_M_DP0,   1),
		CONNECTOR_ID_ITEM(CONNECTOR_M_VGA,   2),
	},
	[DIMENSION_G0M_SOC] = {
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI0, 0),
		CONNECTOR_ID_ITEM(CONNECTOR_M_HDMI1, 1),
		CONNECTOR_ID_ITEM(CONNECTOR_M_DP0,   2),
		CONNECTOR_ID_ITEM(CONNECTOR_M_VGA,   3),
	},
};

bool is_output_type_hdmi(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_HDMIA) {
		return true;
	}

	/*others*/
	return false;
}

bool is_output_type_dvi(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_DVII) {
		return true;
	}

	/*others*/
	return false;
}

bool is_output_type_vga(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_VGA) {
		return true;
	}

	/*others*/
	return false;
}

bool is_output_type_dp(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_DisplayPort) {
		return true;
	}

	/*others*/
	return false;
}

bool is_output_type_edp(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_eDP) {
		return true;
	}

	/*others*/
	return false;
}

bool is_output_type_lvds(struct connector_output_mode *om)
{
	if (!om)
		return false;

	if (om->mode_connector_item == DRM_MODE_CONNECTOR_LVDS) {
		return true;
	}

	/*others*/
	return false;
}

bool is_flags_skip_ddc_detect(struct connector_output_mode *om)
{
	bool skip = false;

	if (!om)
		return skip;

	if (om->flags & CONNECTOR_FLAGS_SKIP_DDC) {
		skip = true;
	}

	return skip;
}

bool is_flags_wait_longer_hpd(struct connector_output_mode *om)
{
	bool wait = false;

	if (!om)
		return wait;

	if (om->flags & CONNECTOR_FLAGS_LONGER_HPD) {
		wait = true;
	}

	return wait;
}

bool is_flags_data_en_plugin(struct connector_output_mode *om)
{
	bool en = false;

	if (!om)
		return en;

	if (om->flags & CONNECTOR_FLAGS_EN_PLUGIN) {
		en = true;
	}

	return en;
}


static int innodpu_get_connector_id(inno_dev *dev, unsigned int reg_module)
{
	chip_type_e plat;
	inno_dev *pdev = NULL;
	int dimension = 0, index = 0;

	if (!dev)
		return -EINVAL;

	pdev = inno_dev_get_parent(dev);
	if (!pdev)
		return -EFAULT;

	plat = hal_get_chiptype(pdev);
	switch (plat) {
	case CHIP_G1_SOC:
		dimension = DIMENSION_G1_SOC;
	break;
	case CHIP_G0_SOC:
		dimension = DIMENSION_G0_SOC;
	break;
	case CHIP_G1P_SOC:
		dimension = DIMENSION_G1P_SOC;
	break;
	case CHIP_G0M_SOC:
		dimension = DIMENSION_G0M_SOC;
	break;
	default:
		dimension = DIMENSION_MAX;
		innodpu_err(dev, "unknown chip type""\n");
	break;
	};

	switch (reg_module) {
	case REG_M_HDMI:
		index = CONNECTOR_M_HDMI0;
	break;
	case REG_M_HDMI1:
		index = CONNECTOR_M_HDMI1;
	break;
	case REG_M_DP:
		index = CONNECTOR_M_DP0;
	break;
	case REG_M_VGA:
		index = CONNECTOR_M_VGA;
	break;
	case REG_M_LVDS:
		index = CONNECTOR_M_LVDS;
	break;
	default:
		index = DPU_CONNECT_NUM;
		innodpu_err(dev, "unknown reg_module\n");
	break;
	};

	if (dimension >= DIMENSION_MAX ||
		index >= DPU_CONNECT_NUM) {
		innodpu_err(dev, "unknown connector id\n");
		return -EFAULT;
	}

	return chip_connector_id[dimension][index];
}

extern int s_dp_vga;
extern int s_hdmi_dvi;
static int innodpu_get_output_index(inno_dev *dev, unsigned int reg_module)
{
	inno_dev *pdev = NULL;
	unsigned char val = 0;
	int connector_id = 0, ret = 0, version = 0;

	if (!dev)
		return -EINVAL;

	pdev  = inno_dev_get_parent(dev);
	if (!pdev)
		return -EFAULT;

	version = hal_hwinfo_version(pdev);
	if (version < 0) {
		innodpu_err(dev, "hwinfo version failed\n");
		return -EFAULT;
	}

	if (version >= 0x7) {
		connector_id = innodpu_get_connector_id(dev, reg_module);
		if (connector_id < 0) {
			innodpu_err(dev, "connector_id failed!!!\n");
			return -EINVAL;
		}

		ret = hal_get_output_mode(pdev, connector_id, &val);
		if (ret < 0) {
			innodpu_err(dev, "hal_get_output_mode failed, Using 0\n");
			val = 0;
		}
	} else {
		if (reg_module == REG_M_HDMI) {
			if (s_hdmi_dvi) {
				ret = (s_hdmi_dvi & BIT(0)) ? 1 : 0;
			} else {
				ret = hal_getflag_hdmi2dvi(pdev, 0);
			}
		} else if (reg_module == REG_M_HDMI1) {
			if (s_hdmi_dvi) {
				ret = (s_hdmi_dvi & BIT(1)) ? 1 : 0;
			} else {
				ret = hal_getflag_hdmi2dvi(pdev, 1);
			}
		} else if (reg_module == REG_M_DP) {
			if (s_dp_vga == 1)
				ret = 1;
			else
				ret = hal_getflag_dp2vga(pdev, 0);
		}

		if (ret > 0)
			val = 1;
		else
			val = 0;
	}

	innodpu_info(dev, DPU_UT_INIT, "connector_id:%d output_id:%d\n", connector_id, val);

	return val;
}

struct connector_output_mode *innodpu_get_connector_output_mode(inno_dev *dev,
		unsigned int reg_module)
{
	int index = 0;
	struct connector_output_mode *output_mode = NULL;

	if (!dev)
		return NULL;

	index = innodpu_get_output_index(dev, reg_module);
	if (index < 0) {
		innodpu_err(dev, "get_output_mode failed!!!, Using the default output:0\n");
		index = 0;
	}

	if (reg_module == REG_M_HDMI ||
		reg_module == REG_M_HDMI1) {
		if (index < ARRAY_SIZE(hdmi_output_mode))
			output_mode = &hdmi_output_mode[index];
	} else if (reg_module == REG_M_DP) {
		if (index < ARRAY_SIZE(dp_output_mode))
			output_mode = &dp_output_mode[index];
	} else {
		output_mode = NULL;
	}

	return output_mode;
}

int innodpu_get_connector_backlight_mode(inno_dev *dev, unsigned int reg_module)
{
	inno_dev *pdev = NULL;
	unsigned char mode = 0;
	int connector_id = 0, ret = 0, version = 0;

	if (!dev)
		return -EINVAL;

	pdev  = inno_dev_get_parent(dev);
	if (!pdev)
		return -EFAULT;

	version = hal_hwinfo_version(pdev);
	if (version < 0) {
		innodpu_err(dev, "hwinfo version failed, Use default DDCCI\n");
		return CONNECTOR_BACKLIGHT_DDCCI;
	}

	if (version >= 0x7) {
		connector_id = innodpu_get_connector_id(dev, reg_module);
		if (connector_id < 0) {
			innodpu_err(dev, "connector_id failed, Use default DDCCI\n");
			return CONNECTOR_BACKLIGHT_DDCCI;
		}

		ret = hal_get_backlight_mode(pdev, connector_id, &mode);
		if (ret < 0) {
			innodpu_err(dev, "hal_get_backlight_mode failed, Use default DDCCI\n");
			mode = CONNECTOR_BACKLIGHT_DDCCI;
		}
	} else {
		mode = CONNECTOR_BACKLIGHT_DDCCI;
	}

	innodpu_info(dev, DPU_UT_INIT, "backlight_mode:%d\n", mode);

	return mode;
}

static int range_pixel_clock_khz(struct edid *edid, u8 *t)
{
	/* unspecified */
	if (t[9] == 0 || t[9] == 255)
		return 0;

	/* 1.4 with CVT support gives us real precision, yay */
	if (edid->revision >= 4 && t[10] == 0x04)
		return (t[9] * 10000) - ((t[12] >> 2) * 250);

	/* 1.3 is pathetic, so fuzz up a bit */
	return t[9] * 10000 + 5001;
}

#define EDID_DETAILED_TIMINGS 4
#define EDID_DETAIL_MONITOR_RANGE 0xfd
int innodpu_detailed_block_monitor_range_replace(u8 *raw_edid)
{
	u8 csum = 0;
	int i, clock;
	struct detailed_timing *timing = NULL;
	struct detailed_non_pixel *data = NULL;
	struct edid *edid = (struct edid *)raw_edid;
	struct detailed_data_monitor_range *range = NULL;

	if (edid == NULL)
		return -1;

	for (i = 0; i < EDID_DETAILED_TIMINGS; i++) {
		timing = &(edid->detailed_timings[i]);
		data = &timing->data.other_data;

		if (data->type != EDID_DETAIL_MONITOR_RANGE)
			continue;

		clock = range_pixel_clock_khz(edid, (u8*)timing);
		range = &data->data.range;
		if (clock > 533000 &&
			clock < 594000) {

			/* Replace Max Supported pixel clock rate in MHz/10 */
			if (edid->revision >= 4 && range->flags == 0x04) {
				range->pixel_clock_mhz = 60;
				range->formula.cvt.data1 = 0;
			} else {
				range->pixel_clock_mhz = 60;
			}

			inno_printk("Max Supported pixel clock rate replace:old-%d khz, new-%d khz\n",
					clock, range_pixel_clock_khz(edid, (u8 *)timing));

			/* Adjust edid checksum to 0 */
			for (i = 0; i < EDID_LENGTH; i++)
				csum += raw_edid[i];

			if (csum) {
				if (csum <= edid->checksum)
					edid->checksum -= csum;
				else
					edid->checksum += 256 - csum;
			}
		}

		return 0;
	}

	return -1;
}

int innodpu_conn_get_monitor_max_clk(u8 *raw_edid)
{
	int i, clock;
	struct detailed_timing *timing = NULL;
	struct detailed_non_pixel *data = NULL;
	struct edid *edid = (struct edid *)raw_edid;

	if (edid == NULL)
		return -1;

	for (i = 0; i < EDID_DETAILED_TIMINGS; i++) {
		timing = &(edid->detailed_timings[i]);
		data = &timing->data.other_data;

		if (data->type != EDID_DETAIL_MONITOR_RANGE)
			continue;

		clock = range_pixel_clock_khz(edid, (u8*)timing);
		return clock;
	}

	return -1;
}
#undef EDID_DETAILED_TIMINGS
#undef EDID_DETAIL_MONITOR_RANGE


#ifdef CONFIG_DRM_INNO_AUDIO
void innodpu_connector_clear_eld(struct drm_connector *connector)
{
	memset(connector->eld, 0, sizeof(connector->eld));

	connector->latency_present[0] = false;
	connector->latency_present[1] = false;
	connector->video_latency[0] = 0;
	connector->audio_latency[0] = 0;
	connector->video_latency[1] = 0;
	connector->audio_latency[1] = 0;
}
#endif
