/*************************************************************************/ /*!
@File			innodpu_dpu.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include "inno_drm_version.h"
#include <linux/component.h>
#include <linux/platform_device.h>
#include "innodpu_dpu.h"
#include "innodpu_common.h"
#include "innogpu.h"
#include "innogpu_drm.h"

bool s_pdp0_debug = false;
module_param(s_pdp0_debug, bool, 0600);
MODULE_PARM_DESC(s_pdp0_debug, "pdp0 debug(default: false)");

bool s_drm_debug = false;
module_param(s_drm_debug, bool, 0600);
MODULE_PARM_DESC(s_drm_debug, "drm debug(default: false)");

bool s_dp_debug = false;
module_param(s_dp_debug, bool, 0600);
MODULE_PARM_DESC(s_dp_debug, "dp debug(default: false)");

bool s_vga_debug = false;
module_param(s_vga_debug, bool, 0600);
MODULE_PARM_DESC(s_vga_debug, "dp debug(default: false)");

unsigned int s_wb_size = 0x200000;
module_param(s_wb_size, uint, 0444);
MODULE_PARM_DESC(s_wb_size, "wb size (default: 0x200000)");

int s_dp_vga = -1;
module_param(s_dp_vga, int, 0444);
MODULE_PARM_DESC(s_dp_vga, "dp to vga enable(-1 not support, 0-dp, 1-vga)");

unsigned int s_hdmi_dvi = 0x00;
module_param(s_hdmi_dvi, uint, 0600);
MODULE_PARM_DESC(s_hdmi_dvi, "hdmi to dvi enable(-1 not support, 0x01-hdmi0 to dvi, 0x02-hdmi1 to dvi)");

int s_hdmi_freq = 0;
module_param(s_hdmi_freq, int, 0644);
MODULE_PARM_DESC(s_hdmi_freq, "Only debug, to set hdmi freq target.");

int s_hdmi_ddcfreq = 0;
module_param(s_hdmi_ddcfreq, int, 0644);
MODULE_PARM_DESC(s_hdmi_ddcfreq, "set hdmi ddcfreq(kHz)");

unsigned int s_dpu_debug = 0x00;
MODULE_PARM_DESC(s_dpu_debug, "Enable debug output, where each bit enables a debug category.\n"
		"\t\tBit 0 (0x01)  will enable INIT messages (drm init code)\n"
		"\t\tBit 1 (0x02)  will enable CORE messages (drm core code)\n"
		"\t\tBit 2 (0x04)  will enable MEM messages (drm MEM code)\n"
		"\t\tBit 3 (0x08)  will enable DPU messages (drm controller code)\n"
		"\t\tBit 4 (0x10)  will enable HDMI messages (hdmi code)\n"
		"\t\tBit 5 (0x20)  will enable DP messages (dp code)\n"
		"\t\tBit 6 (0x40)  will enable LVDS messages (lvds code)\n"
		"\t\tBit 7 (0x80) will enable VGA messages (vga code)\n"
		"\t\tBit 8 (0x100) will enable cursor messages (cursor code\n"
		"\t\tBit 9 (0x200) will enable gamma messages (gamma code)");

module_param_named(s_dpu_debug, s_dpu_debug, uint, 0600);

unsigned int s_dpu_level = 0x48;
MODULE_PARM_DESC(s_dpu_level, "Enable debug output level, default: ERROR/INFO.\n"
		"\t\tBit0  (0x01)  will enable EMERG messages.\n"
		"\t\tBit1  (0x02)  will enable ALERT messages.\n"
		"\t\tBit2  (0x04)  will enable CRIT messages.\n"
		"\t\tBit3  (0x08)  will enable ERROR messages.\n"
		"\t\tBit4  (0x10)  will enable WARN messages.\n"
		"\t\tBit5  (0x20)  will enable NOTICE messages.\n"
		"\t\tBit6  (0x40)  will enable INFO messages.\n"
		"\t\tBit7  (0x80)  will enable DEBUG messages.\n");
module_param_named(s_dpu_level, s_dpu_level, uint, 0600);

extern int innodpu_pdp0_bind(struct device *dev, struct device *master, void *data, bool null_display);
extern void innodpu_pdp0_unbind(struct device *dev, struct device *master, void *data, bool null_display);
extern int innodpu_pdp0_suspend(struct device *dev, int dpu_id);
extern int innodpu_pdp0_resume(struct device *dev, int dpu_id);
extern int innodpu_pdp0_shutdown(struct device *dev, int dpu_id);



static int innodpu_bind(struct device *dev, struct device *master, void *data)
{
	int dpu_id = 0;
	plat_data_t *pdata;
	struct drm_device * drm_dev = data;
	struct innodpu_drm_private *dev_priv;
	unsigned dpu_nums = hal_get_dev_nums(drm_dev->dev, DEV_DPU);
	unsigned db9000_nums = hal_get_dev_nums(drm_dev->dev, DEV_DB9000);

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "Invaild dev_priv\n");
		return -EINVAL;
	}

	pdata = dev->platform_data;
	dpu_id = pdata->dev_idx;
	if (dpu_id > dpu_nums) {
		innodpu_err(drm_dev->dev, "Invaild dpu_id:%d\n", dpu_id);
		return -EINVAL;
	}

	if (!(s_dpu_nums & BIT(dpu_id)) && !dev_priv->drm_nulldisplay)
		return 0;

	if (dpu_id < db9000_nums && !dev_priv->drm_nulldisplay) {
		innodpu_warn(drm_dev->dev, "Not support db9000\n");
		return 0;
	} else {
		return innodpu_pdp0_bind(dev, master, data, dev_priv->drm_nulldisplay);
	}
}

static void innodpu_unbind(struct device *dev, struct device *master, void *data)
{
	int dpu_id = 0;
	plat_data_t *pdata;
	unsigned dpu_nums = hal_get_dev_nums(dev->parent, DEV_DPU);
	unsigned db9000_nums = hal_get_dev_nums(dev->parent, DEV_DB9000);
	struct drm_device * drm_dev = data;
	struct innodpu_drm_private *dev_priv;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "Invaild dev_priv\n");
		return;
	}

	pdata = dev->platform_data;
	dpu_id = pdata->dev_idx;
	if (dpu_id > dpu_nums) {
		innodpu_err(drm_dev->dev, "Invaild dpu_id:%d\n", dpu_id);
		return;
	}

	if (!(s_dpu_nums & BIT(dpu_id)) && !dev_priv->drm_nulldisplay)
		return;

	if (dpu_id < db9000_nums && !dev_priv->drm_nulldisplay) {
		innodpu_warn(drm_dev->dev, "Not support db9000\n");
	} else {
		innodpu_pdp0_unbind(dev, master, data, dev_priv->drm_nulldisplay);
	}
}

static const struct component_ops innodpu_ops = {
	.bind = innodpu_bind,
	.unbind = innodpu_unbind,
};

static int innodpu_platform_probe(struct platform_device *pdev)
{
	return component_add(&pdev->dev, &innodpu_ops);
}

static int innodpu_platform_remove(struct platform_device *pdev)
{
	component_del(&pdev->dev, &innodpu_ops);
	return 0;
}

static int innodpu_dpu_suspend(struct device *dev)
{
	plat_data_t *pdata;
	int dpu_id = 0;
	unsigned dpu_nums = hal_get_dev_nums(dev->parent, DEV_DPU);
	unsigned db9000_nums = hal_get_dev_nums(dev->parent, DEV_DB9000);
	unsigned pdp0_nums = hal_get_dev_nums(dev->parent, DEV_PDP0);

	pdata = dev->platform_data;
	dpu_id = pdata->dev_idx;

	if (dpu_id > dpu_nums) {
		innodpu_err(dev, "Invaild dpu_id:%d\n", dpu_id);
		return -EINVAL;
	}

	if (!(s_dpu_nums & BIT(dpu_id)))
		return 0;

	if (dpu_id < db9000_nums) {
		innodpu_warn(dev, "Not support db9000\n");
		return 0;
	} else {
		if (dpu_id - db9000_nums < pdp0_nums)
			return innodpu_pdp0_suspend(dev, dpu_id);
		else
			innodpu_warn(dev, "Not support dpu_id:%d\n", dpu_id);
	}

	return -EINVAL;
}

static void innodpu_dpu_shutdown(struct platform_device *pdev)
{
	plat_data_t *pdata;
	int dpu_id = 0;
	struct device *dev = &pdev->dev;
	unsigned dpu_nums = hal_get_dev_nums(dev->parent, DEV_DPU);
	unsigned db9000_nums = hal_get_dev_nums(dev->parent, DEV_DB9000);
	unsigned pdp0_nums = hal_get_dev_nums(dev->parent, DEV_PDP0);

	pdata = dev->platform_data;
	dpu_id = pdata->dev_idx;

	if (dpu_id > dpu_nums) {
		innodpu_err(dev, "Invaild dpu_id:%d\n", dpu_id);
		return;
	}

	if (!(s_dpu_nums & BIT(dpu_id)))
		return;

	if (dpu_id < db9000_nums) {
		innodpu_warn(dev, "Not support db9000\n");
		return;
	} else {
		if (dpu_id - db9000_nums < pdp0_nums)
			innodpu_pdp0_shutdown(dev, dpu_id);
		else
			innodpu_warn(dev, "Not support dpu_id:%d\n", dpu_id);
	}

	return;

}

static int innodpu_dpu_resume(struct device *dev)
{
	plat_data_t *pdata;
	int dpu_id = 0;

	unsigned dpu_nums = hal_get_dev_nums(dev->parent, DEV_DPU);
	unsigned db9000_nums = hal_get_dev_nums(dev->parent, DEV_DB9000);
	unsigned pdp0_nums = hal_get_dev_nums(dev->parent, DEV_PDP0);

	pdata = dev->platform_data;
	dpu_id = pdata->dev_idx;

	if (dpu_id > dpu_nums) {
		innodpu_err(dev, "Invaild dpu_id:%d\n", dpu_id);
		return -EINVAL;
	}

	if (!(s_dpu_nums & BIT(dpu_id)))
		return 0;

	if (dpu_id < db9000_nums) {
		innodpu_warn(dev, "Not support db9000\n");
		return 0;
	} else {
		if (dpu_id - db9000_nums < pdp0_nums)
			return innodpu_pdp0_resume(dev, dpu_id);
		else
			innodpu_warn(dev, "Not support dpu_id:%d\n", dpu_id);
	}

	return -EINVAL;

}

static const struct dev_pm_ops innodpu_dpu_pm_ops = {
	SET_SYSTEM_SLEEP_PM_OPS(innodpu_dpu_suspend, innodpu_dpu_resume)
};

struct platform_driver g_innogpu_dpu_driver = {
	.probe = innodpu_platform_probe,
	.remove = innodpu_platform_remove,
	.shutdown = innodpu_dpu_shutdown,
	.driver = {
		.name = INNO_DPU_DEVICE_NAME,
		.pm = &innodpu_dpu_pm_ops,
		},
};
