/*************************************************************************/ /*!
@File			innodpu_drm_drv.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include <linux/platform_device.h>
#include "syscommon.h"
#include "innodpu_drm_gem.h"
#include "innodpu_drm_drv.h"
#include "innodpu_drm_fb.h"
#include "innodpu_common.h"
#include "innodpu_dp_common.h"
#include "innodpu_connector.h"
#include "innodpu_drm_pm.h"
#include "pdp_drm.h"
#include "innogpu_drm.h"
#include "kernel_compatibility.h"

#define SWAP_SHMEM_ZEO
#define DPU_BACKUP_ALL_VRAM

unsigned int s_dpu_nums = 0x0f;
module_param(s_dpu_nums, uint, 0444);
MODULE_PARM_DESC(s_dpu_nums, "dpu driver register(default: 0x0f)");

unsigned int s_hdmi_nums = 0x3;
module_param(s_hdmi_nums, uint, 0444);
MODULE_PARM_DESC(s_hdmi_nums,"hdmi driver register(default: 3)");

unsigned int s_dp_nums = 1;
module_param(s_dp_nums, uint, 0444);
MODULE_PARM_DESC(s_dp_nums, "dp driver register(default: 1)");

unsigned int s_lvds_nums = 1;
module_param(s_lvds_nums, uint, 0444);
MODULE_PARM_DESC(s_lvds_nums, "lvds driver register(default: 1)");

unsigned int s_vga_nums = 1;
module_param(s_vga_nums, uint, 0444);
MODULE_PARM_DESC(s_vga_nums, "vga driver register(default: 0)");

unsigned int s_logo_extime = 5;
module_param(s_logo_extime, uint, 0644);
MODULE_PARM_DESC(s_logo_extime, "Controls the minimum execution"
			" time of processes displayed the logo(default: 5s)");

unsigned int s_hw_cursor = 0x00;
module_param(s_hw_cursor, uint, 0444);
MODULE_PARM_DESC(s_hw_cursor, "hw cursor driver register(default: 0x00)");

/*
+ * G0M
+ * 0-4K+4K
+ *
+ */
unsigned int s_dpu_match = 0x0;
module_param(s_dpu_match, uint, 0444);
MODULE_PARM_DESC(s_dpu_match, "dpu match register(default: 0x00)");

unsigned int s_dpu_gem_imme = 0x2;
module_param(s_dpu_gem_imme, uint, 0444);
MODULE_PARM_DESC(s_dpu_gem_imme, "dpu gem obj clear immediately (2-yes, 0-no (default: 2))");

#if (DRM_VERSION < KERNEL_VERSION(5, 8, 0))
int innogpu_drm_debugfs_init(struct drm_minor *minor);
#else
void innogpu_drm_debugfs_init(struct drm_minor *minor);
#endif

#if (DRM_VERSION < KERNEL_VERSION(4, 12, 0))
void innogpu_drm_debugfs_cleanup(struct drm_minor *minor);
#endif


/*
 * function :innodpu_update_user_id
 * desc:
 */
static __maybe_unused void innodpu_update_user_id(struct innodpu_drm_private *dev_priv)
{
	int i = 0;
	uint64_t name_len = inno_strlen(current->comm);
	unsigned int len = dev_priv->pm_backup_array_len;
	struct innodpu_gem_backuplist *backup_array = dev_priv->pm_backup_array;

	/* no try to update if backup array not init success */
	if (len == 0 || backup_array == NULL) {
		return;
	}

	for (i = 0; i < len; ++i){
		if (!backup_array[i].pm_bakup) {
			continue;
		}

		if (backup_array[i].full_match) {
			if (inno_strncmp(current->comm, backup_array[i].name, name_len) == 0) {
				spin_lock(&dev_priv->user_id_lock);
				backup_array[i].user_id = current->tgid;
				spin_unlock(&dev_priv->user_id_lock);
				return;
			}
		} else {
			if (inno_strstr(current->comm, backup_array[i].name)) {
				spin_lock(&dev_priv->user_id_lock);
				backup_array[i].user_id = current->tgid;
				spin_unlock(&dev_priv->user_id_lock);
				return;
			}
		}
	}

	return;
}

bool innodpu_gem_need_backup(struct innodpu_drm_private *dev_priv, char *obj_name)
{
	int i = 0;
	bool ret = false;
	char *id_pos = NULL;
	pid_t current_tgid = 0, front_user_id = 0;
	unsigned int len = dev_priv->pm_backup_array_len;
	struct innodpu_gem_backuplist *backup_array = dev_priv->pm_backup_array;

#ifdef DPU_BACKUP_ALL_VRAM
	return true;
#endif

	/* no backup if backup array not init success */
	if (len == 0 || backup_array == NULL) {
		return false;
	}

	for (i = 0; i < len; ++i){

		if (!backup_array[i].pm_bakup) {
			continue;
		}

		if (inno_strstr(obj_name, backup_array[i].name)) {
			id_pos = strrchr(obj_name, '-');
			if ((id_pos != NULL) && (inno_sscanf(id_pos, "-%d", &current_tgid) == 1)) {

				spin_lock(&dev_priv->user_id_lock);
				front_user_id = backup_array[i].user_id;
				spin_unlock(&dev_priv->user_id_lock);

				if (current_tgid == front_user_id) {
					ret = true;
					break;
				}
			}
		}
	}

	return ret;
}

/*
 * function:
 * 1. Because the Linux system will restart the new xorg when switching users(not logging out),
 * bug the memory requested by the previous user's Xorg is not released by the UMD, resulting
 * in less adn less memory.
 * 2. UMD confirms that after switching users, the previous user's vram is no longer used.
 * 3. UMD confirms that when switching users, the newly started Xorg must apply for vram first,
 * and can index users according to Xorg's process ID.
 * 4. The DPU applies for s_dpu_shared_mem_size size of the video memory to be used for allowlist processes,
 * shared among multiple users
 * 5. this solution is risky, it is only a temporary solution to the short-term memory, the final solution is
 * waiting for the application to release.
 */

extern struct inno_gem_private *inno_gem_init_shared(struct drm_device *drm_dev,
		struct inno_shared_vram * p_shared_vram_info, unsigned char flags,unsigned long vram_zero_base_dev,
		unsigned long vram_zero_base_cpu, unsigned long vram_zero_size);
extern void inno_gem_share_free(struct drm_device *drm_dev,
		struct inno_shared_vram * pshared_vram);
extern int inno_gem_share_alloc(struct drm_device *drm_dev,
		struct inno_shared_vram * pshared_vram, unsigned long mm_size);

struct innodpu_shmem_allowlist {
	const char *name;
	bool full_match;
};

static __maybe_unused struct innodpu_shmem_allowlist s_shmem_allowlist[] =
{
	{"Xorg", 1},
	{"X", 1},
	{"kwin_x11", 0},
	{"browse", 0},
	{"cdos-desktop", 1},
	{"chrome", 1},
	{"firefox", 0}
};

/*
 * function: Check if the memory requested from the dpu can be shared by multiple users
 */
static __maybe_unused bool innodpu_check_memory_shared(struct innodpu_drm_private *dev_priv)
{
	int i = 0;
	uint64_t name_len = inno_strlen(current->comm);

	if (!dev_priv->has_shared_mem)
		return false;
	for (i=0; i<ARRAY_SIZE(s_shmem_allowlist); ++i){
		if (s_shmem_allowlist[i].full_match) {
			if (inno_strncmp(current->comm, s_shmem_allowlist[i].name, name_len) == 0)
				return true;
		} else {
			if (inno_strstr(current->comm, s_shmem_allowlist[i].name))
				return true;
		}
	}

	return false;
}

static __maybe_unused int innodpu_multiuser_add(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv, void *data, struct drm_file *drm_file)
{
	int ret = 0;
	int retry = 3;
	unsigned long flags = 0;
	struct inno_shared_vram * pshared_vram = &dev_priv->shared_vram_info;
	struct drm_pdp_user_set *puser_add = data;
	struct inno_shared_vram_user *puser = NULL;
	struct inno_shared_vram_user *tmp_user = NULL;

	if(!dev_priv->has_shared_mem) {
		innodpu_err(drm_dev->dev, "innodpu not support multi-user.\n");
		return -EINVAL;
	}

	spin_lock(&pshared_vram->user_idr_lock);
	tmp_user = idr_find(&pshared_vram->uer_idr, puser_add->user_id);
	spin_unlock(&pshared_vram->user_idr_lock);
	if (tmp_user) {
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "innodpu user-id %lld already exits.\n", puser_add->user_id);
		return 0;
	}

	puser = devm_kmalloc(drm_dev->dev, sizeof(*puser), GFP_KERNEL);
	if (!puser) {
		innodpu_err(drm_dev->dev, "innodpu alloc user_info failed, short of memory.\n");
		return -ENOMEM;
	}
	memset(puser, 0, sizeof(*puser));

	flags = GEM_MANAGE_DRM_MM | GEM_FREE_IMME;
	puser->gem_priv = inno_gem_init_shared(drm_dev, pshared_vram, flags,
		dev_priv->vram_zero_base_dev, dev_priv->vram_zero_base_cpu, dev_priv->vram_zero_size);
	puser->user_id = puser_add->user_id;

	do {
		spin_lock(&pshared_vram->user_idr_lock);
		ret = idr_alloc(&pshared_vram->uer_idr, puser, puser_add->user_id, puser_add->user_id+1, GFP_ATOMIC);
		spin_unlock(&pshared_vram->user_idr_lock);
		if (ret == puser_add->user_id)
			break;
	} while (--retry > 0);
	if (retry <=0)
		innodpu_err(drm_dev->dev, "innodpu push user(%p)-%lld error.\n", puser, puser->user_id);

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "innodpu push user(%p)-%lld.\n", puser, puser->user_id);

	return ret;
}

static __maybe_unused int innodpu_multiuser_set(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv, void *data, struct drm_file *drm_file)
{
	struct inno_shared_vram * pshared_vram = &dev_priv->shared_vram_info;
	struct drm_pdp_user_set *puser_set = data;
	struct inno_shared_vram_user *puser = NULL;

	spin_lock(&pshared_vram->user_idr_lock);
	puser = idr_find(&pshared_vram->uer_idr, puser_set->user_id);
	spin_unlock(&pshared_vram->user_idr_lock);
	if (!puser) {
		innodpu_err(drm_dev->dev,"innodpu can't switch user-%lld, need to set.\n", puser_set->user_id);
		return -ENODEV;
	}
	pshared_vram->current_user = puser;

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "innodpu switch user(%p)-%ld.\n", puser, puser->user_id);
	return 0;
}

static __maybe_unused int innodpu_multiuser_del(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv, void *data, struct drm_file *drm_file)
{
	struct inno_shared_vram * pshared_vram = &dev_priv->shared_vram_info;
	struct drm_pdp_user_del *puser_del = data;
	struct inno_shared_vram_user *puser = NULL;

	spin_lock(&pshared_vram->user_idr_lock);
	puser = idr_find(&pshared_vram->uer_idr, puser_del->user_id);
	spin_unlock(&pshared_vram->user_idr_lock);
	if (!puser) {
		innodpu_err(drm_dev->dev,"innodpu can't del user-%lld, need to set.\n", puser_del->user_id);
		return -ENODEV;
	}

	spin_lock(&pshared_vram->user_idr_lock);
	idr_remove(&pshared_vram->uer_idr, puser_del->user_id);
	spin_unlock(&pshared_vram->user_idr_lock);

	inno_gem_fini(drm_dev, puser->gem_priv);

	if (puser == pshared_vram->current_user)
		pshared_vram->current_user = NULL;

	devm_kfree(drm_dev->dev, puser);

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "innodpu del user(%p)-%ld.\n", puser, puser->user_id);
	return 0;
}

static __maybe_unused void innodpu_xorg_monitor_switch_user(struct drm_device *drm_dev,
		struct inno_shared_vram * pshared_vram)
{
	int tgid = current->tgid;
	struct innodpu_drm_private *dev_priv = NULL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return;
	}

	if ((!pshared_vram->current_user) || (pshared_vram->current_user->user_id != tgid)) {
		struct drm_pdp_user_add user_add = {.user_id=tgid, .flags=0};
		struct drm_pdp_user_set user_set = {.user_id=tgid, .flags=0};

#ifdef SWAP_SHMEM_ZEO
		{
			unsigned int vm_size = 0;
			unsigned int vm_offset = 0;
			struct inno_shared_vram * pshared_vram = &dev_priv->shared_vram_info;
			unsigned long vram_zero_size = dev_priv->visiable_gem_priv->vram_zero_size;
			unsigned long vram_share_size = pshared_vram->vram_share_size;
			void * axi_vram_src = (void *)dev_priv->visiable_gem_priv->vram_zero_base_dev;
			void *axi_vram_dst = NULL;

			for (vm_offset = 0; vm_offset < vram_share_size; vm_offset += vm_size) {
				vm_size = ((vram_share_size - vm_offset) < vram_zero_size) ?
					(vram_share_size - vm_offset) : vram_zero_size;
				axi_vram_dst = (void *)pshared_vram->vram_share_devaddr + vm_offset;
				innodpu_dma_memcpy(drm_dev->dev, &axi_vram_src, &axi_vram_dst, &vm_size, 1, GDDR2GDDR);
			}
		}
#endif

		if (innodpu_multiuser_add(drm_dev, dev_priv, &user_add, NULL) < 0)
			return;

		if (pshared_vram->current_user && pshared_vram->current_user->gem_priv)
			flush_work(&pshared_vram->current_user->gem_priv->fb_free_work);
		innodpu_multiuser_set(drm_dev, dev_priv, &user_set, NULL);
	}
}

int innodpu_gem_dumb_create(struct drm_file *drm_file,
		struct drm_device *drm_dev, struct drm_mode_create_dumb *args)
{
	int ret = 0;
	struct inno_shared_vram *pshared_vram;
	struct innodpu_drm_private *dev_priv = NULL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return -EINVAL;
	}
	pshared_vram = &dev_priv->shared_vram_info;

	if ((args->flags & DUMB_USED_GTT) && (!dev_priv->has_gtt_mem)) {
		innodpu_err(drm_dev->dev, "request gtt mem form dpu, but dpu not support\n");
		return -ENOMEM;
	}

	if (dev_priv->has_shared_mem && pshared_vram != NULL &&
		!(args->flags & DUMB_USED_GTT)) {
		if (innodpu_check_memory_shared(dev_priv)) {
			if ((strncmp(current->comm, "Xorg", 4)==0) || (strncmp(current->comm, "X", 1)==0))
				innodpu_xorg_monitor_switch_user(drm_dev, pshared_vram);

			if (pshared_vram->share_is_visiable ||
						(!!pshared_vram->share_is_visiable == !(args->flags & DUMB_USED_INVISIABLE))) {
				if (inno_gem_dumb_create(drm_file, drm_dev, pshared_vram->current_user->gem_priv, args) == 0) {
					innodpu_update_user_id(dev_priv);
					return 0;
				}
			}
		}
	}

	if ((args->flags & DUMB_USED_INVISIABLE) && (dev_priv->has_inv_mem)) {
		ret = inno_gem_dumb_create(drm_file, drm_dev, dev_priv->invisiable_gem_priv, args);
	} else if ((args->flags & DUMB_USED_GTT) && (dev_priv->has_gtt_mem)) {
		ret = inno_gem_dumb_create(drm_file, drm_dev, dev_priv->gtt_gem_priv, args);
	} else {
		ret = inno_gem_dumb_create(drm_file, drm_dev, dev_priv->visiable_gem_priv, args);
	}

	if (ret == 0) {
		innodpu_update_user_id(dev_priv);
	}

	return ret;
}

int innodpu_gem_object_create_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *drm_file)
{
	int ret = 0;
	struct drm_pdp_gem_create *args = data;
	struct inno_shared_vram *pshared_vram;
	struct innodpu_drm_private *dev_priv = NULL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return -EINVAL;
	}
	pshared_vram = &dev_priv->shared_vram_info;

	if ((args->flags & DUMB_USED_GTT) && (!dev_priv->has_gtt_mem)) {
		innodpu_err(drm_dev->dev, "request gtt mem form dpu, but dpu not support\n");
		return -ENOMEM;
	}

	if (dev_priv->has_shared_mem && pshared_vram != NULL &&
		!(args->flags & DUMB_USED_GTT)) {
		if (innodpu_check_memory_shared(dev_priv)) {
			if ((strncmp(current->comm, "Xorg", 4)==0) || (strncmp(current->comm, "X", 1)==0))
				innodpu_xorg_monitor_switch_user(drm_dev, pshared_vram);

			if (pshared_vram->share_is_visiable ||
						(!!pshared_vram->share_is_visiable == !(args->flags & DUMB_USED_INVISIABLE))) {
				if (inno_gem_create_priv(drm_file, drm_dev, pshared_vram->current_user->gem_priv, args) == 0) {
					innodpu_update_user_id(dev_priv);
					return 0;
				}
			}
		}
	}

	// used invisiable gem
	if ((args->flags & DUMB_USED_INVISIABLE) && (dev_priv->has_inv_mem)) {
		ret = inno_gem_create_priv(drm_file, drm_dev, dev_priv->invisiable_gem_priv, args);
	} else if ((args->flags & DUMB_USED_GTT) && (dev_priv->has_gtt_mem)) {
		ret = inno_gem_create_priv(drm_file, drm_dev, dev_priv->gtt_gem_priv, args);
	} else {
		ret = inno_gem_create_priv(drm_file, drm_dev, dev_priv->visiable_gem_priv, args);
	}
	if (ret == 0) {
		innodpu_update_user_id(dev_priv);
	}

	return ret;
}

int innodpu_drm_chipinfo_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *drm_file)
{
	struct drm_pdp_chip_info *args = data;
	struct innodpu_drm_private *dev_priv = NULL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return -EINVAL;
	}

	args->chip_type = hal_get_chiptype(drm_dev->dev);
	args->support_inv = hal_has_inv_mem(drm_dev->dev);
	args->support_shmem = dev_priv->has_shared_mem;

	return 0;
}

int inno_gem_object_mmap_ioctl(struct drm_device *drm_dev, void *data,
		struct drm_file *drm_file)
{
	struct drm_pdp_gem_mmap *args = (struct drm_pdp_gem_mmap *)data;

	return drm_gem_dumb_map_offset(drm_file, drm_dev, args->handle, &args->offset);
}

int inno_gem_object_plane_fd_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	struct innodpu_drm_private *dev_priv = NULL;
	struct drm_pdp_base_fd *args = (struct drm_pdp_base_fd *)data;
	struct drm_plane *plane = NULL;
	struct drm_crtc *crtc = NULL;
	struct drm_gem_object *obj = NULL;
	struct drm_framebuffer *bfb = NULL;
	struct innodpu_pdp0_drm *pdp0_drm = NULL;
	struct drm_display_mode mode = {0};
	unsigned int vblank_us = 16700;
	int ret = -EINVAL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return -1;
	}

	if (!s_dpu_support_plane_fd || !dev_priv->drm_nulldisplay) {
		return 0;
	}
	if (args == NULL) {
		return -1;
	}

	drm_for_each_plane(plane, drm_dev) {
		if (plane->base.id == args->plane_id) {
			ret = 0;
			break;
		}
	}
	if (ret) {
		innodpu_err(drm_dev, "Not found plane id %d\n", args->plane_id);
		return ret;
	}

	if (plane->state && plane->state->crtc) {
		crtc = plane->state->crtc;
		mode = crtc->state->mode;
	} else if (!plane->state && plane->crtc) {
		crtc = plane->crtc;
		mode = crtc->mode;
	} else {
		innodpu_err(drm_dev, "Plane %d not crtc\n", args->plane_id);
		return -EIO;
	}

	/* calc wait time us, by frame_id and frame rate */
	pdp0_drm = crtc_to_pdp0_device(crtc);
	if (mode.clock) {
		vblank_us = 1000000 / drm_mode_vrefresh(&mode);
	}

	if (args->frame_id > atomic64_read(&pdp0_drm->active_count)) {
		vblank_us *= (atomic64_read(&pdp0_drm->active_count) - args->frame_id);
	} else {
		/* at lease wait next active region comming */
		args->frame_id = atomic64_read(&pdp0_drm->active_count) + 1;
	}

	/* wait target active region access */
	pdp0_drm = crtc_to_pdp0_device(crtc);
	if (pdp0_drm->active_wq != NULL) {
		ret = inno_wait_event_interruptible_timeout(pdp0_drm->active_wq,
			atomic64_read(&pdp0_drm->active_count) >= args->frame_id,
			usecs_to_jiffies(vblank_us + 100000)); /* add 100ms as margin */
		if (ret < 0) {
			innodpu_err(drm_dev, "%s wait %d img failed\n", __func__, args->plane_id);
			return ret;
		} else if (ret == 0) {
			innodpu_err(drm_dev, "%s wait %d img valid timeout\n", __func__, args->plane_id);
			return -ETIMEDOUT;
		}
	}

	/* get bfb here */
	if (plane->state && plane->state->fb) {
		bfb = plane->state->fb;
	} else if (!plane->state && plane->fb) {
		bfb = plane->fb;
	} else {
		innodpu_err(drm_dev, "Plane %d not fb\n", args->plane_id);
		return -EIO;
	}

	ret = 0;

	/* plane obj */
	obj = bfb->obj[0];
	args->format = bfb->format->format;
	args->frame_size = obj->size;
	args->width = bfb->width;
	args->height = bfb->height;
	args->obj_idr = obj->name;
	args->frame_id = atomic64_read(&pdp0_drm->active_count);

	/* todo: support cursor_obj */
	args->cursor_obj_idr = 0;
	args->cursor_x = 0;
	args->cursor_y = 0;
	args->cursor_width = 0;
	args->cursor_height = 0;
	args->cursor_format = 0;

	if (args->obj_idr <= 0) {
		innodpu_err(drm_dev, "unknow err of get plane[%d] obj_name[%d]\n", args->plane_id, args->obj_idr);
	}

	return ret;
}


#ifndef CONFIG_VIDEOMODE_HELPERS
/**
 * drm_display_mode_to_videomode - fill in @vm using @dmode,
 * @dmode: drm_display_mode structure to use as source
 * @vm: videomode structure to use as destination
 *
 * Fills out @vm using the display mode specified in @dmode.
 */
void drm_display_mode_to_videomode(
		const struct drm_display_mode *dmode,struct videomode *vm)
{
	vm->hactive = dmode->hdisplay;
	vm->hfront_porch = dmode->hsync_start - dmode->hdisplay;
	vm->hsync_len = dmode->hsync_end - dmode->hsync_start;
	vm->hback_porch = dmode->htotal - dmode->hsync_end;

	vm->vactive = dmode->vdisplay;
	vm->vfront_porch = dmode->vsync_start - dmode->vdisplay;
	vm->vsync_len = dmode->vsync_end - dmode->vsync_start;
	vm->vback_porch = dmode->vtotal - dmode->vsync_end;

	vm->pixelclock = dmode->clock * 1000;

	vm->flags = 0;
	if (dmode->flags & DRM_MODE_FLAG_PHSYNC)
		vm->flags |= DISPLAY_FLAGS_HSYNC_HIGH;
	else if (dmode->flags & DRM_MODE_FLAG_NHSYNC)
		vm->flags |= DISPLAY_FLAGS_HSYNC_LOW;
	if (dmode->flags & DRM_MODE_FLAG_PVSYNC)
		vm->flags |= DISPLAY_FLAGS_VSYNC_HIGH;
	else if (dmode->flags & DRM_MODE_FLAG_NVSYNC)
		vm->flags |= DISPLAY_FLAGS_VSYNC_LOW;
	if (dmode->flags & DRM_MODE_FLAG_INTERLACE)
		vm->flags |= DISPLAY_FLAGS_INTERLACED;
	if (dmode->flags & DRM_MODE_FLAG_DBLSCAN)
		vm->flags |= DISPLAY_FLAGS_DOUBLESCAN;
	if (dmode->flags & DRM_MODE_FLAG_DBLCLK)
		vm->flags |= DISPLAY_FLAGS_DOUBLECLK;
}
#endif

#if (DRM_VERSION >= KERNEL_VERSION(5, 14, 0))
int drm_fb_helper_remove_conflicting_framebuffers(struct apertures_struct *a,
		const char *name, bool primary)
{
#if IS_REACHABLE(CONFIG_FB)
	return remove_conflicting_framebuffers(a, name, primary);
#else
	return 0;
#endif
}
#endif

static void innodpu_kick_off_efifb(struct drm_device *drm_dev)
{
	bool do_free = false;
	struct apertures_struct *a;

	a = alloc_apertures(1);
	if (!a) {
		innodpu_err(drm_dev->dev, "Short of memory,innodrmfb will not take over fbcon !\n");
		return;
	}

	a->ranges[0].base = 0;
	a->ranges[0].size = ~0;
	do_free = true;
	drm_fb_helper_remove_conflicting_framebuffers(a, "innodrmfb", true);
	if (do_free)
		kfree(a);
}

int innodpu_fb_init(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	int ret = 0;

	if (!dev_priv->drm_nulldisplay) {
		innodpu_kick_off_efifb(drm_dev);
		innodpu_info(drm_dev->dev, DPU_UT_INIT, "Start init fbdev device.\n");
		ret = innodpu_fbdev_init(drm_dev);
		if (ret) {
			innodpu_err(drm_dev->dev,"Failed to Init fbdev  (err=%d)\n", ret);
			return ret;
		}
	}

	return 0;
}

void innodpu_fb_fini(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	if (!dev_priv->drm_nulldisplay) {
		innodpu_fbdev_fini(drm_dev);
	}
}

int innodpu_pm_init(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	if (!dev_priv->drm_nulldisplay) {
		dev_priv->pm_nb.notifier_call = innodpu_pm_notifier;
		register_pm_notifier(&dev_priv->pm_nb);
	}

	return 0;
}

void innodpu_pm_fini(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	if (!dev_priv->drm_nulldisplay) {
		unregister_pm_notifier(&dev_priv->pm_nb);
	}
}

static int innodpu_chip_init(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	return innodpu_custom_init(drm_dev, dev_priv);
}

int innodpu_early_load(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv, struct device *dev)
{
	int err = 0;
	dev_priv->dev = dev;
	dev_priv->drm_dev = drm_dev;
	dev_priv->display_enabled = true;
	dev_priv->has_shared_mem = false;
	dev_priv->has_gtt_mem = false;
	dev_priv->invisiable_gem_priv = NULL;
	dev_priv->logo_end_ktime = 0;
	mutex_init(&dev_priv->init_lock);
	mutex_init(&dev_priv->logo_lock);

	err  = innogpu_drm_get_nulldisp(drm_dev, &dev_priv->drm_nulldisplay);
	if (err < 0) {
		innodpu_err(drm_dev->dev, "Get nulldisp failed (err=%d)\n", err);
		return err;
	}
	err = innogpu_drm_has_invmem(drm_dev, &dev_priv->has_inv_mem);
	if (err < 0) {
		innodpu_err(drm_dev->dev, "Get Invmem failed (err=%d)\n", err);
		return err;
	}

	err = innodpu_modeset_early_init(drm_dev, dev_priv);
	if (err) {
		innodpu_err(drm_dev->dev, "Early modeset initialisation failed (err=%d)\n", err);
		goto err_modeset_early_init;
	}

	innodpu_chip_init(drm_dev, dev_priv);
	innodpu_info(drm_dev->dev, DPU_UT_INIT,"Early modeset initialization completed.\n");

	return 0;

err_modeset_early_init:
	return err;
}

void innodpu_early_unload(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0))
	/* Called by drm_dev_fini in Linux 4.11.0 and later */
	drm_vblank_cleanup(drm_dev);
#endif
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
int innogpu_drm_debugfs_init(struct drm_minor *minor)
#else
void innogpu_drm_debugfs_init(struct drm_minor *minor)
#endif
{

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
	int ret = 0;
	ret = innodpu_debugfs_init(minor);
	return ret;
#else
	return;
#endif
}

static int innodpu_vram_init(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	int err = 0;

	innodpu_info(drm_dev->dev, DPU_UT_INIT,"dpu vram init: nulldisp-%d,shmem-%d,imme-%d.\n",
				dev_priv->drm_nulldisplay, s_dpu_has_shared_mem, s_dpu_gem_imme);

	if ((!dev_priv->drm_nulldisplay) && s_dpu_has_shared_mem) {
		err = inno_gem_share_alloc(drm_dev, &dev_priv->shared_vram_info,
					s_dpu_shared_mem_size / hal_get_gpu_core_nums());
		if (err < 0) {
			return -ENOMEM;
		}
		idr_init_base(&dev_priv->shared_vram_info.uer_idr, 0);
		spin_lock_init(&dev_priv->shared_vram_info.user_idr_lock);
		dev_priv->shared_vram_info.current_user = NULL;
		dev_priv->has_shared_mem = true;
	}

	dev_priv->visiable_gem_priv = inno_gem_init(drm_dev, true, s_dpu_gem_imme);
	if (!dev_priv->visiable_gem_priv) {
		innodpu_err(drm_dev->dev, "Inno visiable gem init failed. Short of memory.\n");
		err = -ENOMEM;
		goto err_gem_init;
	}

	if (dev_priv->has_inv_mem) {
		dev_priv->invisiable_gem_priv = inno_gem_init(drm_dev, false, s_dpu_gem_imme);
		if (!dev_priv->invisiable_gem_priv) {
			innodpu_err(drm_dev->dev, "Inno invisiable gem init failed. Short of memory.\n");
			err = -EINVAL;
			goto err_modeset_early_init;
		}
		dev_priv->invisiable_gem_priv->vram_zero_base_dev = dev_priv->visiable_gem_priv->vram_zero_base_dev;
		dev_priv->invisiable_gem_priv->vram_zero_base_cpu = dev_priv->visiable_gem_priv->vram_zero_base_cpu;
		dev_priv->invisiable_gem_priv->vram_zero_size = dev_priv->visiable_gem_priv->vram_zero_size;
	}
	dev_priv->vram_zero_base_dev = dev_priv->visiable_gem_priv->vram_zero_base_dev;
	dev_priv->vram_zero_base_cpu = dev_priv->visiable_gem_priv->vram_zero_base_cpu;
	dev_priv->vram_zero_size = dev_priv->visiable_gem_priv->vram_zero_size;

	if (hal_has_gtt_mem(drm_dev->dev) && s_dpu_has_gtt_mem) {
		dev_priv->gtt_gem_priv = inno_gem_init(drm_dev, false, s_dpu_gem_imme | BIT(2));
		if (!dev_priv->gtt_gem_priv) {
			innodpu_err(drm_dev->dev, "Inno gtt gem init failed. Short of memory.\n");
			err = - EINVAL;
			goto err_modeset_early_init;
		}
		dev_priv->has_gtt_mem = true;
	}
	return err;

err_modeset_early_init:
	if (dev_priv->has_inv_mem && dev_priv->invisiable_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->invisiable_gem_priv);
	if (dev_priv->visiable_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->visiable_gem_priv);
	if (dev_priv->has_gtt_mem && dev_priv->gtt_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->gtt_gem_priv);

err_gem_init:
	if ((!dev_priv->drm_nulldisplay) && s_dpu_has_shared_mem) {
		idr_destroy(&dev_priv->shared_vram_info.uer_idr);
		if (dev_priv->has_shared_mem)
			inno_gem_share_free(drm_dev, &dev_priv->shared_vram_info);
		dev_priv->has_shared_mem = false;
	}
	return err;
}

static void innodpu_vram_fini(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	if (dev_priv->has_inv_mem && dev_priv->invisiable_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->invisiable_gem_priv);
	if (dev_priv->visiable_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->visiable_gem_priv);
	if (dev_priv->has_gtt_mem && dev_priv->gtt_gem_priv)
		inno_gem_fini(drm_dev, dev_priv->gtt_gem_priv);
	if ((!dev_priv->drm_nulldisplay) && s_dpu_has_shared_mem) {
		idr_destroy(&dev_priv->shared_vram_info.uer_idr);
		if (dev_priv->has_shared_mem)
			inno_gem_share_free(drm_dev, &dev_priv->shared_vram_info);
		dev_priv->has_shared_mem = false;
	}
}

int innodpu_late_load(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	int err = 0;

	err = innodpu_vram_init(drm_dev, dev_priv);
	if (err < 0) {
		innodpu_err(drm_dev->dev, "late modeset initialisation failed (err=%d)\n", err);
		goto err_vram_init;
	}

	innodpu_info(drm_dev->dev, DPU_UT_INIT,"drm vblank init.\n");
	// drm_vblank_init：vblank自动释放 (drm_vblank_init_release)
	err = drm_vblank_init(drm_dev, drm_dev->mode_config.num_crtc);
	if (err < 0) {
		innodpu_err(drm_dev->dev, "Failed to complete vblank init (err=%d)\n", err);
		goto err_vblank_init;
	}
#if (DRM_VERSION < KERNEL_VERSION(4, 7, 0))
	drm_dev->vblank_disable_allowed = 1;
#endif

	innodpu_info(drm_dev->dev, DPU_UT_INIT,"modeset late init.\n");
	err = innodpu_modeset_late_init(drm_dev, dev_priv);
	if (err) {
		innodpu_err(drm_dev->dev, "late modeset initialisation failed (err=%d)\n", err);
		goto err_modeset_late_init;
	}

	return err;

err_modeset_late_init:
#if (DRM_VERSION < KERNEL_VERSION(4, 7, 0))
	drm_dev->vblank_disable_allowed = 0;
#endif

err_vblank_init:
	innodpu_vram_fini(drm_dev, dev_priv);

err_vram_init:
	return err;
}

void innodpu_late_unload(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv)
{
	drm_kms_helper_poll_fini(drm_dev);
	drm_mode_config_cleanup(drm_dev);
	innodpu_vram_fini(drm_dev, dev_priv);
}

void innodpu_bind_atomic(struct drm_device * drm_dev, bool lock)
{
	struct innodpu_drm_private *dev_priv = NULL;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "innodpu dev_priv is NULL.\n");
		return;
	}

	if (lock) {
		mutex_lock(&dev_priv->init_lock);
	} else {
		mutex_unlock(&dev_priv->init_lock);
	}
}
