/*************************************************************************/ /*!
@File			innodpu_drm_gem.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include "inno_drm_version.h"
#if (DRM_VERSION >= KERNEL_VERSION(5, 5, 0))
#include <drm/drm_drv.h>
#include <drm/drm_prime.h>
#include <linux/platform_device.h>
#endif

#include <linux/dma-buf.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <linux/capability.h>
#include <drm/drm_mm.h>
#if (DRM_VERSION >= KERNEL_VERSION(5, 5, 0))
#include <drm/drm_device.h>
#include <drm/drm_file.h>
#else
#include <drm/drmP.h>
#endif

#include "innogpu_pci_drv.h"
#include "innogpu.h"
#include "osfunc_common.h"
#include "innodpu_drm_gem.h"
#include "pdp_drm.h"
#include "kernel_compatibility.h"
#include "innogpu_drm.h"
#include "pvrsrv_memallocflags.h"


static bool s_dpu_gem_print = false;
module_param(s_dpu_gem_print, bool, 0600);
MODULE_PARM_DESC(s_dpu_gem_print, "s3/s4 print inno obj suspend data or not (default: false)");

// 在缺页中断中：
// 1. 用户进行访问时-fault，分配内存页，存储到gem链表中
// 2. 在close时，将更新得内存页同步到显存后，释放内存
#define GEM_MMAP_PER_SIZE (0x100000)
#define GEM_ZERO_SIZE     (0x100000)
#define DRM_GEM_IMME_MEMCLEAR


static __maybe_unused void inno_gem_vbuffer_clear(struct inno_gem_private *gem_priv,
	struct inno_gem_object *inno_obj, struct device* pdev);

/* for vga auto adapt */
#define VGA_MAX_WIDTH 1920
#define VGA_MAX_HEIGHT 1200

void pdp0_for_vga_buffer_init(struct inno_gem_private *gem_priv,
	inno_dev* dev, struct role_target *role)
{
	gem_priv->vga_max_width = VGA_MAX_WIDTH;
	gem_priv->vga_max_height = VGA_MAX_HEIGHT;
	gem_priv->vga_vram_size = gem_priv->vga_max_width * gem_priv->vga_max_height * 4;

	if (hal_has_inv_mem(dev)) {
		gem_priv->vga_vram_base_dev = hal_vram_alloc(dev, role, false, gem_priv->vga_vram_size);
		gem_priv->vga_vram_base_cpu = 0;
	} else {
		gem_priv->vga_vram_base_dev = hal_vram_alloc(dev, role, true, gem_priv->vga_vram_size);
		gem_priv->vga_vram_base_cpu = dev_paddr_to_cpu_paddr(dev, gem_priv->vga_vram_base_dev);
	}

	gem_priv->vga_buffer = inno_vzalloc(gem_priv->vga_vram_size);
}

void pdp0_for_vga_buffer_fini(struct inno_gem_private *gem_priv,
	inno_dev* dev, struct role_target *role)
{
	if (gem_priv->vga_vram_base_dev != 0) {
		if (!gem_priv->vga_vram_base_cpu) {
			hal_inv_vram_free(dev, role, gem_priv->vga_vram_base_dev);
		} else {
			hal_vram_free(dev, role, gem_priv->vga_vram_base_dev);
		}
	}

	if (gem_priv->vga_buffer != NULL) {
		inno_vfree(gem_priv->vga_buffer);
	}

	gem_priv->vga_vram_base_dev = 0;
	gem_priv->vga_vram_base_cpu = 0;
	gem_priv->vga_vram_size = 0;
	gem_priv->vga_max_width = 0;
	gem_priv->vga_max_height = 0;
	gem_priv->vga_buffer = NULL;
}

void pdp0_for_vga_buffer_draw(void *buffer, int size, unsigned int width, unsigned int height)
{
	uint32_t *point = (uint32_t *)(buffer);

	inno_memset(buffer, 0, size);
	point[0] = 0xFFFFFFFF;
	point[width - 1] = 0xFFFFFFFF;
	point[(height - 1) * width] = 0xFFFFFFFF;
	point[(height * width) -1] = 0xFFFFFFFF;

	return;
}

bool pdp0_for_vga_buffer_set(struct inno_gem_private *gem_priv, inno_dev* dev,
	unsigned int target_wdith, unsigned int target_height)
{
	unsigned int width = 0;
	unsigned int height = 0;
	void *dst = NULL, *src = NULL;
	int size = 0;

	if (gem_priv->vga_buffer == NULL || gem_priv->vga_vram_base_dev == 0) {
		return false;
	}

	width = target_wdith > gem_priv->vga_max_width ? gem_priv->vga_max_width : target_wdith;
	height = target_height > gem_priv->vga_max_height ? gem_priv->vga_max_height : target_height;
	size = width * height * 4;

	/* draw buffer, if has special needs just modify this func */
	pdp0_for_vga_buffer_draw(gem_priv->vga_buffer, size, width, height);

	/* copy to gram */
	if (!gem_priv->vga_vram_base_cpu) {
		/* invisible copy */
		src = (void *)gem_priv->vga_buffer;
		dst = (void *)gem_priv->vga_vram_base_dev;
		innodpu_dma_memcpy_for_smallbar_sg(dev, &src, &dst, &size, 1, SYS2GDDR);
	} else {
		/* visible copy */
		dst = (void*)cpu_paddr_to_pcie_paddr(dev, gem_priv->vga_vram_base_cpu);
		src = gem_priv->vga_buffer;
		innodpu_dma_memcpy3(dev, &src, &dst, &size, 1, SYS2GDDR);
	}

	return true;
}

static vm_fault_t inno_gem_vm_fault(struct vm_fault *vmf)
{
	struct vm_area_struct *vma = vmf->vma;
	struct drm_gem_object *obj = vma->vm_private_data;
	struct inno_gem_object *inno_obj = to_inno_obj(obj);
	unsigned long pfn = 0;
	unsigned long offset = 0;
	unsigned long paddr = 0;
	int err = VM_FAULT_SIGBUS;
	offset = vmf->address - vma->vm_start; // vmf->pgoff must not used，because fake offset

	if (inno_obj->gem_priv->gtt_mem) {
		struct page *page = NULL;

		if (inno_obj->gtt == NULL) {
			innodpu_err(obj->dev->dev, "gtt data err\n");
			return VM_FAULT_OOM;
		}

		if (offset > inno_obj->gtt->size) {
			innodpu_err(obj->dev->dev, "vm fault offset more than gtt obj size\n");
			return VM_FAULT_OOM;
		}

		page = inno_obj->gtt->page_array[offset / PAGE_SIZE];
		pfn = page_to_pfn(page);
		err  = vmf_insert_pfn(vma, vmf->address, pfn);
	} else if (inno_obj->gem_priv->visiable) {
		paddr = inno_obj->cpu_addr + offset;
		pfn = paddr >> PAGE_SHIFT;
		err = vmf_insert_pfn(vma, vmf->address, pfn);
	} else {
		struct gem_vm_list *pvm;
		unsigned long page_nums;
		int i = 0;

#if 0
		innodpu_info(obj->dev->dev, DPU_UT_MEM, "%s(%d) vm fault visiable-%d, vm_fault dev_addr-%#llx, vm_flags-%#lx, pgoff-%#lx, inno_obj-%pK",
			inno_obj->name, obj->name, inno_obj->gem_priv->visiable, inno_obj->dev_addr + offset,
			vma->vm_flags, vmf->pgoff, inno_obj);
#endif

		pvm = kzalloc(sizeof(*pvm), GFP_KERNEL);
		if (!pvm) {
			innodpu_err(obj->dev->dev, "VM Fault alloc memory failed.\n");
			return VM_FAULT_OOM;
		}
		pvm->dev_addr  = inno_obj->dev_addr + offset;
		pvm->offset = offset;
		pvm->size = PAGE_SIZE;
		page_nums = pvm->size >>PAGE_SHIFT;
#ifdef GEM_USED_VMALLOC
		pvm->vaddr = vmalloc(pvm->size);
#else
		pvm->vaddr = kzalloc(pvm->size, GFP_KERNEL);
#endif
		if (!pvm->vaddr) {
			innodpu_err(obj->dev->dev, "vmalloc Fault alloc memory failed.\n");
			return VM_FAULT_OOM;
		}
#if 0 // TBD
		{	// Invisiable vram->visiable vram->system_mem
			//void *axi_vram_src = inno_obj->gem_priv->vram_resv_base;
			void *axi_vram_src = (void *)inno_obj->dev_addr + pvm->offset; // invisiable vram
			void *axi_vram_dst = (void *)inno_obj->gem_priv->vram_resv_base; // TBD: visiable vram
			unsigned long axi_vram_cpu = dev_paddr_to_cpu_paddr(obj->dev->dev->parent, inno_obj->gem_priv->vram_resv_base);
			void *src = (void *)cpu_paddr_to_pcie_paddr(obj->dev->dev, axi_vram_cpu);
			void *dst = (void *)pvm->vaddr;
			//innodpu_err(obj->dev->dev, "src-%pK, dst-%pK, paddr-%#lx", axi_vram_src, axi_vram_dst, axi_vram_cpu);

			// 1. 先从AXI DMA拷贝到VISIABLE VRAM，然后通过VISIABEL VRAM拷贝到SYSTEM RAM
			mutex_lock(&inno_obj->gem_priv->gem_mm_lock);
			innodpu_dma_memcpy(obj->dev->dev->parent, &axi_vram_src, &axi_vram_dst, &pvm->size, 1, GDDR2GDDR);
			mutex_unlock(&inno_obj->gem_priv->gem_mm_lock);
			innodpu_dma_memcpy3(obj->dev->dev->parent, &src, &dst, &pvm->size, 1, GDDR2SYS);
		}
#else
		{
			void *axi_vram_src = (void *)inno_obj->dev_addr + pvm->offset; // invisiable vram
			void *dst = (void *)pvm->vaddr;
#ifdef GEM_USED_VMALLOC
			innodpu_dma_memcpy_for_smallbar_sg(obj->dev->dev, &axi_vram_src, &dst, &pvm->size, 1, GDDR2SYS);
#else
			innodpu_dma_memcpy_for_smallbar(obj->dev->dev, &axi_vram_src, &dst, &pvm->size, 1, GDDR2SYS);
#endif
		}
#endif
		for (i=0; i < page_nums; ++i) {
#ifdef GEM_USED_VMALLOC
			struct page * page = vmalloc_to_page(pvm->vaddr + (i * PAGE_SIZE));
#else
			struct page * page = virt_to_page(pvm->vaddr + (i * PAGE_SIZE));
#endif
			unsigned long pfn = page_to_pfn(page);
			err  = vmf_insert_pfn(vma, vmf->address + (i * PAGE_SIZE), pfn);
		}
		INIT_LIST_HEAD(&pvm->list);
		mutex_lock(&inno_obj->vm_lock);
		list_add_tail(&pvm->list, &inno_obj->vm_head);
		mutex_unlock(&inno_obj->vm_lock);
	}

#if (DRM_VERSION >= KERNEL_VERSION(4, 20, 0))
	return err;
#endif

	switch (err) {
	case 0:
	case -EBUSY:
		return VM_FAULT_NOPAGE;
	case -ENOMEM:
		return VM_FAULT_OOM;
	default:
		return VM_FAULT_SIGBUS;
	}
}

// default is drm_gem_vm_open
static void inno_gem_vm_open(struct vm_area_struct *vma)
{
	struct drm_gem_object *obj = vma->vm_private_data;
	if (!obj || !obj->dev) {
		return;
	}

	drm_gem_object_get(obj);
}

static __attribute__ ((unused)) void inno_gem_write2file(const char * name, void  *addr, int length)
{
	struct file *fp;
	loff_t pos = 0;

	fp = filp_open(name, O_RDWR | O_CREAT | O_APPEND, 0644);
	if (IS_ERR(fp)) {
		DRM_ERROR("create file error\n");
		return;
	}
	kernel_write(fp, addr, length, &pos);

	filp_close(fp , NULL);
}

static __attribute__ ((unused))void inno_gem_read2file(const char * name, void *addr, int length, loff_t * pos)
{
	struct file *fp;

	fp = filp_open(name, O_RDONLY, 0644);
	if (IS_ERR(fp)) {
		DRM_ERROR("create file error\n");
		return;
	}
	kernel_read(fp, addr, length, pos);

	filp_close(fp , NULL);
}
static __attribute__ ((unused)) unsigned char *pimgbuf[GEM_MSYNC_PER_SIZE] = {'\0'};

static void inno_gem_wdma_sync(struct inno_gem_object *inno_obj, unsigned int cnt)
{
	struct drm_gem_object *obj = &inno_obj->base;

	if (!cnt)
		return;
#if 0
	innodpu_dma_memcpy3(obj->dev->dev->parent, inno_obj->vm_wsync_src,
		inno_obj->vm_wsync_dst, inno_obj->vm_wsync_len, cnt, SYS2GDDR);
	// GDDR2GDDR
	mutex_lock(&inno_obj->gem_priv->gem_mm_lock);
	innodpu_dma_memcpy(obj->dev->dev->parent, inno_obj->axi_vram_src,
		inno_obj->axi_vram_dst, inno_obj->vm_wsync_len, cnt, GDDR2GDDR);
	mutex_unlock(&inno_obj->gem_priv->gem_mm_lock);
#else
	{
#ifdef GEM_USED_VMALLOC
		innodpu_dma_memcpy_for_smallbar_sg(obj->dev->dev, inno_obj->vm_wsync_src,
			inno_obj->axi_vram_dst, inno_obj->vm_wsync_len, cnt, SYS2GDDR);
#else
		innodpu_dma_memcpy_for_smallbar(obj->dev->dev, inno_obj->vm_wsync_src,
			inno_obj->axi_vram_dst, inno_obj->vm_wsync_len, cnt, SYS2GDDR);
#endif
	}
#endif
	while(cnt--) {
#ifdef GEM_USED_VMALLOC
		vfree(inno_obj->pvm_arr[cnt]->vaddr);
#else
		kfree(inno_obj->pvm_arr[cnt]->vaddr);
#endif
		kfree(inno_obj->pvm_arr[cnt]);
	}
}

// default is drm_gem_vm_close
static void inno_gem_vm_close(struct vm_area_struct *vma)
{
	struct drm_gem_object *obj = vma->vm_private_data;
	struct inno_gem_object *inno_obj = to_inno_obj(obj);
	struct gem_vm_list *pvm = NULL, *tmpn = NULL;
	unsigned char cnt = 0;

	// 1. sync modified buffer
	// 2. free memory
	mutex_lock(&inno_obj->vm_lock);
	list_for_each_entry_safe(pvm, tmpn, &inno_obj->vm_head, list) {
		// Invisiable vram<-visiable vram<-system_mem
		inno_obj->axi_vram_dst[cnt] = (void *)inno_obj->dev_addr + pvm->offset; // invisiable vram
		inno_obj->vm_wsync_src[cnt] = (void *)pvm->vaddr;
		inno_obj->vm_wsync_len[cnt] = pvm->size;
		inno_obj->pvm_arr[cnt] = pvm;
		list_del(&pvm->list);
#ifdef GEM_USED_VMALLOC
		inno_gem_wdma_sync(inno_obj, cnt+1);
#else
		if (cnt == 255) { // 1M
			inno_gem_wdma_sync(inno_obj, 256);
		}
		++cnt;
#endif
	}
	mutex_unlock(&inno_obj->vm_lock);
	inno_gem_wdma_sync(inno_obj, cnt);

	// 3. put object
	drm_gem_object_put_unlocked(obj);
}

struct vm_operations_struct inno_gem_vm_ops = {
	.open = inno_gem_vm_open,
	.close = inno_gem_vm_close,
	.fault = inno_gem_vm_fault,
};







#if (DRM_VERSION > KERNEL_VERSION(5, 2, 0)) // TBD
static void inno_gem_func_free(struct drm_gem_object *obj)
{
	inno_gem_object_free_priv(obj);
	return;
}
static int inno_gem_func_open(struct drm_gem_object *obj, struct drm_file *file)
{
	return 0;
}
static void inno_gem_func_close(struct drm_gem_object *obj, struct drm_file *file)
{
		return;
}
static void inno_gem_func_print_info(struct drm_printer *p, unsigned int indent,
		const struct drm_gem_object *obj)
{
	return;
}

static struct dma_buf *inno_gem_func_export(struct drm_gem_object *obj, int flags)
{
#if (DRM_VERSION > KERNEL_VERSION(5, 4, 0))
	return inno_gem_prime_export(obj, flags);
#else
	return inno_gem_prime_export(obj->dev, obj, flags);
#endif
}

#if 0
static struct sg_table *inno_gem_func_get_sg_table(struct drm_gem_object *obj)
{

}

/*
 * called by drm_gem_mmap, used drm_gem_prime_mmap=>drm_drv->fops->mmap
 **/
static int inno_gem_func_mmap(struct drm_gem_object *obj, struct vm_area_struct *vma)
{
	return 0;
}
#endif


static struct drm_gem_object_funcs inno_gem_funcs = {
	.free = inno_gem_func_free,
	.open = inno_gem_func_open,
	.close = inno_gem_func_close,
	.print_info = inno_gem_func_print_info,
	.export = inno_gem_func_export, // replace &drm_drv.gem_prime_export
	.vm_ops = &inno_gem_vm_ops, // replace &drm_drv.gem_vm_ops
#if 0
	/* not support now */
	.get_sg_table = inno_gem_func_get_sg_table,
	.mmap = inno_gem_func_mmap,
#endif
};
#endif



#if (DRM_VERSION <= KERNEL_VERSION(5, 2, 0))
struct reservation_object * inno_gem_prime_res_obj(struct drm_gem_object *gem_obj)
{
	struct inno_gem_object *inno_obj;

	inno_obj = to_inno_obj(gem_obj);

	return inno_obj->resv;
}
#endif


/**
 * @gem_create_object: constructor for gem objects
 * called by dumb_create or pdp_gem_create
 *
 * Hook for allocating the GEM object struct, for use by the CMA and
 * SHMEM GEM helpers.
 */
static struct inno_gem_object *inno_gem_private_object_create(
		struct drm_device *drm_dev, size_t size, struct dma_resv *resv)
{
	struct inno_gem_object *inno_obj;
	inno_obj = kzalloc(sizeof(*inno_obj), GFP_KERNEL);
	if (!inno_obj)
		return NULL;
	inno_obj->dfd = 0;
	INIT_LIST_HEAD(&inno_obj->vm_head);
	mutex_init(&inno_obj->vm_lock);

#if (DRM_VERSION < KERNEL_VERSION(5, 2, 0))
	if (!resv)
		dma_resv_init(&inno_obj->_resv);
#else
	inno_obj->base.resv = resv;
#endif
	drm_gem_private_object_init(drm_dev, &inno_obj->base, size);

	return inno_obj;
}

static __maybe_unused void inno_show_dpu_vram_usage(struct drm_device *drm_dev,struct inno_gem_private *gem_priv)
{
	unsigned long long used = 0;
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;

	innodpu_err(drm_dev->dev,"------ dpu vram usage info[visiable-%d] ------\n", gem_priv->visiable);
	mutex_lock(&gem_priv->gem_mm_lock);
	list_for_each_entry_safe(inno_obj, tmp, &gem_priv->gem_mm_list, gem_mm_node) {
		used += inno_obj->base.size;
		innodpu_err(drm_dev->dev,"\t name:%-40s(%d),[0x%-10llx + 0x%-10lx]\n",
				inno_obj->name, inno_obj->base.name, inno_obj->dev_addr, inno_obj->base.size);
	}
	mutex_unlock(&gem_priv->gem_mm_lock);
	innodpu_err(drm_dev->dev,"------ dpu vram total used = %#llx ------\n",used);
}

struct drm_gem_object *inno_gem_object_create(struct drm_device *drm_dev,
		struct inno_gem_private *gem_priv, size_t size, u32 flags)
{
	struct inno_gem_object *inno_obj;
	struct drm_gem_object *gem_obj;
	int err = 0;
	struct role_target role;
	struct innodpu_drm_private *dev_priv;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return NULL;
	}

	// 1 .create drm_gem_object *gem_obj and called drm_gem_object_init
	inno_obj = inno_gem_private_object_create(drm_dev, size, NULL);
	if (!inno_obj) {
		innodpu_err(drm_dev->dev, "Failed create inno_gem_object, short of system memory\n");
		return NULL;
	}
	gem_obj = &inno_obj->base;

	inno_obj->name =  kasprintf(GFP_KERNEL, "%s-%d", current->comm ,current->pid);
	if (inno_obj->name == NULL) {
		innodpu_err(drm_dev->dev, "Failed create obj-name, short of memory\n");
		err = -ENOMEM;
		goto err_name;
	}

	// 2. create drm node
	if (gem_priv->manage_drm_mm) { // 使用drm mm方式管理显存
		struct drm_mm_node *node = &inno_obj->node;
		mutex_lock(&gem_priv->gem_mm_lock);
		err = drm_mm_insert_node(&gem_priv->vram, node, size);
		mutex_unlock(&gem_priv->gem_mm_lock);
		if (err) {
			innodpu_warn(drm_dev->dev,"drm mm short of vram memory, err-%ld(is -28) size-%ld\n", err, size);
			goto err_vram_alloc;
		}
		inno_obj->dev_addr = inno_obj->node.start;
	} else { // 使用list方式管理显存
		role.vram_role = HAL_VRAM_ROLE_DPU;
		role.id = dev_priv->role_id;
		role.sub_id = 0;
		mutex_lock(&gem_priv->gem_mm_lock);

		if (gem_priv->gtt_mem) {
			uint64_t gtt_flags = s_dpu_gtt_mem_clear ? PVRSRV_MEMALLOCFLAG_ZERO_ON_ALLOC : 0;

			inno_obj->gtt = hal_gtt_alloc(drm_dev->dev, &role, size, gtt_flags);
			if (!inno_obj->gtt) {
				mutex_unlock(&gem_priv->gem_mm_lock);
				innodpu_err(drm_dev->dev,"list short of gtt memory, size-%ld\n", size);
				goto err_vram_alloc;
			}
		} else {
			inno_obj->dev_addr = hal_vram_alloc(drm_dev->dev, &role, gem_priv->visiable, size);
			if (!inno_obj->dev_addr) {
				mutex_unlock(&gem_priv->gem_mm_lock);
				innodpu_err(drm_dev->dev,"list short of vram memory, size-%ld\n", size);
				goto err_vram_alloc;
			}
		}

		list_add_tail(&inno_obj->gem_mm_node, &gem_priv->gem_mm_list);
		mutex_unlock(&gem_priv->gem_mm_lock);
	}

#if (DRM_VERSION >= KERNEL_VERSION(5, 2, 0))
	if (!gem_obj->funcs) {
		gem_obj->funcs = &inno_gem_funcs;
	}
	inno_obj->resv = inno_obj->base.resv;
#else
	inno_obj->resv = &inno_obj->_resv;
#endif

	//drsrc = to_platform_device(dev->dev)->dev.platform_data;
	// 3. alloc device	memory
	inno_obj->gem_priv = gem_priv;
	if (gem_priv->visiable)
		inno_obj->cpu_addr = dev_paddr_to_cpu_paddr(drm_dev->dev, inno_obj->dev_addr);
	else
		inno_obj->cpu_addr = 0;

	inno_obj->dma_map_export_host_addr = false;

	// 创建 DRM_IOCTL_GEM_FLINK 提供其他模块使用
	mutex_lock(&drm_dev->object_name_lock);
	if (!gem_obj->name) {
		gem_obj->name = idr_alloc(&drm_dev->object_name_idr, gem_obj, 1, 0, GFP_KERNEL);
	}
	mutex_unlock(&drm_dev->object_name_lock);

	INIT_LIST_HEAD(&inno_obj->fb_pm_node);
	mutex_lock(&gem_priv->fb_pm_lock);
	list_add_tail(&inno_obj->fb_pm_node, &gem_priv->fb_pm_list);
	mutex_unlock(&gem_priv->fb_pm_lock);

	innodpu_info(drm_dev->dev, DPU_UT_MEM,
		"(%s)%s Visiable-%d size-%#llx, paddr-%#llx, daddr-%#llx, inno_obj-%#llx\n",
		inno_obj->base.name, inno_obj->name, gem_priv->visiable, gem_obj->size,
		inno_obj->cpu_addr, inno_obj->dev_addr, inno_obj);

	/* clear buffer to zero when not shared memory and not gtt mem */
	if (!gem_priv->manage_drm_mm && !gem_priv->gtt_mem) {
		inno_gem_vbuffer_clear(gem_priv, inno_obj, gem_obj->dev->dev);
	}

	return &inno_obj->base;

err_vram_alloc:
	kfree(inno_obj->name);
err_name:
	kfree(inno_obj);
	return NULL;
}

int inno_gem_create_priv(struct drm_file *file_priv, struct drm_device *drm_dev,
		struct inno_gem_private *gem_priv, void *data)
{
	int ret = 0;
	struct drm_pdp_gem_create *args = data;
	struct drm_gem_object *gem_obj;
	struct inno_gem_object *inno_obj;

	// 2. create gem_object and  alloc device memory
	gem_obj = inno_gem_object_create(drm_dev, gem_priv, PAGE_ALIGN(args->size), args->flags);
	if (!gem_obj) {
		innodpu_warn(drm_dev->dev, "Pid=%d Visiable-%d Input: flags-%#x size-%llu error\n",
			task_pid_nr(current), gem_priv->visiable, args->flags,args->size);
		return -ENOMEM;
	}
	inno_obj = to_inno_obj(gem_obj);

	inno_obj->flags = args->flags;
	ret = drm_gem_handle_create(file_priv, gem_obj, &args->handle);
	if (ret) {
		innodpu_warn(drm_dev->dev, "Pid=%d Visiable-%d Input: flags-%#x size-%llu create handle error\n",
			task_pid_nr(current), gem_priv->visiable, args->flags,args->size);
		goto err_handle;
	}

	drm_gem_object_put_unlocked(gem_obj);

	innodpu_info(drm_dev->dev, DPU_UT_MEM,
		"%s(%d) Visiable-%d Gttem-%d Input: size-%llu create handle. ret-%d, inno_obj-%pK flags-0x%x\n",
		inno_obj->name, gem_obj->name, gem_priv->visiable, gem_priv->gtt_mem, args->size, ret, inno_obj, inno_obj->flags);
	innodpu_info(drm_dev->dev, DPU_UT_MEM, DRM_VRAM_FMT, DRM_VRAM_ARG(inno_obj));

	return ret;

err_handle:
	drm_gem_object_put_unlocked(gem_obj);
	inno_gem_object_free_priv(gem_obj);
	return ret;
}

int inno_gem_dumb_create(struct drm_file *drm_file, struct drm_device *drm_dev,
		struct inno_gem_private *gem_priv, struct drm_mode_create_dumb *args)
{
	int ret = 0;
	__u32 handle;
	__u32 pitch;
	__u64 size;
	struct inno_gem_object *inno_obj;
	struct drm_gem_object *gem_obj;

	// 1. width align - // TBD
#ifdef HW_ALIGN
	pitch = ALIGN(DIV_ROUND_UP(args->width * args->bpp, 8), 16);
#else
	pitch = args->width * (ALIGN(args->bpp, 8) >> 3);
#endif
	size = PAGE_ALIGN(pitch * args->height);


	// 2. create gem_object and  alloc device memory
	gem_obj = inno_gem_object_create(drm_dev, gem_priv, size, 0);
	if (!gem_obj) {
		innodpu_err(drm_dev->dev, "Pid=%d Visiable-%d Input: %ux%u bpp-%u flags-%#x pitch-%u size-%llu error\n",
			task_pid_nr(current), gem_priv->visiable, args->width, args->height,
			args->bpp, args->flags, args->pitch, args->size);
		return -ENOMEM;
	}
	inno_obj = to_inno_obj(gem_obj);

	inno_obj->flags = args->flags;
	// 3. create gem handle
	ret = drm_gem_handle_create(drm_file, gem_obj, &handle);
	if (ret) {
		innodpu_err(drm_dev->dev, "Pid=%d Visiable-%d Input: %ux%u bpp-%u flags-%#x create handle error\n",
			task_pid_nr(current), gem_priv->visiable, args->width, args->height, args->bpp, args->flags);
		goto err_handle;
	}

	args->handle = handle;
	args->pitch = pitch;
	args->size = size;

	drm_gem_object_put_unlocked(gem_obj);

	innodpu_info(drm_dev->dev, DPU_UT_MEM,
		"%s(%d) Visiable-%d Input: %ux%u bpp-%u." \
		"Output: handle-%u pitch-%u size-%llu. ret-%d, daddr-%#llx inno_obj-%pK, flags-0x%x\n",
		inno_obj->name, gem_obj->name, gem_priv->visiable,
		args->width, args->height, args->bpp, args->handle,
		args->pitch, args->size, ret,inno_obj->dev_addr, inno_obj, inno_obj->flags);
	innodpu_info(drm_dev->dev, DPU_UT_MEM, DRM_VRAM_FMT, DRM_VRAM_ARG(inno_obj));

	return ret;

err_handle:
	drm_gem_object_put_unlocked(gem_obj);
	inno_gem_object_free_priv(gem_obj);
	return ret;
}


void inno_sysmem_thaw_free(struct inno_gem_private *gem_priv)
{
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;

	if (gem_priv->manage_drm_mm) {
		const struct drm_mm_node *entry = NULL;

		drm_mm_for_each_node(entry, &gem_priv->vram) {
			inno_obj = container_of(entry, struct inno_gem_object, node);
			if (inno_obj->suspend_data) {
				innodpu_info(inno_obj->base.dev->dev, DPU_UT_MEM,
					"free (%d)%s, flags %#x\n", inno_obj->base.name, inno_obj->name, inno_obj->flags);
				vfree((void *)inno_obj->suspend_data);
				inno_obj->suspend_data = NULL;
			}
		}
	} else {
		mutex_lock(&gem_priv->fb_pm_lock);
		list_for_each_entry_safe(inno_obj, tmp, &gem_priv->fb_pm_list, fb_pm_node) {
			if (inno_obj->suspend_data) {
				innodpu_info(inno_obj->base.dev->dev, DPU_UT_MEM,
					"free (%d)%s, flags %#x\n", inno_obj->base.name, inno_obj->name, inno_obj->flags);
				vfree((void *)inno_obj->suspend_data);
				inno_obj->suspend_data = NULL;
			}
		}
		mutex_unlock(&gem_priv->fb_pm_lock);
	}
}

static __maybe_unused void inno_gem_vbuffer_clear(struct inno_gem_private *gem_priv,
	struct inno_gem_object *inno_obj, struct device* pdev)
{
	int vm_offset = 0, vm_size = 0;
	void *axi_vram_src = NULL, *axi_vram_dst = NULL;

#ifndef SUPPORT_DMA_TRANSFER
	return;
#endif

	for (vm_offset = 0; vm_offset < inno_obj->base.size; vm_offset += vm_size) {
		vm_size = ((inno_obj->base.size - vm_offset) < gem_priv->vram_zero_size) ?
			(inno_obj->base.size - vm_offset) : gem_priv->vram_zero_size;
		axi_vram_src = (void *)gem_priv->vram_zero_base_dev;
		axi_vram_dst = (void *)inno_obj->dev_addr + vm_offset;

		innodpu_dma_memcpy(pdev, &axi_vram_src, &axi_vram_dst, &vm_size, 1, GDDR2GDDR);
	}
	return;
}

static __maybe_unused void inno_gem_obj_recover(struct inno_gem_object *inno_obj, struct inno_gem_private *gem_priv)
{
	struct drm_gem_object *obj;
	struct device *pdev = NULL;
	void *src = NULL;
	void *dst = NULL;
	int size = 0;

	obj = &inno_obj->base;
	pdev = obj->dev->dev;
	size = (int)(inno_obj->base.size);

#ifdef DRM_PM_GEM_VERIFY_MEMBACK
	{
		char _tmpname[100] = {'\0'};
		sprintf(_tmpname, "/tmp/recover-%d_%s", obj->name, inno_obj->name);
		inno_gem_write2file(_tmpname, inno_obj->suspend_data, inno_obj->base.size);
	}
#endif

	innodpu_info(inno_obj->base.dev->dev, DPU_UT_MEM,
		"recover (%d)%s, flags %#x\n", inno_obj->base.name, inno_obj->name, inno_obj->flags);
	if (gem_priv->visiable) {
		dst = (void*)cpu_paddr_to_pcie_paddr(pdev, inno_obj->cpu_addr);
		src = (void*)inno_obj->suspend_data;
		innodpu_dma_memcpy3(pdev, &src, &dst, &size, 1, SYS2GDDR);
	} else {
		void *axi_vram_dst = NULL;

		src = (void *)inno_obj->suspend_data;
		axi_vram_dst = (void *)inno_obj->dev_addr; // invisiable vram
		innodpu_dma_memcpy_for_smallbar_sg(obj->dev->dev, &src, &axi_vram_dst, &size, 1, SYS2GDDR);
	}
}

void inno_gem_recover(struct inno_gem_private *gem_priv)
{
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;

	if (gem_priv->manage_drm_mm) {
		const struct drm_mm_node *entry = NULL;
		drm_mm_for_each_node(entry, &gem_priv->vram) {
			inno_obj = container_of(entry, struct inno_gem_object, node);
			if (inno_obj->suspend_data) {
				if (s_dpu_gem_print) {
					inno_printk("DEBUG-RECOVER: (%d)%s inno_obj suspend data %p\n",
						inno_obj->base.name, inno_obj->name, inno_obj->suspend_data);
				}
				if (inno_is_vmalloc_addr(inno_obj->suspend_data)) {
					inno_gem_obj_recover(inno_obj, gem_priv);
					vfree((void *)inno_obj->suspend_data);
				} else {
					inno_printk("WARN: %s %d inno_obj suspend data invalid %p\n",
						__func__, __LINE__, inno_obj->suspend_data);
				}
				inno_obj->suspend_data = NULL;
			}
		}
	} else {
		mutex_lock(&gem_priv->fb_pm_lock);
		list_for_each_entry_safe(inno_obj, tmp, &gem_priv->fb_pm_list, fb_pm_node) {
			if (inno_obj->suspend_data) {
				if (s_dpu_gem_print) {
					inno_printk("DEBUG-RECOVER: (%d)%s inno_obj suspend data %p\n",
						inno_obj->base.name, inno_obj->name, inno_obj->suspend_data);
				}
				if (inno_is_vmalloc_addr(inno_obj->suspend_data)) {
					inno_gem_obj_recover(inno_obj, gem_priv);
					vfree((void *)inno_obj->suspend_data);
				} else {
					inno_printk("WARN: %s %d inno_obj suspend data invalid %p\n",
						__func__, __LINE__, inno_obj->suspend_data);
				}
				inno_obj->suspend_data = NULL;
			}
		}
		mutex_unlock(&gem_priv->fb_pm_lock);
	}
}

void inno_gem_recover_zero_vram(struct inno_gem_private *gem_priv)
{
	struct drm_device *drm_dev = NULL;
	void *src = NULL, *dst = NULL;
	int ret = 0, size = 0;

	if (!gem_priv->visiable) {
		DRM_ERROR("%s: dpu zero vram just in visable vram\n", __func__);
		return;
	}

	if (gem_priv->vram_zero_bakup_buffer == NULL) {
		DRM_ERROR("%s: invalid zero bakup buffer, not recover\n", __func__);
		return;
	}

	drm_dev = gem_priv->drm_dev;
	size = gem_priv->vram_zero_size;

	/* zero vram is only in visable vram */
	dst = (void*)cpu_paddr_to_pcie_paddr(drm_dev->dev, gem_priv->vram_zero_base_cpu);
	src = (void*)gem_priv->vram_zero_bakup_buffer;
	ret = innodpu_dma_memcpy3(drm_dev->dev, &src, &dst, &size, 1, SYS2GDDR);

	if (ret) {
		innodpu_err(drm_dev->dev, "innodpu_dma_memcpy3 recover zero-mem failed\n");
	}

	/* release buffer here */
	vfree(gem_priv->vram_zero_bakup_buffer);
	gem_priv->vram_zero_bakup_buffer = NULL;

	return;
}

static void inno_gem_obj_backup(struct inno_gem_object *inno_obj, struct inno_gem_private *gem_priv)
{
	struct drm_gem_object *obj;
	struct device *pdev = NULL;
	void *src = NULL, *dst = NULL;
	int size = 0, ret = 0;

	obj = &inno_obj->base;
	inno_obj->suspend_data = vmalloc(inno_obj->base.size);

	if (inno_obj->suspend_data == NULL) {
		DRM_ERROR("%s: vmalloc failed\n", __func__);
		return;
	}

	if (s_dpu_gem_print) {
		inno_printk("DEBUG-BACKUP: (%d)%s inno_obj suspend data %p\n",
			inno_obj->base.name, inno_obj->name, inno_obj->suspend_data);
	}

	pdev = obj->dev->dev;
	size = (int)(inno_obj->base.size);

	innodpu_info(inno_obj->base.dev->dev, DPU_UT_MEM,
		"backup (%d)%s, flags %#x\n", inno_obj->base.name, inno_obj->name, inno_obj->flags);

	if (gem_priv->visiable) {
		src = (void*)cpu_paddr_to_pcie_paddr(pdev, inno_obj->cpu_addr);
		dst = (void*)inno_obj->suspend_data;
		ret = innodpu_dma_memcpy3(pdev, &src, &dst, &size, 1, GDDR2SYS);
	} else {
		void *axi_vram_src = NULL;

		dst = (void *)inno_obj->suspend_data;
		axi_vram_src = (void *)inno_obj->dev_addr;
		ret = innodpu_dma_memcpy_for_smallbar_sg(obj->dev->dev, &axi_vram_src, &dst, &size, 1, GDDR2SYS);
	}

#ifdef DRM_PM_GEM_VERIFY_MEMBACK
	if (ret == 0) {
		char _tmpname[100] = {'\0'};
		sprintf(_tmpname, "/tmp/%d_%s.bin", obj->name, inno_obj->name);
		inno_gem_write2file(_tmpname, inno_obj->suspend_data, inno_obj->base.size);
	}
#endif
	if (ret) {
		innodpu_err(obj->dev->dev,"innodpu_dma_memcpy3 suspend failed\n");
	}

	return;
}

void inno_gem_backup_shmem(struct inno_gem_private *gem_priv)
{
	struct inno_gem_object *inno_obj = NULL;
	const struct drm_mm_node *entry = NULL;
	struct drm_device *drm_dev = gem_priv->drm_dev;
	struct innodpu_drm_private *dev_priv = NULL;

	if (drm_dev)
		dev_priv = innogpu_drm_to_display_private(drm_dev);

	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return;
	}

	drm_mm_for_each_node(entry, &gem_priv->vram) {
		inno_obj = container_of(entry, struct inno_gem_object, node);
		if (innodpu_gem_need_backup(dev_priv, inno_obj->name)) {
			inno_gem_obj_backup(inno_obj, gem_priv);
		}
	}
}

void inno_gem_backup(struct inno_gem_private *gem_priv)
{
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;
	struct drm_device *drm_dev = gem_priv->drm_dev;
	struct innodpu_drm_private *dev_priv = NULL;

	if (drm_dev)
		dev_priv = innogpu_drm_to_display_private(drm_dev);

	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return;
	}

	mutex_lock(&gem_priv->fb_pm_lock);
	list_for_each_entry_safe(inno_obj, tmp, &gem_priv->fb_pm_list, fb_pm_node) {
		if (innodpu_gem_need_backup(dev_priv, inno_obj->name)) {
			inno_gem_obj_backup(inno_obj, gem_priv);
		}
	}
	mutex_unlock(&gem_priv->fb_pm_lock);
}

void inno_gem_backup_zero_vram(struct inno_gem_private *gem_priv)
{
	struct drm_device *drm_dev = NULL;
	void *src = NULL, *dst = NULL;
	int ret = 0, size = 0;

	if (!gem_priv->visiable) {
		DRM_ERROR("%s: dpu zero vram just in visable vram\n", __func__);
		return;
	}

	gem_priv->vram_zero_bakup_buffer = vmalloc(gem_priv->vram_zero_size);
	if (gem_priv->vram_zero_bakup_buffer == NULL) {
		DRM_ERROR("%s: vmalloc failed\n", __func__);
		return;
	}

	drm_dev = gem_priv->drm_dev;
	size = gem_priv->vram_zero_size;

	src = (void*)cpu_paddr_to_pcie_paddr(drm_dev->dev, gem_priv->vram_zero_base_cpu);
	dst = gem_priv->vram_zero_bakup_buffer;
	ret = innodpu_dma_memcpy3(drm_dev->dev, &src, &dst, &size, 1, GDDR2SYS);

	if (ret) {
		innodpu_err(drm_dev->dev, "innodpu_dma_memcpy3 backup zero-mem failed\n");
	}

	return;
}

void inno_gem_suspend(struct inno_gem_private *gem_priv)
{

	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;

	mutex_lock(&gem_priv->fb_pm_lock);
	list_for_each_entry_safe(inno_obj, tmp, &gem_priv->fb_pm_list, fb_pm_node) {
		inno_gem_obj_backup(inno_obj, gem_priv);
	}
	mutex_unlock(&gem_priv->fb_pm_lock);
}


static __maybe_unused void inno_fd_clear_func(struct work_struct *work)
{
	struct inno_gem_private *gem_priv = container_of(work, struct inno_gem_private, fb_clear_work);
	void __iomem *vaddr = NULL;

	if (gem_priv->visiable) {
		vaddr = (void __iomem *)inno_ioremap_wc_portable(gem_priv->vram_zero_base_cpu, gem_priv->vram_zero_size);

		OSDeviceMemSet(vaddr, 0, gem_priv->vram_zero_size);
		iounmap(vaddr);
	}
}

/* 释放inno_gem_object资源，及占用的显存空间 */
static void inno_gem_object_free(struct inno_gem_object *inno_obj)
{
	struct inno_gem_private *gem_priv = inno_obj->gem_priv;
	struct drm_gem_object *gem_obj = &inno_obj->base;
	struct role_target role;
	struct drm_device *drm_dev = gem_obj->dev;
	struct innodpu_drm_private *dev_priv = NULL;

	if (drm_dev)
		dev_priv = innogpu_drm_to_display_private(drm_dev);

	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return;
	}

	drm_gem_free_mmap_offset(gem_obj);

#if (DRM_VERSION < KERNEL_VERSION(5, 2, 0))
	if (&inno_obj->_resv == inno_obj->resv)
		dma_resv_fini(&inno_obj->_resv);
#endif
	if (gem_obj->import_attach) {
		drm_prime_gem_destroy(gem_obj, inno_obj->sgt);
		gem_obj->import_attach = NULL;
	}

	if (gem_priv->manage_drm_mm) {
		mutex_lock(&gem_priv->gem_mm_lock);
		drm_mm_remove_node(&inno_obj->node);
		mutex_unlock(&gem_priv->gem_mm_lock);;
	} else {
		role.vram_role = HAL_VRAM_ROLE_DPU;
		role.id = dev_priv->role_id;
		role.sub_id = 0;
		mutex_lock(&gem_priv->gem_mm_lock);
		if (gem_priv->gtt_mem) {
			hal_gtt_free(gem_obj->dev->dev, &role, inno_obj->gtt);
		} else if (gem_priv->visiable) {
			hal_vram_free(gem_obj->dev->dev, &role, inno_obj->dev_addr);
		} else {
			hal_inv_vram_free(gem_obj->dev->dev, &role, inno_obj->dev_addr);
		}
		list_del(&inno_obj->gem_mm_node);
		mutex_unlock(&gem_priv->gem_mm_lock);
	}
	drm_gem_object_release(&inno_obj->base);
	kfree(inno_obj->name);
	kfree(inno_obj);
}

static void inno_fb_free_func(struct work_struct *work)
{
	struct inno_gem_private *gem_priv = container_of(work, struct inno_gem_private, fb_free_work);
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;
	struct drm_gem_object *gem_obj;
	struct device *pdev = NULL;
	void __iomem *vaddr = NULL;

	mutex_lock(&gem_priv->fb_free_lock);
	list_for_each_entry_safe(inno_obj, tmp, &gem_priv->fb_free_list, fb_free_node) {
		gem_obj = &inno_obj->base;
		pdev = gem_obj->dev->dev;
		list_del(&inno_obj->fb_free_node);
		innodpu_info(pdev, DPU_UT_MEM,
				"%s(%d) Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx, inno_obj-%pK, is_fbdev_obj:%d\n",
				inno_obj->name, gem_obj->name, inno_obj->gem_priv->visiable, inno_obj->base.size,
				inno_obj->cpu_addr, inno_obj->dev_addr, inno_obj, inno_obj->is_fbdev_obj);

		/* just clear share memory when free */
		if (gem_priv->manage_drm_mm) {
			if (inno_obj->is_fbdev_obj || (gem_priv->vram_zero_base_dev == 0)) {
				vaddr = (void __iomem *)inno_ioremap_wc_portable(inno_obj->cpu_addr, inno_obj->base.size);

				OSDeviceMemSet(vaddr, 0, inno_obj->base.size);
				iounmap(vaddr);
			} else {
				inno_gem_vbuffer_clear(gem_priv, inno_obj, pdev);
			}
		}
		inno_gem_object_free(inno_obj);
	}
	mutex_unlock(&gem_priv->fb_free_lock);
}

void inno_gem_object_free_priv(struct drm_gem_object *gem_obj)
{
	struct inno_gem_object *inno_obj = to_inno_obj(gem_obj);
	struct inno_gem_private *gem_priv = inno_obj->gem_priv;

	if (!gem_priv) { // create by inno_gem_prime_import_sg_table
		if (gem_obj->import_attach) {
			drm_prime_gem_destroy(gem_obj, inno_obj->sgt);
			kfree(inno_obj);
		}
		return;
	}
	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
		"(%d)%s Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx, inno_obj-%#llx\n",
		inno_obj->base.name, inno_obj->name, inno_obj->gem_priv->visiable, inno_obj->base.size,
		inno_obj->cpu_addr, inno_obj->dev_addr, inno_obj);

	/* delete from power manager(s3/s4) list */
	mutex_lock(&gem_priv->fb_pm_lock);
	list_del(&inno_obj->fb_pm_node);
	mutex_unlock(&gem_priv->fb_pm_lock);

	if (!gem_priv->imme_release) {
		mutex_lock(&gem_priv->fb_free_lock);
		list_add_tail(&inno_obj->fb_free_node, &gem_priv->fb_free_list);
		mutex_unlock(&gem_priv->fb_free_lock);
		schedule_work(&gem_priv->fb_free_work);
	} else {
#ifdef DRM_GEM_IMME_MEMCLEAR // used test memclear
		/* just clear share memory when free */
		if (gem_priv->manage_drm_mm) {
			inno_gem_vbuffer_clear(gem_priv, inno_obj, gem_obj->dev->dev);
		}
#endif
		inno_gem_object_free(inno_obj);
	}

	return ;
}

#if 1
static int inno_gem_prime_attach(struct dma_buf *dma_buf,
#if (DRM_VERSION < KERNEL_VERSION(4, 19, 0))
								 struct device *dev,
#endif
								 struct dma_buf_attachment *attach)
{
	struct drm_gem_object *gem_obj = dma_buf->priv;
	struct inno_gem_object *inno_obj = to_inno_obj(gem_obj);

	/* Restrict access to Rogue */
	if (WARN_ON(!gem_obj->dev->dev) ||
		((gem_obj->dev->dev != attach->dev->parent) &&
		(gem_obj->dev->dev != attach->dev)))
		return -EPERM;

	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
			"%s(%d) Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx, inno_obj-%pK\n",
			inno_obj->name, gem_obj->name, inno_obj->gem_priv->visiable, inno_obj->base.size,
			inno_obj->cpu_addr, inno_obj->dev_addr, inno_obj);

	return 0;
}

static void inno_gem_prime_detach(struct dma_buf *dma_buf, struct dma_buf_attachment *attach)
{
	struct drm_gem_object *gem_obj = dma_buf->priv;
	struct inno_gem_object *inno_obj = to_inno_obj(gem_obj);

	/* Restrict access to Rogue */
	if (WARN_ON(!gem_obj->dev->dev) ||
		((gem_obj->dev->dev != attach->dev->parent) &&
		 (gem_obj->dev->dev != attach->dev)))
		return;

	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
			"%s(%d) Visiable-%d  size-%#lx, paddr-%#llx, daddr-%#llx, inno_obj-%pK\n",
			inno_obj->name, gem_obj->name,  inno_obj->gem_priv->visiable, inno_obj->base.size,
			inno_obj->cpu_addr, inno_obj->dev_addr, inno_obj);
}

static struct sg_table *inno_gem_prime_map_dma_buf(struct dma_buf_attachment *attach,
												   enum dma_data_direction dir)
{
	struct drm_gem_object *gem_obj = attach->dmabuf->priv;
	struct inno_gem_object *inno_obj = to_inno_obj(gem_obj);
	struct inno_gem_private *gem_priv = NULL;
	struct sg_table *sgt = NULL;
	struct scatterlist *sg = NULL;
	int ret = 0, i = 0;

	if (inno_obj == NULL) {
		return NULL;
	}
	gem_priv = inno_obj->gem_priv;

	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
			"Pid=%d Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx\n",
			task_pid_nr(current), inno_obj->gem_priv->visiable, inno_obj->base.size,
			inno_obj->cpu_addr, inno_obj->dev_addr);

	sgt = kmalloc(sizeof(*sgt), GFP_KERNEL);
	if (!sgt)
		return NULL;

	if (inno_obj->dma_map_export_host_addr) {
		if (sg_alloc_table(sgt, 1, GFP_KERNEL))
			goto err_free_sgt;
		sg_dma_address(sgt->sgl) = inno_obj->cpu_addr;
		sg_dma_len(sgt->sgl) = gem_obj->size;
	} else if (!gem_priv->gtt_mem) {
		if (sg_alloc_table(sgt, 1, GFP_KERNEL))
			goto err_free_sgt;
		sg_dma_address(sgt->sgl) = inno_obj->dev_addr;
		sg_dma_len(sgt->sgl) = gem_obj->size;
	} else {
		ret = sg_alloc_table_from_pages(sgt, (struct page **)inno_obj->gtt->page_array,
			inno_obj->gtt->page_num, 0, inno_obj->gtt->size, GFP_KERNEL);
		if (ret)
			goto err_free_sgt;

		inno_for_each_sg(sgt->sgl, sg, sgt->nents, i) {
			uint64_t gtt_addr = cpu_paddr_to_gtt_paddr(gem_obj->dev->dev,
				(uint64_t)inno_dma_map_page(gem_obj->dev->dev, sg_page(sg), 0, sg->length));
			sg_dma_address(sg) = gtt_addr;
			sg_dma_len(sg) = sg->length;
		}
	}

	return sgt;

err_free_sgt:
	kfree(sgt);
	return NULL;
}

static void inno_gem_prime_unmap_dma_buf(struct dma_buf_attachment *attach,
										 struct sg_table *sgt, enum dma_data_direction dir)
{
	sg_free_table(sgt);
	kfree(sgt);
}

static void inno_gem_dmabuf_release(struct dma_buf *dma_buf)
{
	// kref_read(&gem_obj->refcount) should be 2
	drm_gem_dmabuf_release(dma_buf);
}



static int inno_gem_prime_mmap(struct dma_buf *dma_buf, struct vm_area_struct *vma)
{
	struct drm_gem_object *obj = dma_buf->priv;
	struct inno_gem_object *inno_obj = NULL;
	int err;

	inno_obj = to_inno_obj(obj);

	// drm_gem_mmap_obj will set:
	// vma->vm_ops = dev->driver->gem_vm_ops; // used gem_vm_ops to mmap
	// vma->vm_private_data = obj;
	mutex_lock(&obj->dev->struct_mutex);
	err = drm_gem_mmap_obj(obj, obj->size, vma);
#if defined(CONFIG_CPU_RK3588) || defined(CONFIG_ARCH_EMEISWORD)
	if (!inno_obj->gem_priv->gtt_mem) {
		vma->vm_page_prot = pgprot_noncached(vm_get_page_prot(vma->vm_flags));
	}
#endif
	mutex_unlock(&obj->dev->struct_mutex);

	return err;
}

#if (DRM_VERSION < KERNEL_VERSION(5, 6, 0))
static void *inno_gem_prime_kmap(struct dma_buf *dma_buf, unsigned long page_num)
{
#ifdef CONFIG_KALLSYMS
	inno_info("[%s %d]: map called by %pf\n", __func__, __LINE__, __builtin_return_address(0));
#endif

	return NULL;
}
#endif
#if (DRM_VERSION < KERNEL_VERSION(4, 19, 0))
static void *inno_gem_prime_kmap_atomic(struct dma_buf *dma_buf, unsigned long page_num)
{
	return NULL;
}

#endif


static const struct dma_buf_ops s_inno_gem_prime_dmabuf_ops = {
	.attach = inno_gem_prime_attach,
	.detach = inno_gem_prime_detach,
	.map_dma_buf = inno_gem_prime_map_dma_buf,
	.unmap_dma_buf = inno_gem_prime_unmap_dma_buf,
	.release = inno_gem_dmabuf_release,
#if (DRM_VERSION >= KERNEL_VERSION(4, 12, 0))
#if (DRM_VERSION < KERNEL_VERSION(4, 19, 0))
	.map_atomic = inno_gem_prime_kmap_atomic,
#endif
#if (DRM_VERSION < KERNEL_VERSION(5, 6, 0))
	.map = inno_gem_prime_kmap,
#endif
#else
	.kmap_atomic = inno_gem_prime_kmap_atomic,
	.kmap = inno_gem_prime_kmap,
#endif
	.mmap = inno_gem_prime_mmap,
};

struct dma_buf *inno_gem_prime_export(
#if (DRM_VERSION < KERNEL_VERSION(5, 4, 0))
										 struct drm_device *dev,
#endif
										 struct drm_gem_object *gem_obj, int flags)
{
	struct inno_gem_object *inno_obj = to_inno_obj(gem_obj);
	struct dma_buf *dbuf = NULL;
#if (DRM_VERSION >= KERNEL_VERSION(4, 1, 0))
	DEFINE_DMA_BUF_EXPORT_INFO(export_info);
#endif

#if (DRM_VERSION >= KERNEL_VERSION(4, 1, 0))
	export_info.ops = &s_inno_gem_prime_dmabuf_ops; // drm_gem_prime_dmabuf_ops
	export_info.size = gem_obj->size;
	export_info.flags = flags;
	export_info.resv = inno_obj->resv;
	export_info.priv = gem_obj;
#if 0
	// 添加一个in_fence 和 一个 out_fence
	if (inno_gem_fence_create(inno_obj, inno_obj->resv) < 0) {
		DRM_ERROR("inno_gem_fence_create Error\n");
		return NULL;
	}
#endif

#if (DRM_VERSION >= KERNEL_VERSION(5, 4, 0))
	dbuf = drm_gem_dmabuf_export(gem_obj->dev, &export_info);
#else
#if (DRM_VERSION >= KERNEL_VERSION(4, 9, 0))
	dbuf = drm_gem_dmabuf_export(dev, &export_info);
#else
	dbuf = dma_buf_export(&export_info);
#endif
#endif
#else
	dbuf =
		dma_buf_export(gem_obj, &inno_gem_prime_dmabuf_ops, gem_obj->size, flags, inno_obj->resv);
#endif

	return dbuf;
}
#endif

/* 注：不能优化成 drm_gem_prime_import, mpv播放器时在麒麟会导致引用计数增加 */
struct drm_gem_object *
inno_gem_prime_import(struct drm_device *drm_dev, struct dma_buf *dma_buf)
{
	struct drm_gem_object *obj = dma_buf->priv;

	if (obj->dev == drm_dev) {
		BUG_ON(dma_buf->ops != &s_inno_gem_prime_dmabuf_ops);

		/*
		 * The dmabuf is one of ours, so return the associated
		 * PDP GEM object, rather than create a new one.
		 */
		drm_gem_object_get(obj);
		return obj;
	}

	return drm_gem_prime_import(drm_dev, dma_buf);
}


/**
 * @gem_prime_import_sg_table:
 *
 * Optional hook used by the PRIME helper functions
 * drm_gem_prime_import() respectively drm_gem_prime_import_dev().
 */
struct drm_gem_object *
inno_gem_prime_import_sg_table(struct drm_device *drm_dev,
		struct dma_buf_attachment *attach, struct sg_table *sgt)
{
#if 1
	struct inno_gem_object *inno_obj_dup = NULL;
	struct inno_gem_object *inno_obj = NULL;
	struct dma_buf *dmabuf = attach->dmabuf;

	inno_obj = to_inno_obj(dmabuf->priv);
	if (inno_obj == NULL) {
		return NULL;
	}

	inno_obj_dup = inno_gem_private_object_create(drm_dev, attach->dmabuf->size, attach->dmabuf->resv);
	if (!inno_obj_dup) {
		return NULL;
	}

	inno_obj_dup->sgt = sgt;

	/* support gtt mem */
	if (inno_obj->gem_priv->gtt_mem) {
		inno_obj_dup->gtt = inno_obj->gtt;
		inno_obj_dup->gem_priv = inno_obj->gem_priv;
	} else if (inno_obj_dup->sgt->nents != 1) {
		/* We only expect a single entry when gem is card memory */
		kfree(inno_obj_dup);
		return NULL;
	} else {
		inno_obj_dup->dev_addr = sg_dma_address(inno_obj_dup->sgt->sgl);
	}

	// inno_obj->cpu_addr = dev_paddr_to_cpu_paddr(dev->dev->parent, inno_obj->dev_addr);
	inno_obj_dup->resv = attach->dmabuf->resv;
	return &inno_obj_dup->base;
#else
	struct inno_gem_object *inno_obj;
	struct drm_gem_object *gem_obj = NULL;
	struct dma_buf *dmabuf = attach->dmabuf;

	gem_obj = dmabuf->priv;
	inno_obj = to_inno_obj(gem_obj);
	if (gem_obj->dev == drm_dev) {
		drm_gem_object_get(gem_obj);
		innodpu_err(drm_dev->dev,
			"%s(%d)   inno_obj-%pK kref is %ld\n",
			inno_obj->name, gem_obj->name, inno_obj, kref_read(&gem_obj->refcount));

		return gem_obj;
	} else {
		innodpu_err(drm_dev->dev, "obj sg stable is error\n");
	}
#endif
	return NULL;
}

int inno_gem_share_alloc(struct drm_device *drm_dev,
		struct inno_shared_vram * pshared_vram, unsigned long mm_size)
{
	int ret = 0;
	struct innodpu_drm_private *dev_priv = NULL;
	struct role_target role;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return -EINVAL;
	}

	role.vram_role = HAL_VRAM_ROLE_DPU;
	role.id = dev_priv->role_id;
	role.sub_id = 0;

	if (hal_has_inv_mem(drm_dev->dev)) {
			pshared_vram->vram_share_devaddr
				= hal_vram_alloc(drm_dev->dev, &role, false, mm_size);
			if (!pshared_vram->vram_share_devaddr) {
				innodpu_err(drm_dev->dev, "share momory invisable alloc failed. size %#llx\n", mm_size);
				return -ENOMEM;
			}
			pshared_vram->vram_share_cpuaddr = 0;
			pshared_vram->vram_share_size = mm_size;
			pshared_vram->share_is_visiable = false;
	} else {
		pshared_vram->vram_share_devaddr = hal_vram_alloc(drm_dev->dev, &role, true, mm_size);
		if (!pshared_vram->vram_share_devaddr) {
			innodpu_err(drm_dev->dev, "share momory visiable alloc failed. size %#llx\n", mm_size);
			return -ENOMEM;
		}

		pshared_vram->vram_share_cpuaddr =
			dev_paddr_to_cpu_paddr(drm_dev->dev, pshared_vram->vram_share_devaddr);
		pshared_vram->vram_share_size = mm_size;
		pshared_vram->share_is_visiable = true;
	}

	innodpu_info(drm_dev->dev, DPU_UT_MEM,
		"%s alloc multi-user shared memory ret-%d\nvisiable-%d baddr-%#llx, paddr-%#llx, size is %#llx ret\n",
		drm_dev->driver->name, ret, pshared_vram->share_is_visiable,
		pshared_vram->vram_share_devaddr, pshared_vram->vram_share_cpuaddr);

	return ret;
}

void inno_gem_share_free(struct drm_device *drm_dev,
		struct inno_shared_vram * pshared_vram)
{
	struct innodpu_drm_private *dev_priv = NULL;
	struct role_target role;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return;
	}

	role.vram_role = HAL_VRAM_ROLE_DPU;
	role.id =  dev_priv->role_id;
	role.sub_id = 0;

	if (pshared_vram->share_is_visiable) {
		hal_vram_free(drm_dev->dev, &role, pshared_vram->vram_share_devaddr);
	} else {
		hal_inv_vram_free(drm_dev->dev, &role, pshared_vram->vram_share_devaddr);
	}
}

/*
 * flags:
 *		BIT(0): 0-GEM_MANAGE_LIST 1-GEM_MANAGE_DRM_MM
 *		BIT(1): 0-GEM_FREE_ZERO, 1-GEM_FREE_IMME // 是否立即释放？0-立即释放，不清零  1-延迟释放，会清零
 */
struct inno_gem_private *inno_gem_init_shared(struct drm_device *drm_dev,
		struct inno_shared_vram * p_shared_vram_info, unsigned char flags,unsigned long vram_zero_base_dev,
		unsigned long vram_zero_base_cpu, unsigned long vram_zero_size)
{
	struct inno_gem_private *gem_priv = NULL;

	gem_priv = kzalloc(sizeof(*gem_priv), GFP_KERNEL);
	if (!gem_priv) {
		innodpu_err(drm_dev->dev, "Gem init is failed\n");
		return NULL;
	}
	gem_priv->drm_dev = drm_dev;
	gem_priv->visiable = p_shared_vram_info->share_is_visiable;
	if (flags & GEM_FREE_IMME) {
		gem_priv->imme_release = true;
	} else {
		gem_priv->imme_release = false;
	}
	if (flags & GEM_MANAGE_DRM_MM) {
		gem_priv->manage_drm_mm = true;
	} else {
		gem_priv->manage_drm_mm = false;
	}
	gem_priv->vram_baddr = p_shared_vram_info->vram_share_devaddr;
	gem_priv->size = p_shared_vram_info->vram_share_size;
	gem_priv->vram_zero_base_cpu = vram_zero_base_cpu;
	gem_priv->vram_zero_base_dev = vram_zero_base_dev;
	gem_priv->vram_zero_size = vram_zero_size;

	/* used for s3/s4 iter to backup vram */
	mutex_init(&gem_priv->fb_pm_lock);
	INIT_LIST_HEAD(&gem_priv->fb_pm_list);

	/* used for delay free(memclear) */
	mutex_init(&gem_priv->fb_free_lock);
	INIT_LIST_HEAD(&gem_priv->fb_free_list);
	INIT_WORK(&gem_priv->fb_free_work, inno_fb_free_func);

	/* initialize a drm-mm allocator to manager vram */
	mutex_init(&gem_priv->gem_mm_lock);
	drm_mm_init(&gem_priv->vram, gem_priv->vram_baddr, gem_priv->size);

	innodpu_info(drm_dev->dev, DPU_UT_INIT, "Gem init share visiable-%d vram_baddr-%#llx, dpu-size is %#llx\n",
		gem_priv->visiable, gem_priv->vram_baddr, gem_priv->size);

	return gem_priv;
}


/*
 * flags:
 *		BIT(0): 0-GEM_MANAGE_LIST 1-GEM_MANAGE_DRM_MM
 *		BIT(1): 0-GEM_FREE_ZERO, 1-GEM_FREE_IMME free imme ？ 0-yes not clear buffer : 1-not clear buffer
 *		BIT(2): 0-NOT_GTT_MEM, 1-GTT_MEM
 */
struct inno_gem_private *inno_gem_init(struct drm_device *drm_dev, bool visiable,
		unsigned char flags)
{
	struct inno_gem_private *gem_priv = NULL;
	bool is_gtt_mem = false;
#if defined(SUPPORT_ION) && !defined(SUPPORT_GEM_ALLOC)
#else
	struct innodpu_drm_private *dev_priv = NULL;
	struct role_target role;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return NULL;
	}
#endif

	/* gtt mem */
	is_gtt_mem = (flags & BIT(2)) >> 2;
	if (is_gtt_mem) {
		if (!hal_has_gtt_mem(drm_dev->dev)) {
			innodpu_warn(drm_dev->dev, "Gem not has gtt vram\n");
			return NULL;
		}
	} else if (!hal_has_inv_mem(drm_dev->dev) && !visiable) {
		innodpu_warn(drm_dev->dev, "Gem not has invisiable vram\n");
		return NULL;
	}

	gem_priv = kzalloc(sizeof(*gem_priv), GFP_KERNEL);
	if (!gem_priv) {
		innodpu_err(drm_dev->dev, "Gem init visiable-%d is failed\n", visiable);
		return NULL;
	}
	gem_priv->drm_dev = drm_dev;
	gem_priv->visiable = visiable;
	gem_priv->imme_release = flags & BIT(1);
	gem_priv->gtt_mem = is_gtt_mem;

	/* 使用DRM_MM管理显存，还是使用LIST维护显存列表 */
	if (flags & GEM_MANAGE_DRM_MM) {
		gem_priv->manage_drm_mm = true;
	} else{
		INIT_LIST_HEAD(&gem_priv->gem_mm_list);
		gem_priv->manage_drm_mm = false;
	}

	mutex_init(&gem_priv->gem_mm_lock);
	mutex_init(&gem_priv->fb_free_lock);
	INIT_LIST_HEAD(&gem_priv->fb_free_list);
	INIT_WORK(&gem_priv->fb_free_work, inno_fb_free_func);

	mutex_init(&gem_priv->fb_pm_lock);
	INIT_LIST_HEAD(&gem_priv->fb_pm_list);

#if defined(SUPPORT_ION) && !defined(SUPPORT_GEM_ALLOC)
	innodpu_info(drm_dev->dev, DPU_UT_INIT,
		"%s has no directly allocatable memory; the memory is managed by ION\n", drm_dev->driver->name);
#else
	role.vram_role = HAL_VRAM_ROLE_DPU;
	role.id = dev_priv->role_id;
	role.sub_id = 0;
	if (visiable && !gem_priv->gtt_mem) {
		gem_priv->size = 0x6400000; /* 100Mbytes, this is a fake value */
		gem_priv->vram_zero_base_dev = hal_vram_alloc(drm_dev->dev, &role, gem_priv->visiable, GEM_ZERO_SIZE);
		gem_priv->vram_zero_base_cpu = dev_paddr_to_cpu_paddr(drm_dev->dev, gem_priv->vram_zero_base_dev);
		gem_priv->vram_zero_size = GEM_ZERO_SIZE;

		if (!dev_priv->drm_nulldisplay) {
			pdp0_for_vga_buffer_init(gem_priv, drm_dev->dev, &role);
		}

		INIT_WORK(&gem_priv->fb_clear_work, inno_fd_clear_func);
		schedule_work(&gem_priv->fb_clear_work);
	}

	if (!visiable && !gem_priv->gtt_mem) {
		gem_priv->size = 0x19000000; /* 400Mbytes, this is a fake value */
	}

	innodpu_info(drm_dev->dev, DPU_UT_MEM, "Gem init visiable-%d GttMem-%d vram_baddr-%#lx, dpu-size is %#lx\n",
		visiable, gem_priv->gtt_mem, gem_priv->vram_baddr, gem_priv->size);
#endif
	return gem_priv;
}

void inno_gem_fini(struct drm_device *drm_dev, struct inno_gem_private *gem_priv)
{
	struct innodpu_drm_private *dev_priv = NULL;
	struct inno_gem_object *inno_obj = NULL, *tmp = NULL;
	struct role_target role;

	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return;
	}

	role.vram_role = HAL_VRAM_ROLE_DPU;
	role.id = dev_priv->role_id;
	role.sub_id = 0;

	if (gem_priv->manage_drm_mm) {
		const struct drm_mm_node *entry = NULL;
		flush_work(&gem_priv->fb_free_work);

		mutex_lock(&gem_priv->gem_mm_lock);
		drm_mm_for_each_node(entry, &gem_priv->vram) {
			struct inno_gem_object *inno_obj =
					container_of(entry, struct inno_gem_object, node);
			innodpu_warn(drm_dev->dev,"Visiable-%d memory not free: vram_baddr-%#llx, size %#lx\n",
							gem_priv->visiable, inno_obj->dev_addr, inno_obj->base.size);
		}
		drm_mm_takedown(&gem_priv->vram);
		mutex_unlock(&gem_priv->gem_mm_lock);
	} else {
		if (gem_priv->visiable && gem_priv->vram_zero_base_dev && !gem_priv->gtt_mem) {
			hal_vram_free(drm_dev->dev, &role, gem_priv->vram_zero_base_dev);
			gem_priv->vram_zero_base_dev = gem_priv->vram_zero_base_cpu = 0;
			if (!dev_priv->drm_nulldisplay) {
				pdp0_for_vga_buffer_fini(gem_priv, drm_dev->dev, &role);
			}
		}
		flush_work(&gem_priv->fb_free_work);
		if (list_empty(&gem_priv->gem_mm_list))
			goto out_free;

		mutex_lock(&gem_priv->gem_mm_lock);
		list_for_each_entry_safe(inno_obj, tmp, &gem_priv->gem_mm_list, gem_mm_node) {
			if (gem_priv->gtt_mem) {
				hal_gtt_free(drm_dev->dev, &role, inno_obj->gtt);
			} else if (gem_priv->visiable) {
				hal_vram_free(drm_dev->dev, &role, inno_obj->dev_addr);
			} else {
				hal_inv_vram_free(drm_dev->dev, &role, inno_obj->dev_addr);
			}
			list_del(&inno_obj->gem_mm_node);
			kfree(inno_obj->name);
			kfree(inno_obj);
		}
		mutex_unlock(&gem_priv->gem_mm_lock);
	}

out_free:
	mutex_destroy(&gem_priv->gem_mm_lock);
	mutex_destroy(&gem_priv->fb_free_lock);
	mutex_destroy(&gem_priv->fb_pm_lock);
	kfree(gem_priv);
}

int inno_gem_vram_count(int id, void *ptr, void *data)
{
	struct drm_gem_object *obj = (struct drm_gem_object *)ptr;
	struct inno_gem_object *inno_obj = to_inno_obj(obj);
	struct drm_dpu_vram_count *m = (struct drm_dpu_vram_count *)data;

	if (inno_obj->gem_priv->visiable) {
		m->visiable_vram_usage += obj->size;
	} else {
		m->invisiable_vram_usage += obj->size;
	}

	return 0;
}

int inno_gem_object_dump_vram_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *drm_file)
{
	struct drm_dpu_vram_count *args = (struct drm_dpu_vram_count *)data;
	struct innodpu_drm_private *dev_priv = NULL;
	
	dev_priv = innogpu_drm_to_display_private(drm_dev);
	if (!dev_priv) {
		innodpu_err(drm_dev->dev, "dev priv is NULL\n");
		return -EINVAL;
	}
	innodpu_warn(drm_dev->dev, "DRM_PDP_GEM_GET legacy,use DRM_PDP_CHIP_INFO instead\n");

	args->visiable_vram_size = dev_priv->visiable_gem_priv->size;
	args->visiable_vram_usage = 0;

	mutex_lock(&drm_dev->object_name_lock);
	idr_for_each(&drm_dev->object_name_idr, inno_gem_vram_count, args);
	mutex_unlock(&drm_dev->object_name_lock);

	if (dev_priv->invisiable_gem_priv) {
		args->invisiable_vram_size = dev_priv->invisiable_gem_priv->size;
		args->invisiable_vram_usage = 0;
		mutex_lock(&drm_dev->object_name_lock);
		idr_for_each(&drm_dev->object_name_idr, inno_gem_vram_count, args);
		mutex_unlock(&drm_dev->object_name_lock);
	} else {
		args->invisiable_vram_size = 0;
		args->invisiable_vram_usage = 0;
	}
	return 0;
}

static int
inno_gem_lookup_our_object(struct drm_file *drm_file, u32 handle, struct drm_gem_object **objp)
{
	struct drm_gem_object *obj;

	obj = drm_gem_object_lookup(drm_file, handle);
	if (!obj)
		return -ENOENT;

	if (obj->import_attach) {
		/*
		 * The dmabuf associated with the object is not one of ours.
		 * Our own buffers are handled differently on import.
		 */
		drm_gem_object_put_unlocked(obj);
		return -EINVAL;
	}

	*objp = obj;
	return 0;
}

// TBD: cpu access need mmap
int inno_gem_object_cpu_prep_ioctl(struct drm_device *dev, void *data, struct drm_file *drm_file)
{
	struct drm_pdp_gem_cpu_prep *args = (struct drm_pdp_gem_cpu_prep *)data;
	struct drm_gem_object *gem_obj;
	struct inno_gem_object *inno_obj;
	bool write = ! !(args->flags & PDP_GEM_CPU_PREP_WRITE);
	bool wait = !(args->flags & PDP_GEM_CPU_PREP_NOWAIT);
	int err = 0;

	if (args->flags & ~(PDP_GEM_CPU_PREP_READ | PDP_GEM_CPU_PREP_WRITE | PDP_GEM_CPU_PREP_NOWAIT)) {
		DRM_ERROR("invalid flags: %#08x\n", args->flags);
		return -EINVAL;
	}

	mutex_lock(&dev->struct_mutex);
	err = inno_gem_lookup_our_object(drm_file, args->handle, &gem_obj);
	if (err)
		goto exit_unlock;

	inno_obj = to_inno_obj(gem_obj);
	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
			"Pid=%d Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx\n",
			task_pid_nr(current), inno_obj->gem_priv->visiable, gem_obj->size,
			inno_obj->cpu_addr, inno_obj->dev_addr);

	if (inno_obj->cpu_prep) {
		err = -EBUSY;
		goto exit_unref;
	}

	if (wait) {
		long lerr;

		lerr = dma_resv_wait_timeout_rcu(inno_obj->resv, write, true, 30 * HZ);
		if (!lerr)
			err = -EBUSY;
		else if (lerr < 0)
			err = lerr;
	} else {
		if (!dma_resv_test_signaled_rcu(inno_obj->resv, write))
			err = -EBUSY;
	}
	if (!err)
		inno_obj->cpu_prep = true;

exit_unref:
	drm_gem_object_put(gem_obj);
exit_unlock:
	mutex_unlock(&dev->struct_mutex);

	return err;
}

int inno_gem_object_cpu_fini_ioctl(struct drm_device *dev, void *data, struct drm_file *drm_file)
{
	struct drm_pdp_gem_cpu_fini *args = (struct drm_pdp_gem_cpu_fini *)data;
	struct drm_gem_object *gem_obj;
	struct inno_gem_object *inno_obj;
	int err = 0;

	if (args->pad) {
		DRM_ERROR("invalid pad (this should always be 0)\n");
		return -EINVAL;
	}

	mutex_lock(&dev->struct_mutex);
	err = inno_gem_lookup_our_object(drm_file, args->handle, &gem_obj);
	if (err)
		goto exit_unlock;

	inno_obj = to_inno_obj(gem_obj);
	innodpu_info(gem_obj->dev->dev, DPU_UT_MEM,
			"Pid=%d Visiable-%d  size-%#llx, paddr-%#llx, daddr-%#llx\n",
			task_pid_nr(current), inno_obj->gem_priv->visiable, gem_obj->size,
			inno_obj->cpu_addr, inno_obj->dev_addr);

	if (!inno_obj->cpu_prep) {
		err = -EINVAL;
		goto exit_unref;
	}

	inno_obj->cpu_prep = false;

exit_unref:
	drm_gem_object_put(gem_obj);
exit_unlock:
	mutex_unlock(&dev->struct_mutex);

	return err;
}

int inno_gem_object_inv_get_ioctl(struct drm_device *dev, void *data, struct drm_file *drm_file)
{
	struct drm_dpu_gem_inv_get *args = (struct drm_dpu_gem_inv_get *)data;
	struct drm_gem_object *gem_obj;
	struct inno_gem_object *inno_obj;
	int err = 0;

	mutex_lock(&dev->struct_mutex);
	err = inno_gem_lookup_our_object(drm_file, args->handle, &gem_obj);
	if (err)
		goto exit_unlock;

	inno_obj = to_inno_obj(gem_obj);
	if (!err)
		args->isInv = !inno_obj->gem_priv->visiable;

	drm_gem_object_put(gem_obj);
exit_unlock:
	mutex_unlock(&dev->struct_mutex);

	return err;
}

struct dma_resv *inno_gem_get_resv(struct drm_gem_object *obj)
{
	return (to_inno_obj(obj)->resv);
}

unsigned long inno_gem_get_dev_addr(struct drm_gem_object * obj)
{
	struct inno_gem_object *inno_obj = to_inno_obj(obj);

	return (unsigned long)inno_obj->dev_addr;
}

unsigned long inno_gem_get_cpu_addr(struct drm_gem_object *obj)
{
	struct inno_gem_object *inno_obj = to_inno_obj(obj);

	return (unsigned long)inno_obj->cpu_addr;
}
