#include <drm/drm_connector.h>
#include <drm/drm_encoder.h>
#include "innodpu_common.h"
#include "hal_interface.h"
#include "innodpu_compatibility.h"
#include "innodpu_print.h"
#include "innodpu_hdmi.h"
#include "innodpu_i2c.h"
#include "g1_soc_hdmi_i2c.h"
#include "g0_soc_hdmi_i2c.h"
#include "g0m_soc_hdmi_i2c.h"
#include "g1p_soc_hdmi_i2c.h"

#define G0_HDMI_I2C_NUM  3
#define G1_HDMI_I2C_NUM  3
#define G1P_HDMI_I2C_NUM 3
#define G0M_HDMI_I2C_NUM 3


struct hdmi_i2c_st g0_hdmi_i2c[G0_HDMI_I2C_NUM];
struct hdmi_i2c_st g0m_hdmi_i2c[G0M_HDMI_I2C_NUM];
struct hdmi_i2c_st g1_hdmi_i2c[G1_HDMI_I2C_NUM];
struct hdmi_i2c_st g1p_hdmi_i2c[G1P_HDMI_I2C_NUM];


int g1_soc_hdmi_i2c_init(struct hdmi_chip_t *chip)
{
	struct i2c_adapter *adap;
	struct i2c_algo_bit_data *bit_data;

	if((chip->id > G1_HDMI_I2C_NUM))
		return -1;

	memset(&g1_hdmi_i2c[chip->id], 0, sizeof(struct hdmi_i2c_st));
	chip->i2c = &g1_hdmi_i2c[chip->id];
	inno_mutex_init(&chip->i2c->i2c_mlock);

	adap = &chip->adapter;
	bit_data = &chip->bit_data;
	bit_data->setsda = g1_soc_hdmi_i2c_gpio_setsda_val;
	bit_data->setscl = g1_soc_hdmi_i2c_gpio_setscl_val;
	bit_data->getscl = g1_soc_hdmi_i2c_gpio_getscl;
	bit_data->getsda = g1_soc_hdmi_i2c_gpio_getsda;
	bit_data->pre_xfer = g1_soc_hdmi_i2c_gpio_pre_xfer;
	bit_data->post_xfer = g1_soc_hdmi_i2c_gpio_post_xfer;
	bit_data->udelay = 5; /* 100 kHz */
	bit_data->timeout = HZ / 10; /* 100 ms */
	bit_data->data = chip;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "%s-i2c", chip->name);
	adap->algo_data = bit_data;
	adap->class = I2C_CLASS_HWMON | I2C_CLASS_SPD;
	adap->dev.parent = chip->dev;
	adap->dev.of_node = inno_get_dev_ofnode(chip->dev);
	adap->nr = -1;

	return i2c_bit_add_numbered_bus(adap);
}

void g1_soc_hdmi_i2c_fini(struct hdmi_chip_t *chip)
{
	inno_mutex_lock(&chip->i2c->i2c_mlock);
	i2c_del_adapter(&chip->adapter);
	inno_mutex_unlock(&chip->i2c->i2c_mlock);

	chip->i2c->support_i2c = false;
	return;
}

int g0_soc_hdmi_i2c_init(struct hdmi_chip_t *chip)
{
	struct i2c_adapter *adap;
	struct i2c_algo_bit_data *bit_data;

	if (chip->id > G0_HDMI_I2C_NUM) {
		return -1;
	}

	memset(&g0_hdmi_i2c[chip->id], 0, sizeof(struct hdmi_i2c_st));
	chip->i2c = &g0_hdmi_i2c[chip->id];
	inno_mutex_init(&chip->i2c->i2c_mlock);

	adap = &chip->adapter;
	bit_data = &chip->bit_data;
	bit_data->setsda = g0_soc_hdmi_i2c_gpio_setsda_val;
	bit_data->setscl = g0_soc_hdmi_i2c_gpio_setscl_val;
	bit_data->getscl = g0_soc_hdmi_i2c_gpio_getscl;
	bit_data->getsda = g0_soc_hdmi_i2c_gpio_getsda;
	bit_data->pre_xfer = g0_soc_hdmi_i2c_gpio_pre_xfer;
	bit_data->post_xfer = g0_soc_hdmi_i2c_gpio_post_xfer;
	bit_data->udelay = 5; /* 100 kHz */
	bit_data->timeout = HZ / 10; /* 100 ms */
	bit_data->data = chip;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "%s-i2c", chip->name);
	adap->algo_data = bit_data;
	adap->class = I2C_CLASS_HWMON | I2C_CLASS_SPD;
	adap->dev.parent = chip->dev;
	adap->dev.of_node = inno_get_dev_ofnode(chip->dev);
	adap->nr = -1;

	if(!i2c_bit_add_numbered_bus(adap)) {
		chip->i2c->support_i2c = true;
		return 0;
	}
	return -1;
}

void g0_soc_hdmi_i2c_fini(struct hdmi_chip_t *chip)
{
	inno_mutex_lock(&chip->i2c->i2c_mlock);
	i2c_del_adapter(&chip->adapter);
	inno_mutex_unlock(&chip->i2c->i2c_mlock);

	chip->i2c->support_i2c = false;
	return;
}

int g1p_soc_hdmi_i2c_init(struct hdmi_chip_t *chip)
{
	struct i2c_adapter *adap;
	struct i2c_algo_bit_data *bit_data;

	if(chip->id > G1P_HDMI_I2C_NUM)
		return -1;

	memset(&g1p_hdmi_i2c[chip->id], 0, sizeof(struct hdmi_i2c_st));
	chip->i2c = &g1p_hdmi_i2c[chip->id];
	inno_mutex_init(&chip->i2c->i2c_mlock);

	adap = &chip->adapter;
	bit_data = &chip->bit_data;
	bit_data->setsda = g1p_soc_hdmi_i2c_gpio_setsda_val;
	bit_data->setscl = g1p_soc_hdmi_i2c_gpio_setscl_val;
	bit_data->getscl = g1p_soc_hdmi_i2c_gpio_getscl;
	bit_data->getsda = g1p_soc_hdmi_i2c_gpio_getsda;
	bit_data->pre_xfer = g1p_soc_hdmi_i2c_gpio_pre_xfer;
	bit_data->post_xfer = g1p_soc_hdmi_i2c_gpio_post_xfer;
	bit_data->udelay = 5; /* 100 kHz */
	bit_data->timeout = HZ / 10; /* 100 ms */
	bit_data->data = chip;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "%s-i2c", chip->name);
	adap->algo_data = bit_data;
	adap->class = I2C_CLASS_HWMON | I2C_CLASS_SPD;
	adap->dev.parent = chip->dev;
	adap->dev.of_node = inno_get_dev_ofnode(chip->dev);
	adap->nr = -1;

	if(!i2c_bit_add_numbered_bus(adap)) {
		chip->i2c->support_i2c = true;
		return 0;
	}
	return -1;
}

void g1p_soc_hdmi_i2c_fini(struct hdmi_chip_t *chip)
{
	inno_mutex_lock(&chip->i2c->i2c_mlock);
	i2c_del_adapter(&chip->adapter);
	inno_mutex_unlock(&chip->i2c->i2c_mlock);

	chip->i2c->support_i2c = false;
	return;
}

int g0m_soc_hdmi_i2c_init(struct hdmi_chip_t *chip)
{
	struct i2c_adapter *adap;
	struct i2c_algo_bit_data *bit_data;

	if (chip->id > G0M_HDMI_I2C_NUM) {
		return -1;
	}

	memset(&g0m_hdmi_i2c[chip->id], 0, sizeof(struct hdmi_i2c_st));
	chip->i2c = &g0m_hdmi_i2c[chip->id];
	inno_mutex_init(&chip->i2c->i2c_mlock);

	adap = &chip->adapter;
	bit_data = &chip->bit_data;
	bit_data->setsda = g0m_soc_hdmi_i2c_gpio_setsda_val;
	bit_data->setscl = g0m_soc_hdmi_i2c_gpio_setscl_val;
	bit_data->getscl = g0m_soc_hdmi_i2c_gpio_getscl;
	bit_data->getsda = g0m_soc_hdmi_i2c_gpio_getsda;
	bit_data->pre_xfer = g0m_soc_hdmi_i2c_gpio_pre_xfer;
	bit_data->post_xfer = g0m_soc_hdmi_i2c_gpio_post_xfer;
	bit_data->udelay = 5; /* 100 kHz */
	bit_data->timeout = HZ / 10; /* 100 ms */
	bit_data->data = chip;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "%s-i2c", chip->name);
	adap->algo_data = bit_data;
	adap->class = I2C_CLASS_HWMON | I2C_CLASS_SPD;
	adap->dev.parent = chip->dev;
	adap->dev.of_node = inno_get_dev_ofnode(chip->dev);
	adap->nr = -1;

	if(!i2c_bit_add_numbered_bus(adap)) {
		chip->i2c->support_i2c = true;
		return 0;
	}
	return -1;
}

void g0m_soc_hdmi_i2c_fini(struct hdmi_chip_t *chip)
{
	inno_mutex_lock(&chip->i2c->i2c_mlock);
	i2c_del_adapter(&chip->adapter);
	inno_mutex_unlock(&chip->i2c->i2c_mlock);

	chip->i2c->support_i2c = false;
	return;
}

#ifdef HWI2C
static u32 g0m_soc_hdmi_hwi2c_func(struct i2c_adapter *adapter)
{
	return I2C_FUNC_I2C | I2C_FUNC_SMBUS_EMUL;
}

extern int g0m_soc_hdmi_hwi2c_xfer(struct i2c_adapter *adap,
		struct i2c_msg *msgs, int num);

static const struct i2c_algorithm g0m_algorithm = {
	.master_xfer = g0m_soc_hdmi_hwi2c_xfer,
	.functionality = g0m_soc_hdmi_hwi2c_func,
};

int g0m_soc_hdmi_i2c_init(struct hdmi_chip_t *chip)
{
	struct i2c_adapter *adap;

	if (chip->id > G0M_HDMI_I2C_NUM) {
		return -1;
	}

	memset(&g0m_hdmi_i2c[chip->id], 0, sizeof(struct hdmi_i2c_st));
	chip->i2c = &g0m_hdmi_i2c[chip->id];
	inno_mutex_init(&chip->i2c->i2c_mlock);

	init_completion(&chip->comp);

	adap = &chip->adapter;

	adap->owner = THIS_MODULE;
	snprintf(adap->name, sizeof(adap->name), "%s-hwi2c", chip->name);
	adap->algo = &g0m_algorithm;
	adap->class = I2C_CLASS_DDC;
	adap->dev.parent = chip->dev;
	adap->dev.of_node = inno_get_dev_ofnode(chip->dev);
	adap->nr = -1;

	if(!i2c_add_adapter(adap)) {
		chip->i2c->support_i2c = true;
		return 0;
	}

	return -1;
}

void g0m_soc_hdmi_i2c_fini(struct hdmi_chip_t *chip)
{
	inno_mutex_lock(&chip->i2c->i2c_mlock);
	i2c_del_adapter(&chip->adapter);
	inno_mutex_unlock(&chip->i2c->i2c_mlock);

	chip->i2c->support_i2c = false;
	return;
}
#endif

