/*************************************************************************/ /*!
@File			innogpu_drm.c
@Title
@Copyright		Copyright (c) Innosilicon Technology Ltd. All Rights Reserved
@Description
@License		Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#include "inno_mm.h"
#include "innogpu_drm.h"
#include "pvr_sync_ioctl_drm.h"
#include "innodpu_drm_drv.h"
#include "innodpu_drm_pm.h"
#include "innodpu_common.h"
#include "innodpu_connector.h"
#include "module_common.h"
#include <linux/platform_device.h>
#include "pvr_drm.h"
#include "pvrsrv.h"
extern struct platform_driver g_innogpu_pvrsrvkm_driver;
#ifdef CONFIG_DRM_INNO_AUDIO
extern struct platform_driver g_innoaudio_playback_driver;
#endif

extern void innodpu_poll_logo_execute(struct drm_device *drm_dev,
		struct innodpu_drm_private *dev_priv, int nr);

// sync from function innodpu_gem_object_create_ioctl
static int innogpu_drm_gem_create_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_mmap_ioctl
static int innogpu_drm_gem_mmap_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_cpu_prep_ioctl
static int innogpu_drm_gem_cpu_prep_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_cpu_fini_ioctl
static int innogpu_drm_gem_cpu_fini_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_dump_vram_ioctl
static int innogpu_drm_gem_dump_vram_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_inv_get_ioctl
static int innogpu_drm_gem_inv_get_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function inno_gem_object_plane_fd_ioctl
static int innogpu_drm_gem_plane_fd_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function innodpu_drm_chipinfo_ioctl
static int innogpu_drm_chipinfo_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file);

// sync from function innodpu_component_bind
static int innogpu_drm_component_bind(struct device *dev);

// sync from function innodpu_component_unbind
static void innogpu_drm_component_unbind(struct device *dev);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
static int innogpu_drm_debugfs_init(struct drm_minor *minor);
#else
static void innogpu_drm_debugfs_init(struct drm_minor *minor);
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0))
void innogpu_drm_debugfs_cleanup(struct drm_minor *minor)
#endif
/*
 * Casue pvr/auio/dpu merge as pvr.ko,and dpu depend on pmbus,
 * so when make split, insmod ko should follow this order:
 * innogpu.ko innodma.ko innopmbus.ko innosrv.ko;
 */

unsigned int s_dpu_driver_group = 0x0;
module_param(s_dpu_driver_group, uint, 0444);
MODULE_PARM_DESC(s_dpu_driver_group, "dpu sub drivers group select(default: 0)");

#define SUB_DRIVER_GROUP_NUMBER   (8U)
static struct platform_driver *s_innogpu_drm_sub_drivers[][SUB_DRIVER_GROUP_NUMBER] = {
	{
		&g_innogpu_pvrsrvkm_driver,
#ifdef CONFIG_DRM_INNO_AUDIO
		&g_innoaudio_playback_driver,
#endif
		&g_innogpu_hdmi_driver,
		&g_innogpu_dp_driver,
		&g_innogpu_vga_driver,
		//&g_innogpu_lvds_driver,
		&g_innogpu_vkms_driver,
		&g_innogpu_dpu_driver,
	},

	{
		&g_innogpu_pvrsrvkm_driver,
		&g_innogpu_hdmi_driver,
		&g_innogpu_dp_driver,
		&g_innogpu_vga_driver,
		//&g_innogpu_lvds_driver,
		&g_innogpu_vkms_driver,
		&g_innogpu_dpu_driver,
	},
};

static const struct component_master_ops s_innogu_drm_component_ops = {
	.bind = innogpu_drm_component_bind,
	.unbind = innogpu_drm_component_unbind,
};

int drm_pvr_srvkm_init(struct drm_device *dev, void *arg, struct drm_file *psDRMFile)
{
	struct drm_pvr_srvkm_init_data *data = arg;
	struct pvr_drm_private *priv = innogpu_drm_to_pvr_private(dev);
	int iErr = 0;

	switch (data->init_module)
	{
		case PVR_SRVKM_SYNC_INIT:
		{
			iErr = PVRSRVDeviceSyncOpen(priv->dev_node, psDRMFile);
			break;
		}
		case PVR_SRVKM_SERVICES_INIT:
		{
			iErr = PVRSRVDeviceServicesOpen(priv->dev_node, psDRMFile);
			break;
		}
		default:
		{
			PVR_DPF((PVR_DBG_ERROR, "%s: invalid init_module (%d)",
			        __func__, data->init_module));
			iErr = -EINVAL;
		}
	}

	return iErr;
}

/*
 * private ioctls:
 * merge pvr_drm_ioctls and s_innodpu_ioctls
 */
static const struct drm_ioctl_desc s_innogpu_drm_ioctls[] = {
	DRM_IOCTL_DEF_DRV(PVR_SRVKM_CMD, PVRSRV_BridgeDispatchKM,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
	DRM_IOCTL_DEF_DRV(PVR_SRVKM_INIT, drm_pvr_srvkm_init,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
#if defined(SUPPORT_NATIVE_FENCE_SYNC) && !defined(USE_PVRSYNC_DEVNODE)
	DRM_IOCTL_DEF_DRV(PVR_SYNC_RENAME_CMD, pvr_sync_rename_ioctl,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
	DRM_IOCTL_DEF_DRV(PVR_SYNC_FORCE_SW_ONLY_CMD, pvr_sync_force_sw_only_ioctl,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
	DRM_IOCTL_DEF_DRV(PVR_SW_SYNC_CREATE_FENCE_CMD, pvr_sw_sync_create_fence_ioctl,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
	DRM_IOCTL_DEF_DRV(PVR_SW_SYNC_INC_CMD, pvr_sw_sync_inc_ioctl,
			DRM_RENDER_ALLOW | DRM_UNLOCKED | DRM_RENDER_ALLOW),
#endif

	DRM_IOCTL_DEF_DRV(PDP_GEM_CREATE, innogpu_drm_gem_create_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_MMAP, innogpu_drm_gem_mmap_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_CPU_PREP, innogpu_drm_gem_cpu_prep_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_CPU_FINI, innogpu_drm_gem_cpu_fini_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_GET, innogpu_drm_gem_dump_vram_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_INV_GET, innogpu_drm_gem_inv_get_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_GEM_BASE_FD, innogpu_drm_gem_plane_fd_ioctl,
			INNODPU_IOCTL_FLAGS),
	DRM_IOCTL_DEF_DRV(PDP_CHIP_INFO, innogpu_drm_chipinfo_ioctl,
			INNODPU_IOCTL_FLAGS),
};

static long innogpu_drm_ioctl_check(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	struct drm_file *drm_file = filp->private_data;
	struct drm_device *drm_dev = drm_file->minor->dev;
	struct innodpu_drm_private *display_priv = NULL;

	display_priv = innogpu_drm_to_display_private(drm_dev);

	if (hal_gpuchip_is_ovheat(drm_dev->dev)) {
		innodpu_info(drm_dev->dev, DPU_UT_CORE, "ovheat, no timing settings\n");
		return -EFAULT;
	}

	if (display_priv && drm_file->minor && (drm_file->minor->type == DRM_MINOR_PRIMARY))
		innodpu_poll_logo_execute(drm_dev, display_priv, DRM_IOCTL_NR(cmd));

	return 0;
}

static long innogpu_drm_ioctl(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	long ret = 0;

	ret = innogpu_drm_ioctl_check(filp, cmd, arg);
	if (ret >= 0)
		return drm_ioctl(filp, cmd, arg);

	return ret;
}

#ifdef CONFIG_COMPAT
static long innogpu_drm_compat_ioctl(struct file *filp,
		unsigned int cmd, unsigned long arg)
{
	int ret = 0;

	ret = innogpu_drm_ioctl_check(filp, cmd, arg);
	if (ret >= 0)
		return drm_compat_ioctl(filp, cmd, arg);

	return ret;
}
#endif

/*
 * 
 * Because a system tries to open DRM before the driver registration is complete.
 * it causes the sytem to crash.
 */
static int innogpu_drm_check_bind_finish(struct drm_device *drm_dev)
{
	innodpu_bind_atomic(drm_dev, true);
	innodpu_bind_atomic(drm_dev, false);
	return 0;
}

static int innogpu_pvr_drm_open(struct drm_device *drm_dev, struct drm_file *dfile)
{
#if (PVRSRV_DEVICE_INIT_MODE != PVRSRV_LINUX_DEV_INIT_ON_CONNECT) || defined(__INNO_CONTAINER__)
	struct pvr_drm_private *priv = innogpu_drm_to_pvr_private(drm_dev);
	int err = 0;
#endif

	if (!try_module_get(THIS_MODULE)) {
		DRM_ERROR("failed to get module reference\n");
		return -ENOENT;
	}

#if (PVRSRV_DEVICE_INIT_MODE != PVRSRV_LINUX_DEV_INIT_ON_CONNECT) || defined(__INNO_CONTAINER__)
	err = PVRSRVDeviceServicesOpen(priv->dev_node, dfile);
	if (err)
		module_put(THIS_MODULE);

	return err;
#else
	return 0;
#endif
}

static void innogpu_pvr_drm_release(struct drm_device *drm_dev, struct drm_file *dfile)
{
	struct pvr_drm_private *priv = drm_dev->dev_private;

	PVRSRVDeviceRelease(priv->dev_node, dfile);

	module_put(THIS_MODULE);
}

static int innogpu_drm_file_open (struct drm_device *drm_dev,
		struct drm_file *drm_file)
{
	int ret = 0;

	/*
	 * Xorg starts opening the DRM device before fbdev initialization is complete,
	 * leading to kernel panic.
	 */
	ret = innogpu_drm_check_bind_finish(drm_dev);
	ret = innogpu_pvr_drm_open(drm_dev, drm_file);

	return ret;
}

static void innogpu_drm_file_release (struct drm_device *drm_dev,
		struct drm_file *drm_file)
{
	innogpu_pvr_drm_release(drm_dev,drm_file);

	return;
}

static bool innogpu_pvrsrv_mmap_judge(u64 offset)
{
	/*
	 * Use the formula: vm_pgoff * page_size / 4k < MMAP_THRESHOLD;
	 * to judge this mmap gem_obj belong to dpu or pvr.
	 * But this judge formula is based on empirical values (test on x86/LoonArch/Phytium),
	 * which may not be scientific.
	 */
	if (((offset * inno_page_size) >> 12) < MMAP_THRESHOLD) {
		return true;
	}

	return false;
}

static __maybe_unused bool innogpu_judge_gtt_mem(struct drm_gem_object *obj)
{
	struct inno_gem_object *inno_obj = NULL;

	if (obj == NULL) {
		DRM_ERROR("gem_obj pointer invalid\n");
		return false;
	}

	inno_obj = to_inno_obj(obj);
	if (inno_obj->gem_priv->gtt_mem) {
		return true;
	}

	return false;
}

/*
 * private mmap:
 * merge PVRSRV_MMap and inno_drm_gem_mmap
 */
static int innogpu_drm_mmap(struct file *filp,
		struct vm_area_struct *vma)
{
	int ret = 0;

	if (innogpu_pvrsrv_mmap_judge(vma->vm_pgoff)) {
		ret = PVRSRV_MMap(filp, vma);
	} else {
		ret = drm_gem_mmap(filp, vma);
#if defined(CONFIG_CPU_RK3588) || defined(CONFIG_ARCH_EMEISWORD)
		/* vma->vm_private_data will point to gem_obj after excute drm_gem_mmap_obj */
		if (!innogpu_judge_gtt_mem(vma->vm_private_data)) {
			vma->vm_page_prot = pgprot_noncached(vm_get_page_prot(vma->vm_flags));
		}
#endif
	}

	return ret;
}

const struct file_operations s_innogpu_drm_fops = {\
	.owner = THIS_MODULE,
	.open = drm_open,
	.release = drm_release,
	/*.unlocked_ioctl = drm_ioctl,*/
	/*.compat_ioctl = drm_compat_ioctl,*/
	.unlocked_ioctl = innogpu_drm_ioctl,
#ifdef CONFIG_COMPAT
	.compat_ioctl = innogpu_drm_compat_ioctl,
#endif
	.poll = drm_poll,
	.read = drm_read,
	.llseek = noop_llseek,
	.mmap = innogpu_drm_mmap,
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 12, 0))
	.fasync = drm_fasync,
#endif
};

static int innogpu_drm_gem_dumb_map_offset(struct drm_file *file, struct drm_device *dev,
	u32 handle, u64 *offset)
{
	int ret = 0;

	ret = drm_gem_dumb_map_offset(file, dev, handle, offset);

	if (ret == 0) {
		if (innogpu_pvrsrv_mmap_judge(*offset >> PAGE_SHIFT)) {
			DRM_ERROR("map dumb offset invalid[%lld], will cause mmap err\n", *offset);
		}
	}

	return ret;
}

/*
 * drm management:
 * merge pvr_drm_generic_driver and s_innodpu_drm_driver
 */
struct drm_driver g_innogpu_drm_driver = {
	// drm management
	.name = INNOGPU_DRM_DRIVER_NAME,
	.desc = INNOGPU_DRM_DRIVER_DESC,
	.date = INNOGPU_DRM_DRIVER_DATE,
	.major = INNOGPU_DRM_DRIVER_MAJ,
	.minor = INNOGPU_DRM_DRIVER_MIN,
	.patchlevel = INNOGPU_DRM_DRIVER_BUILD,
	.driver_features = DRIVER_GEM | DRIVER_MODESET | DRIVER_ATOMIC | DRIVER_PRIME | DRIVER_RENDER,

	.fops = &s_innogpu_drm_fops,
	.ioctls = s_innogpu_drm_ioctls,
	.num_ioctls = ARRAY_SIZE(s_innogpu_drm_ioctls),
	.open = innogpu_drm_file_open,
	.postclose = innogpu_drm_file_release,
	.debugfs_init = innogpu_drm_debugfs_init, // TBD innodpu_debugfs_init

	#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0))
	.debugfs_cleanup = innogpu_drm_debugfs_cleanup,
	#endif

	// vram management
	.gem_create_object = NULL,
	.prime_handle_to_fd = drm_gem_prime_handle_to_fd, // drm_gem_prime_handle_to_fd()
	.prime_fd_to_handle = drm_gem_prime_fd_to_handle, // drm_gem_prime_fd_to_handle()
	.dumb_create = innodpu_gem_dumb_create,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 12, 0))
//	.dumb_destroy = drm_gem_dumb_destroy,
#else
	.dumb_destroy = drm_gem_dumb_destroy,
#endif
	.dumb_map_offset = innogpu_drm_gem_dumb_map_offset,
	.gem_prime_import = inno_gem_prime_import,
	.gem_prime_import_sg_table = inno_gem_prime_import_sg_table, // called by gem_prime_import()
	// TBD: gem_prime_mmap current not used, because s_inno_gem_prime_dmabuf_ops.mmap not drm_gem_prime_dmabuf_ops
	.gem_prime_mmap = drm_gem_prime_mmap, // called by default drm_gem_prime_dmabuf_ops.drm_gem_dmabuf_mmap

	// used gem_object.func when version > 5.2.0
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(5, 2, 0))
	/*
	 * commit 805dc614d58a8fb069ed079005e591247df85246
	 * Refs: v5.2-rc5-864-g805dc614d58a
	 * gem_free_object/gem_free_object_unlocked => &drm_gem_object_funcs.free instead.
	 * gem_open_object => &drm_gem_object_funcs.open instead.
	 * gem_close_object => &drm_gem_object_funcs.close instead.
	 * gem_print_info => &drm_gem_object_funcs.print_info instead.
	 * gem_prime_get_sg_table =>&drm_gem_object_funcs.get_sg_table instead.
	 * gem_vm_ops => &drm_gem_object_funcs.vm_ops instead.
	 * gem_prime_export=>&drm_gem_object_funcs.export
	 */
	.gem_prime_get_sg_table = NULL, // called by drm_gem_prime_dmabuf_ops->map_dma_buf, TBD
	.gem_prime_res_obj = inno_gem_prime_res_obj, // drm_gem_prime_export called to get resv


#if (LINUX_VERSION_CODE <= KERNEL_VERSION(5, 7, 0))
	/*
	 * commit ad0f449bebc79b01583c711684fefcdc9620320a
	 * Refs: v5.7-rc1-615-gad0f449bebc7
	 * - gem_print_info
	 * commit 45d58b40292b16ab847497dcd299e315a2ad7956
	 * Refs: v4.14-rc3-611-g45d58b40292b
	 * + gem_print_info
	 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 15, 0))
	/* .gem_print_info = NULL, */
#endif
	/*
	 * commit 1a9458aeb8eb48bfa5f9b3e7682bddc28fd0b85e
	 * Refs: v5.7-rc1-489-g1a9458aeb8eb
	*/
	.gem_free_object = inno_gem_object_free_priv,
#endif
	/*
	 * commit d693def4fd1c23f1ca5aed1afb9993b3a2069ad2
	 * Refs: v5.9-rc5-1077-gd693def4fd1c
	*/
	.gem_open_object = NULL,
	.gem_close_object = NULL,
	.gem_prime_pin = NULL, // &drm_gem_object_funcs.pin, drm_gem_prime_dmabuf_ops.attach
	.gem_prime_unpin = NULL, // &drm_gem_object_funcs.unpin, drm_gem_prime_dmabuf_ops.detach
	.gem_prime_vunmap = NULL, // &drm_gem_object_funcs.vmap, drm_gem_prime_dmabuf_ops.vunmap
	.gem_prime_vmap = NULL, // &drm_gem_object_funcs.vunmap, drm_gem_prime_dmabuf_ops.vmap
	.gem_vm_ops = &inno_gem_vm_ops,

	/*
	 * commit e4fa8457b2197118538a1400b75c898f9faaf164
	 * Refs: v5.2-rc5-870-ge4fa8457b219
	 * -struct dma_buf * (*gem_prime_export)(struct drm_device *dev, struct drm_gem_object *obj, int flags);
	 * +struct dma_buf * (*gem_prime_export)(struct drm_gem_object *obj, int flags);
	 */
	//.gem_prime_export = drm_gem_prime_export, // gem_ops is drm_gem_prime_dmabuf_ops
	.gem_prime_export = inno_gem_prime_export,

#endif // end if (LINUX_VERSION_CODE <= KERNEL_VERSION(5, 2, 0))
};

#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
static int innogpu_drm_debugfs_init(struct drm_minor *minor)
#else
static void innogpu_drm_debugfs_init(struct drm_minor *minor)
#endif
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 8, 0))
	int ret = 0;
	ret = innodpu_debugfs_init(minor);
	return ret;
#else
	innodpu_debugfs_init(minor);
	return;
#endif
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0))
void innogpu_drm_debugfs_cleanup(struct drm_minor *minor)
{
	innodpu_debugfs_cleanup(minor);
}
#endif


// sync innodpu_gem_object_create_ioctl
static int innogpu_drm_gem_create_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return innodpu_gem_object_create_ioctl(drm_dev, data, file);
}

// sync inno_gem_object_mmap_ioctl
static int innogpu_drm_gem_mmap_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	struct drm_pdp_gem_mmap *args = (struct drm_pdp_gem_mmap *)data;

	return innogpu_drm_gem_dumb_map_offset(file, drm_dev, args->handle, &args->offset);
}

// sync inno_gem_object_cpu_prep_ioctl
static int innogpu_drm_gem_cpu_prep_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return inno_gem_object_cpu_prep_ioctl(drm_dev, data, file);
}

// sync inno_gem_object_cpu_fini_ioctl
static int innogpu_drm_gem_cpu_fini_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return inno_gem_object_cpu_fini_ioctl(drm_dev, data, file);
}

// sync inno_gem_object_dump_vram_ioctl
static int innogpu_drm_gem_dump_vram_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return inno_gem_object_dump_vram_ioctl(drm_dev, data, file);
}

// sync inno_gem_object_inv_get_ioctl
static int innogpu_drm_gem_inv_get_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return inno_gem_object_inv_get_ioctl(drm_dev, data, file);
}

// sync inno_gem_object_plane_fd_ioctl
static int innogpu_drm_gem_plane_fd_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return inno_gem_object_plane_fd_ioctl(drm_dev, data, file);
}

// sync innodpu_drm_chipinfo_ioctl
static int innogpu_drm_chipinfo_ioctl(struct drm_device *drm_dev,
		void *data, struct drm_file *file)
{
	return innodpu_drm_chipinfo_ioctl(drm_dev, data, file);
}

static struct device_dma_parameters *innogpu_drm_get_dma_params(struct drm_device * drm_dev)
{
	return &(((struct pvr_drm_private *)innogpu_drm_to_pvr_private(drm_dev))->dma_parms);
}

static int innogpu_drm_component_bind(struct device *dev) //platform_device->device
{
	int ret = 0;
	bool drm_nulldisplay = false;
	struct innogpu_drm_private *drm_private = NULL;
	struct pvr_drm_private *pvr_priv = NULL;
	struct innodpu_drm_private *display_priv = NULL;
	struct drm_device *drm_dev = NULL;
	struct device *parent = dev->parent;
	struct platform_device *pdev = inno_to_platform_device(dev);
	plat_data_t *drm_plat_data = dev_get_platdata(dev);

	if (!parent) {
		innodpu_err(dev, "innogpu pcie device is  NULL.\n");
		return -EINVAL;
	}
#ifndef SUPPORT_DMA_TRANSFER
	if (hal_has_inv_mem(parent)) {
		innodpu_err(parent,"Error: Inno env not support dma, but card support inv memory.\n");
	} else {
		innodpu_err(parent,"Warn: Inno env not support dma, innodpu not clear/back/recover vram.\n");
	}
#endif
	if (hal_get_nulldisplay()) {
		drm_nulldisplay = true;
	} else {
		drm_nulldisplay = (drm_plat_data->dev_idx > 0) ? true : false;
	}

	innodpu_info(parent, DPU_UT_INIT, "Start init innosilicon gpu with %s. Version is 1.0.0\n",
			drm_nulldisplay ? "nulldisp" : "display");

	// drm_dev->dev needs pcie_dev->device
	drm_dev = drm_dev_alloc(&g_innogpu_drm_driver, parent);
	if (IS_ERR(drm_dev)) {
		innodpu_err(parent, "innogpu drm device alloc failed.\n");
		return PTR_ERR(drm_dev);
	}
	inno_platform_set_drvdata(pdev, drm_dev);	// platform_device pdev->dev->driver_data = drm_dev;

	drm_private = kzalloc(sizeof(struct innogpu_drm_private), GFP_KERNEL);
	if (!drm_private) {
		innodpu_err(parent, "Alloc dev_priv failed. Short of memory.\n");
		ret = -ENOMEM;
		goto err_drm_private_alloc;
	}
	drm_dev->dev_private = drm_private;

#if (DRM_VERSION < KERNEL_VERSION(5, 14, 0))
	drm_dev->pdev = drm_plat_data->pdev_rsrc->pdev;
#endif

	drm_private->drm_plat_data = drm_plat_data;
	drm_private->drm_nulldisplay = drm_nulldisplay;
	pvr_priv = innogpu_drm_to_pvr_private(drm_dev);
	display_priv = innogpu_drm_to_display_private(drm_dev);
	if (hal_has_inv_mem(parent))
		drm_private->drm_has_invmem = true;
	else
		drm_private->drm_has_invmem = false;
#ifdef SUPPORT_DMA_TRANSFER
	drm_private->drm_support_dmatrans = true;
#else
	drm_private->drm_support_dmatrans = false;
#endif

	innodpu_info(parent, DPU_UT_INIT, "Start early load innosilicon drm device.\n");
	ret = innodpu_early_load(drm_dev, display_priv, dev);
	if (ret) {
		innodpu_err(parent, "Early load innosilicon drm device Failed.\n");
		goto err_drm_early_load;
	}

	innodpu_info(parent, DPU_UT_INIT, "Start bind component device.\n");
	ret = component_bind_all(dev, drm_dev);
	if (ret) {
		innodpu_err(parent, "Bind all innosilicon dpu device Failed.\n");
		goto err_drm_bind_all;
	}

	if (!dev->dma_parms)
		dev->dma_parms = innogpu_drm_get_dma_params(drm_dev); // TBD: &priv->dma_parms;
	dma_set_max_seg_size(dev, DMA_BIT_MASK(32)); // TBD: maybe risk!!!

	innodpu_info(parent, DPU_UT_INIT, "Start late load innosilicon drm device.\n");
	ret = innodpu_late_load(drm_dev, display_priv);
	if (ret) {
		innodpu_err(parent, "Late load innosilicon drm device Failed.\n");
		goto err_drm_late_load;
	}
	innogpu_drm_set_irq(drm_dev, true);

	ret = drm_dev_register(drm_dev, 0);
	if (ret) {
		innodpu_err(parent, "Drm device register Failed.\n");
		goto err_drm_dev_register;
	}

	ret = innodpu_fb_init(drm_dev, display_priv);
	if (ret) {
		innodpu_err(parent, "Drm device register Failed.\n");
		goto err_fb_init;
	}

	ret = innodpu_pm_init(drm_dev, display_priv);
	if (ret) {
		innodpu_err(parent, "Drm device register Failed.\n");
		goto err_pm_init;
	}

	return 0;

err_pm_init:
	innodpu_fb_fini(drm_dev, display_priv);
err_fb_init:
	drm_atomic_helper_shutdown(drm_dev);
	drm_dev_unregister(drm_dev);
err_drm_dev_register:
	innodpu_late_unload(drm_dev, display_priv);
err_drm_late_load:
	component_unbind_all(dev, drm_dev);
err_drm_bind_all:
	innodpu_early_unload(drm_dev, display_priv);
err_drm_early_load:
	kfree(drm_private);
	drm_private = NULL;
	drm_dev->dev_private = NULL;
err_drm_private_alloc:
	drm_dev_put(drm_dev);		// drm_dev_put = drm_dev_fini + kfree(drm_dev)
	inno_platform_set_drvdata(pdev, NULL);

	return ret;
}

static void innogpu_drm_component_unbind(struct device *dev)
{
	struct platform_device *pdev = inno_to_platform_device(dev);
	struct drm_device *drm_dev = inno_platform_get_drvdata(pdev);
	struct innogpu_drm_private *drm_private = drm_dev->dev_private;
	struct innodpu_drm_private *display_priv = NULL;

	display_priv = innogpu_drm_to_display_private(drm_dev);
	if (display_priv) {
		innodpu_pm_fini(drm_dev, display_priv);
		innodpu_fb_fini(drm_dev, display_priv);
		drm_atomic_helper_shutdown(drm_dev);
		drm_dev_unregister(drm_dev);
		innodpu_late_unload(drm_dev, display_priv);
		component_unbind_all(dev, drm_dev);
		innodpu_early_unload(drm_dev, display_priv);

		kfree(drm_private);
		drm_dev->dev_private = NULL;
	}
	drm_dev_put(drm_dev);		// drm_dev_put = drm_dev_fini + kfree(drm_dev)
	inno_platform_set_drvdata(pdev, NULL);
}

static int innogpu_drm_compare_dev(struct device *dev, void *data)
{
	return (dev == (struct device *)data);
}

/*
 * TBD!!!
 * dev ：sub device
 * data : struct innogpu_drm_match * infomation。
 * tmatch.id : drm index
 * tmatch.name : s_innogpu_drm_sub_drivers subdevice platform name
 * tmatch.subdev:  not used
 * tmatch.pdev : pcie device ( current dev->parent)
 * return 0 if the device doesn't match and non-zero if it does
 */
static int innogpu_drm_match_dev(struct device *dev, void *data)
{
	struct platform_device *pdev = inno_to_platform_device(dev);
	struct innogpu_drm_match *tmatch = data;
	plat_data_t *plat_data = dev_get_platdata(dev);
	int reg_vkms_nums = hal_get_dev_nums(tmatch->pdev, DEV_VKMS);
	int nulldisplay_pipe_num = hal_get_nulldisplay_drm_pipe_num();
	bool nulldisplay = hal_get_nulldisplay();
	int display_nums = hal_get_dev_nums(tmatch->pdev, DEV_DPU) - reg_vkms_nums;
	bool display_device = false;
	bool pvr_device = false;
	int ret = 0;
	int did = 0;

	/* filter out non-drm devices */
	if (IS_ERR_OR_NULL(plat_data) || strcmp(pdev->name, tmatch->name)) {
		return 0;
	}

#ifdef CONFIG_DRM_INNO_AUDIO
	if(!strcmp(pdev->name, INNO_AUDIO_DEVICE_NAME))
		return 1;
#endif

	display_device = !strcmp(pdev->name, INNO_DPU_DEVICE_NAME);
	pvr_device = !strcmp(pdev->name, INNO_GPU_DEVICE_NAME);

	/* pvr */
	if (pvr_device) {
		did = plat_data->dev_idx;
		if (did == tmatch->id) {
			return 1;
		} else {
			return 0;
		}
	}

	/* display */
	if (nulldisplay) { /* nulldisplay crtc + vkms */
			did = plat_data->dev_idx;
			if (did >= 0 && ((did / nulldisplay_pipe_num) == tmatch->id))
				ret = 1;
	} else {
		if (tmatch->id > 0) { /* nulldisplay drm_device */
			if (display_device) /*dpu device*/
				did = plat_data->dev_idx - display_nums;
			else	/*vkms device*/
				did = plat_data->dev_idx;
			if ((did >= 0) && ((did / nulldisplay_pipe_num) == (tmatch->id - 1)))
				ret = 1;
		} else {
			ret = 1;
			if (display_device) { /* nulldisplay / display drm_device */
				if (plat_data->dev_idx < display_nums)	/* really display dpu_device */
					ret = 1;
				else /* nulldisplay dpu_device */
					ret = 0;
			}
		}
	}

	innodpu_info(dev, DPU_UT_INIT , "drm_device-%d plat_index-%d. %s-%s match:%d\n",
			tmatch->id, plat_data->dev_idx, pdev->name, tmatch->name, ret);

	return ret;
}


int innogpu_drm_irq_enabled(struct drm_device *drm_dev, bool *enable)
{
	__maybe_unused struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private)
		return -EINVAL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0))
	drm_private = drm_dev->dev_private;
	*enable = drm_private->irq_enabled;
#else
	*enable = drm_dev->irq_enabled;
#endif

	return 0;
}

/*
 * struct platform_device *pdev : drm platform device
 */
static int innogpu_drm_probe(struct platform_device *pdev)
{
	struct component_match *match = NULL;
	struct device *dev = &pdev->dev;
	int i;
	struct device *tmpdev = NULL;
	struct device *subdev = NULL;
	struct platform_driver *drv = NULL;
	struct innogpu_drm_match tmatch;

	plat_data_t *drm_plat_data = dev_get_platdata(dev);
	plat_data_t *dev_plat_data = NULL;

	innodpu_info(&pdev->dev, DPU_UT_INIT, "Git   Current   Branch: %s \n", GIT_BRANCH);
	innodpu_info(&pdev->dev, DPU_UT_INIT, "Git   Current   CommitID: %s \n", GIT_COMMIT_ID);
	innodpu_info(&pdev->dev, DPU_UT_INIT, "Git   Current   CommitID (dirty) : %s \n", GIT_DIRTY);

	{
		struct hw_board_info board1 = HW_BOARD_INFO_ITEM("ZY", "YF27_1"); //VGA+HDMI2VGA+DP2LVDS
		struct hw_board_info board2 = HW_BOARD_INFO_ITEM("ZY", "YF27_2"); //VGA+HDMI2VGA+DP2VGA

		if (innodpu_is_odm_pcb_match(dev->parent, &board1) || innodpu_is_odm_pcb_match(dev->parent, &board2)) {
			innodpu_info(&pdev->dev, DPU_UT_INIT, "match odm[ZY] pcb\n");
			s_dpu_driver_group = 1;
		}
	}

	/* match compnoent device */
	for (i = 0; i < SUB_DRIVER_GROUP_NUMBER; i++) {
		drv = s_innogpu_drm_sub_drivers[s_dpu_driver_group][i];
		if (!drv)
			continue;

		tmpdev = NULL;
		tmatch.id = drm_plat_data->dev_idx;
		tmatch.name = drv->driver.name;
		tmatch.pdev = dev->parent;

		innodpu_info(&pdev->dev, DPU_UT_INIT, "loop-%d, drm_plat_data is %pK, tmatch id %d name-%s",
					i, drm_plat_data, tmatch.id, tmatch.name);

		if (tmatch.id > 0 || hal_get_nulldisplay()) {
			if (((&g_innogpu_vkms_driver != drv)) && (&g_innogpu_dpu_driver != drv) && (&g_innogpu_pvrsrvkm_driver != drv))
					continue;
		} else {
			if (((&g_innogpu_hdmi_driver == drv)) && (s_hdmi_nums == 0)) {
				innodpu_info(&pdev->dev, DPU_UT_INIT, "hdmi_disable\n");
				continue;
			}

			if ((&g_innogpu_dp_driver == drv) && (s_dp_nums == 0)) {
				innodpu_info(&pdev->dev, DPU_UT_INIT, "dp_disable\n");
				continue;
			}

			if (((&g_innogpu_vga_driver == drv)) && (s_vga_nums == 0)) {
				innodpu_info(&pdev->dev, DPU_UT_INIT, "vga_disable\n");
				continue;
			}

			if ((&g_innogpu_vkms_driver == drv))
				continue;
		}

		do {
			subdev = bus_find_device(&platform_bus_type,
					tmpdev, &tmatch, (void *)innogpu_drm_match_dev);
			put_device(tmpdev);
			tmpdev = subdev;
			if (!subdev) {
				break;
			}

			dev_plat_data = (plat_data_t *)dev_get_platdata(subdev);
			innodpu_info(&pdev->dev, DPU_UT_INIT, "%s%d, pdev_idx:%d, dev_plat_data %p\n", drv->driver.name,
					dev_plat_data->dev_idx, dev_plat_data->pdev_idx, dev_plat_data);

			if (drm_plat_data->pdev_idx != dev_plat_data->pdev_idx) {
				continue;
			}
			innodpu_info(&pdev->dev, DPU_UT_INIT, "%s%d, match.\n",
					drv->driver.name, dev_plat_data->dev_idx);
			device_link_add(dev, subdev, DL_FLAG_STATELESS);
			component_match_add(dev, &match, innogpu_drm_compare_dev, subdev);
		} while (true);
	}
	if (match == NULL || dev == NULL) {
		if (match == NULL)
			DRM_ERROR("match is NULL\n");
		else
			DRM_ERROR("dev is NULL\n");

		return 0;
	}
	return component_master_add_with_match(dev, &s_innogu_drm_component_ops, match);
}

static int innogpu_drm_remove(struct platform_device *pdev)
{
	component_master_del(&pdev->dev, &s_innogu_drm_component_ops);

	return 0;
}
static void platform_subdrivers_unload(void)
{
	int i;

	for (i = SUB_DRIVER_GROUP_NUMBER - 1; i >= 0; i--) {
		if (s_innogpu_drm_sub_drivers[0][i] != NULL) {
			platform_driver_unregister(s_innogpu_drm_sub_drivers[0][i]);
		}
	}
}
static struct platform_driver s_innogpu_drm_platform_driver = {
	.probe = innogpu_drm_probe,
	.remove = innogpu_drm_remove,
	.driver = {
		.owner = THIS_MODULE,
		.name = INNO_DRM_DEVICE_NAME,
		.pm = &innogpu_drm_pm_ops, // replace innodpu_drm_pm_ops
	},
};

#if defined(NO_HARDWARE) || !defined(CONFIG_DRM_INNO_SRVKM)
static int __init innogpu_drm_init(void)
#else
int innogpu_drm_init(void)
#endif
{
	int i = 0;
	int ret = 0;
	ret = PVRSRVDriverInit(); // TBD: support for multi-core
	if (ret)
		return ret;
	// TBD: pvr_init ???  PVRSRVDriverInit, need move to probe
	for (i = 0; i < SUB_DRIVER_GROUP_NUMBER; i++) {
		if (s_innogpu_drm_sub_drivers[0][i] != NULL) {
			platform_driver_register(s_innogpu_drm_sub_drivers[0][i]);
		}
	}

	ret = platform_driver_register(&s_innogpu_drm_platform_driver);
	if (ret) {
		goto err;
	}
	return 0;

err:
	platform_subdrivers_unload();
	PVRSRVDriverDeinit();
	return ret;
}

#if defined(NO_HARDWARE) || !defined(CONFIG_DRM_INNO_SRVKM)
static void __exit innogpu_drm_exit(void)
#else
void innogpu_drm_exit(void)
#endif
{
	PVRSRV_DATA *psPVRSRVData = PVRSRVGetPVRSRVData();
	DRM_DEBUG_DRIVER("\n");
	psPVRSRVData->bUnload = IMG_TRUE;
	platform_driver_unregister(&s_innogpu_drm_platform_driver);
	platform_subdrivers_unload();
	PVRSRVDriverDeinit();
	DRM_DEBUG_DRIVER("done\n");
}

#if defined(NO_HARDWARE) || !defined(CONFIG_DRM_INNO_SRVKM)
late_initcall(innogpu_drm_init);
module_exit(innogpu_drm_exit);

MODULE_LICENSE("Dual MIT/GPL");
MODULE_AUTHOR(INNOGPU_DRM_DRIVER_AUTHOR);
MODULE_DESCRIPTION(INNOGPU_DRM_DRIVER_DESC);
#endif

/*************************************
*innogpu_get_drm_from_pdev
*let platform_device->dev convert to drm_device
*struct device : platfrom_device->dev
*return 		: struct drm_device * ddev or NULL
**************************************/
struct drm_device * innogpu_get_drm_from_pdev(struct device *plat_dev)
{
	plat_data_t *drm_plat_data = NULL;
	if (!plat_dev)
		return NULL;
	drm_plat_data = inno_dev_get_platdata(plat_dev);
	
	if (!drm_plat_data)
		return NULL;
	else
		return  drm_plat_data->pdev_rsrc->ddev;
}

/****************
 *innogpu_get_ddev_from_plat_dev
 *let platform_device->dev convert to drm_device
 *struct device : platfrom_device->dev
 *return : drm_device->dev or NULL
 * */
void * innogpu_get_ddev_from_plat_dev(struct device * plat_dev)
{
	plat_data_t *drm_plat_data = NULL;
	struct drm_device * ddev = NULL;
	if (!plat_dev)
		return NULL;
	drm_plat_data = inno_dev_get_platdata(plat_dev);
	
	if (!drm_plat_data)
		return NULL;
	else {
		ddev = drm_plat_data->pdev_rsrc->ddev;
		return ddev ? ddev->dev : NULL;
	}
}

int innogpu_drm_set_irq(struct drm_device *drm_dev, bool enable)
{
	__maybe_unused struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private)
		return -EINVAL;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 15, 0))
	drm_private = drm_dev->dev_private;
	drm_private->irq_enabled = enable;
#else
	drm_dev->irq_enabled = enable;
#endif

	return 0;
}

int innogpu_drm_get_nulldisp(struct drm_device *drm_dev, bool *nulldisp)
{
	struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private)
		return -EINVAL;

	drm_private = drm_dev->dev_private;
	*nulldisp = drm_private->drm_nulldisplay;

	return 0;
}

int innogpu_drm_has_invmem(struct drm_device *drm_dev, bool *has_invmem)
{
	struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private)
		return -EINVAL;

	drm_private = drm_dev->dev_private;
	*has_invmem = drm_private->drm_has_invmem;

	return 0;
}

int innogpu_drm_get_platdata(struct drm_device *drm_dev, plat_data_t **drm_plat_data)
{
	__maybe_unused struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private)
		return -EINVAL;

	drm_private = drm_dev->dev_private;
	*drm_plat_data = drm_private->drm_plat_data;

	return 0;
}

void * innogpu_drm_to_pvr_private(struct drm_device *drm_dev)
{
	__maybe_unused struct innogpu_drm_private *drm_private = NULL;

	if (!drm_dev->dev_private) {
		inno_printk(KERN_ERR "%s %d  drm_dev:%p dev_private: %p \n", __func__, __LINE__, drm_dev, drm_dev->dev_private);
		return NULL;
	}
	drm_private = drm_dev->dev_private;
	return &(drm_private->pvr_priv);
}

void * innogpu_drm_to_display_private(struct drm_device *drm_dev)
{
	__maybe_unused struct innogpu_drm_private *drm_private = NULL;

	if (drm_dev == NULL) {
		inno_printk(KERN_ERR "%s %d  drm_dev:%p \n", __func__, __LINE__, drm_dev);
		return NULL;
	}

	if (!drm_dev->dev_private) {
		inno_printk(KERN_ERR "%s %d  drm_dev:%p dev_private: %p \n", __func__, __LINE__, drm_dev, drm_dev->dev_private);
		return NULL;
	}

	drm_private = drm_dev->dev_private;
	return &(drm_private->display_priv);
}
