
#include <linux/version.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0))
#include <drm/drm_drv.h>
#include <drm/drm_print.h>
#include <linux/mod_devicetable.h>
#include <linux/dma-mapping.h>
#include <linux/of.h>
#include <linux/slab.h>
#else
#include <drm/drmP.h>
#endif

#include <linux/module.h>
#include <linux/platform_device.h>

#include "module_common.h"
#include "pvr_drv.h"
#include "pvrmodule.h"
#include "sysinfo.h"
#include "pvrsrv_error.h"
#include "pvrsrv.h"

#include <linux/component.h>
#include "innogpu_drm.h"

/* This header must always be included last */
#include "kernel_compatibility.h"

#if defined(SUPPORT_DISPLAY_CLASS)
/* Display class interface */
#include "kerneldisplay.h"
EXPORT_SYMBOL(INNOGPU_DCRegisterDevice);
EXPORT_SYMBOL(INNOGPU_DCUnregisterDevice);
EXPORT_SYMBOL(INNOGPU_DCDisplayConfigurationRetired);
EXPORT_SYMBOL(INNOGPU_DCDisplayHasPendingCommand);
EXPORT_SYMBOL(INNOGPU_DCImportBufferAcquire);
EXPORT_SYMBOL(INNOGPU_DCImportBufferRelease);

/* Physmem interface (required by LMA DC drivers) */
#include "physheap.h"
EXPORT_SYMBOL(PhysHeapAcquireByUsage);
EXPORT_SYMBOL(INNOGPU_PhysHeapRelease);
EXPORT_SYMBOL(INNOGPU_PhysHeapGetType);
EXPORT_SYMBOL(INNOGPU_PhysHeapGetCpuPAddr);
EXPORT_SYMBOL(INNOGPU_PhysHeapGetSize);
EXPORT_SYMBOL(INNOGPU_PhysHeapCpuPAddrToDevPAddr);

EXPORT_SYMBOL(INNOGPU_PVRSRVGetDriverStatus);
EXPORT_SYMBOL(INNOGPU_PVRSRVSystemInstallDeviceLISR);
EXPORT_SYMBOL(INNOGPU_PVRSRVSystemUninstallDeviceLISR);

#include "pvr_notifier.h"
EXPORT_SYMBOL(INNOGPU_PVRSRVCheckStatus);

#include "pvr_debug.h"
EXPORT_SYMBOL(INNOGPU_PVRSRVGetErrorString);
EXPORT_SYMBOL(INNOGPU_PVRSRVGetDeviceInstance);
#endif /* defined(SUPPORT_DISPLAY_CLASS) */

#if defined(SUPPORT_RGX)
#include "rgxapi_km.h"
#if defined(SUPPORT_SHARED_SLC)
EXPORT_SYMBOL(INNOGPU_RGXInitSLC);
#endif
EXPORT_SYMBOL(INNOGPU_RGXHWPerfConnect);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfDisconnect);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfControl);
#if defined(HWPERF_PACKET_V2C_SIG)
EXPORT_SYMBOL(INNOGPU_RGXHWPerfConfigureCounters);
#else
EXPORT_SYMBOL(INNOGPU_RGXHWPerfConfigMuxCounters);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfConfigureAndEnableCustomCounters);
#endif
EXPORT_SYMBOL(INNOGPU_RGXHWPerfDisableCounters);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfAcquireEvents);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfReleaseEvents);
EXPORT_SYMBOL(INNOGPU_RGXHWPerfConvertCRTimeStamp);
#if defined(SUPPORT_KERNEL_HWPERF_TEST)
EXPORT_SYMBOL(INNOGPU_OSAddTimer);
EXPORT_SYMBOL(INNOGPU_OSEnableTimer);
EXPORT_SYMBOL(INNOGPU_OSDisableTimer);
EXPORT_SYMBOL(INNOGPU_OSRemoveTimer);
#endif
#endif

#include "inno_misc.h"
#include "inno_srvkm.h"
INNO_EXT_SYM(inno_pmr_unref);
INNO_EXT_SYM(inno_pmr_ref);
INNO_EXT_SYM(inno_pmr_lock_sys_phys_addr);
INNO_EXT_SYM(inno_pmr_unlock_sys_phys_addr);
INNO_EXT_SYM(inno_stats_remove_mem_alloc_record);
INNO_EXT_SYM(inno_stats_decr_mem_alloc_stat);
INNO_EXT_SYM(inno_pmr_write_bytes);
INNO_EXT_SYM(inno_pmr_read_bytes);


#if defined(__arm64__) || defined(__aarch64__) || defined(PVRSRV_DEVMEM_TEST_SAFE_MEMSETCPY)
#include "osfunc_common.h"

#if defined(__GNUC__)
/* Workarounds for assumptions made that memory will not be mapped uncached
 * in kernel or user address spaces on arm64 platforms (or other testing).
 * */

EXPORT_SYMBOL(INNOGPU_DeviceMemSet);
EXPORT_SYMBOL(INNOGPU_DeviceMemCopy);

#else /* defined __GNUC__ */

EXPORT_SYMBOL(INNOGPU_DeviceMemSetBytes);
EXPORT_SYMBOL(INNOGPU_DeviceMemCopyBytes);

#endif /* defined __GNUC__ */

#endif

#include "syscommon.h"
INNO_EXT_SYM(gpu_extern_vram_flag);
#include "osfunc.h"
INNO_EXT_SYM(OSMemoryBarrier);

/* This header must always be included last */
#include "kernel_compatibility.h"
/* Values used to configure the PVRSRV_DEVICE_INIT_MODE tunable (Linux-only) */
#define PVRSRV_LINUX_DEV_INIT_ON_PROBE   1
#define PVRSRV_LINUX_DEV_INIT_ON_OPEN    2
#define PVRSRV_LINUX_DEV_INIT_ON_CONNECT 3

static int innogpu_pvr_bind(struct device *dev,
		struct device *master, void *data);
static void innogpu_pvr_unbind(struct device *dev,
		struct device *master, void *data);

static DEFINE_MUTEX(g_device_mutex);

extern struct platform_driver g_innogpu_pvrsrvkm_driver;
static const struct component_ops s_innogpu_pvr_ops = {
	.bind = innogpu_pvr_bind,
	.unbind = innogpu_pvr_unbind,
};

static int innogpu_pvr_drm_load(struct device *dev, unsigned long flags)
{
	struct platform_device *pdev = inno_to_platform_device(dev);
	plat_data_t *pdata = inno_platform_get_data(pdev);
	struct drm_device *drm_dev = NULL;
	struct pvr_drm_private *priv = NULL;
	enum PVRSRV_ERROR_TAG srv_err;
	int err, deviceId;

#if defined(NO_HARDWARE)
	drm_dev = inno_platform_get_drvdata(pdev);
	pdata->pdev_rsrc = NULL;
#else
	drm_dev = pdata->pdev_rsrc->ddev;
#endif
	if (drm_dev == NULL)
		DRM_ERROR("%s %d drm_dev is NULL\n", __func__, __LINE__);

	priv = innogpu_drm_to_pvr_private(drm_dev);
	BUG_ON(!priv);
	DRM_DEBUG_DRIVER("device %p\n", drm_dev->dev);

	/*
	dev_set_drvdata(drm_dev->dev, drm_dev); // dropped, drm_dev->dev is pcie device, foribidden to use.
	*/

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3, 12, 0))
	/*
	 * Older kernels do not have render drm_minor member in drm_device,
	 * so we fallback to primary node for device identification
	 */
	deviceId = drm_dev->primary->index;
#else
	if (drm_dev->render)
		deviceId = drm_dev->render->index;
	else /* when render node is NULL, fallback to primary node */
		deviceId = drm_dev->primary->index;
#endif

	/* dropped
	priv = kzalloc(sizeof(*priv), GFP_KERNEL);
	if (!priv) {
		err = -ENOMEM;
		goto err_exit;
	}
	drm_dev->dev_private = priv;

	// TBD ???
	if (!drm_dev->dev->dma_parms)
		drm_dev->dev->dma_parms = &priv->dma_parms;
	dma_set_max_seg_size(drm_dev->dev, DMA_BIT_MASK(32));
	*/

	mutex_lock(&g_device_mutex);

	srv_err = PVRSRVCommonDeviceCreate(dev, deviceId, &priv->dev_node);
	if (srv_err != PVRSRV_OK) {
		DRM_ERROR("failed to create device node for device %p (%s)\n",
			  drm_dev->dev, INNOGPU_PVRSRVGetErrorString(srv_err));
		if (srv_err == PVRSRV_ERROR_PROBE_DEFER)
			err = -EPROBE_DEFER;
		else
			err = -ENODEV;
		goto err_unset_dma_parms;
	}
	err = PVRSRVDeviceInit(priv->dev_node);
	if (err) {
		DRM_ERROR("device %p initialisation failed (err=%d)\n",
			  drm_dev->dev, err);
		goto err_device_destroy;
	}
	/*
	drm_mode_config_init(drm_dev); // dropped
	*/

#if defined(SUPPORT_FWLOAD_ON_PROBE)
	srv_err = PVRSRVCommonDeviceInitialise(priv->dev_node);
	if (srv_err != PVRSRV_OK) {
		err = -ENODEV;
		DRM_ERROR("device %p initialisation failed (err=%d)\n",
			  drm_dev->dev, err);
		goto err_device_deinit;
	}
#endif

	mutex_unlock(&g_device_mutex);

	return 0;

#if defined(SUPPORT_FWLOAD_ON_PROBE)
err_device_deinit:
	/*
	drm_mode_config_cleanup(drm_dev); // dropped
	*/
	PVRSRVDeviceDeinit(priv->dev_node);
#endif
err_device_destroy:
	PVRSRVCommonDeviceDestroy(priv->dev_node);
err_unset_dma_parms:
	mutex_unlock(&g_device_mutex);
	/* dropped
	if (drm_dev->dev->dma_parms == &priv->dma_parms)
		drm_dev->dev->dma_parms = NULL;
	kfree(priv);

err_exit:
	*/
	return err;
}


static int innogpu_pvr_drm_unload(struct drm_device *drm_dev)
{
	struct pvr_drm_private *priv = innogpu_drm_to_pvr_private(drm_dev);

	BUG_ON(!priv);

	DRM_DEBUG_DRIVER("device %p\n", drm_dev->dev);

	/*
	drm_mode_config_cleanup(drm_dev); // dropped
	*/

	PVRSRVDeviceDeinit(priv->dev_node);

	mutex_lock(&g_device_mutex); // TBD???
	PVRSRVCommonDeviceDestroy(priv->dev_node);
	mutex_unlock(&g_device_mutex);

	/* dropped
	if (drm_dev->dev->dma_parms == &priv->dma_parms)
		drm_dev->dev->dma_parms = NULL;

	kfree(priv);
	drm_dev->dev_private = NULL;
	*/
	return 0;
}

/**
 * innogpu_pvr_bind - innosilicon pvr-driver initialization function
 * sync from function pvr_probe
 * @dev:
 * parent(dev->parent) is &pci_dev.dev
 * @master: component master
 * @data: point of struct drm_device
 *
 * This function initializes the innosilicon pvr device
 *
 * Returns:
 * 0 if it is OK, errno otherwise.
 */
static int innogpu_pvr_bind(struct device *dev,
		struct device *master, void *data)
{
	struct drm_device *drm_dev = (struct drm_device *)data;
	struct platform_device *pdev = inno_to_platform_device(dev);
#if defined(NO_HARDWARE)
#else
	plat_data_t *pdata = inno_platform_get_data(pdev);
	pdata->pdev_rsrc->ddev = drm_dev;
	if (pdata->pdev_rsrc->pdev == NULL) {
		DRM_ERROR("pdev_rsrc->pdev pci_device is NULL");
	}
#endif
	inno_platform_set_drvdata(pdev, drm_dev);
	/*
	drm_dev->platformdev = pdev; // dropped
	*/
	BUG_ON(!data);
	return innogpu_pvr_drm_load(dev, 0);
}

/**
 *innogpu_pvr_unbind - innosilicon pvr-driver deinitialization function
 * sync from function pvr_remove
 *@dev:
 *parent(dev->parent) is &pci_dev.dev
 *@master: component master
 *@data: point of struct drm_device
 *
 */
 static void innogpu_pvr_unbind(struct device *dev,
		struct device *master, void *data)
{
	struct platform_device *pdev = inno_to_platform_device(dev);
	struct drm_device *drm_dev = (struct drm_device *)data;

	BUG_ON(!data);

	innogpu_pvr_drm_unload(drm_dev);
	inno_platform_set_drvdata(pdev, NULL);
}

/**
 *innogpu_pvr_unbind - sync from function pvr_init
 *
 */
static int innogpu_pvr_probe(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	DRM_DEBUG_DRIVER("\n");
	return component_add(&pdev->dev, &s_innogpu_pvr_ops);
}
#if 0
static void pvr_devices_unregister(void)
{
#if defined(MODULE) && !defined(PVR_LDM_PLATFORM_PRE_REGISTERED)
	unsigned int i;

	BUG_ON(!pvr_devices);

	for (i = 0; i < pvr_num_devices && pvr_devices[i]; i++)
		platform_device_unregister(pvr_devices[i]);

	kfree(pvr_devices);
	pvr_devices = NULL;
#endif /* defined(MODULE) && !defined(PVR_LDM_PLATFORM_PRE_REGISTERED) */
}
#endif
/**
 *innogpu_pvr_unbind - sync from function pvr_exit
 *
 */
static int innogpu_pvr_remove(struct platform_device *pdev)
{
	BUG_ON(!pdev);
	component_del(&pdev->dev, &s_innogpu_pvr_ops);
	return 0;
}


static void innogpu_pvr_shutdown(struct platform_device *pdev)
{
	struct drm_device *drm_dev = inno_platform_get_drvdata(pdev); //innogpu_drm_component_bind  inno_platform_set_drvdata(pdev, drm_dev);
	if (!drm_dev) {
		DRM_ERROR("device %p shutdown not bind\n", &pdev->dev);
		return;
	}
	
	DRM_DEBUG_DRIVER("device %p\n", &pdev->dev);

	PVRSRVDeviceShutdown(drm_dev);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0))
static const struct of_device_id innogpu_pvr_of_ids[] = {
#if defined(SYS_RGX_OF_COMPATIBLE)
	{ .compatible = SYS_RGX_OF_COMPATIBLE, },
#endif
	{},
};

#if !defined(CHROMIUMOS_KERNEL) || !defined(MODULE)
MODULE_DEVICE_TABLE(of, innogpu_pvr_of_ids);
#endif
#endif

static struct platform_device_id innogpu_pvr_platform_ids[] = {
#if defined(SYS_RGX_DEV_NAME)
	{ SYS_RGX_DEV_NAME, 0 },
#endif
	{ }
};

#if !defined(CHROMIUMOS_KERNEL) || !defined(MODULE)
MODULE_DEVICE_TABLE(platform, innogpu_pvr_platform_ids);
#endif

struct platform_driver g_innogpu_pvrsrvkm_driver = {
	.driver = {
		.name		= SYS_RGX_DEV_NAME,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0))
		.of_match_table	= of_match_ptr(innogpu_pvr_of_ids),
#endif
		.pm		= &pvr_pm_ops,
	},
	.id_table		= innogpu_pvr_platform_ids,
	.probe			= innogpu_pvr_probe,
	.remove			= innogpu_pvr_remove,
	.shutdown		= innogpu_pvr_shutdown,
};

MODULE_FIRMWARE("innogpu/rgx.fw.35.2.1632.23");
MODULE_FIRMWARE("innogpu/rgx.sh.35.2.1632.23");
MODULE_FIRMWARE("innogpu/rgx.fw.35.4.1632.23");
MODULE_FIRMWARE("innogpu/rgx.sh.35.4.1632.23");

